/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FUNCTIONS_H
#define FUNCTIONS_H

#include<Structures.h>
#include<map>
#include<iostream>
#include<cmath>
#include<map>
#include<algorithm>
//#include<unordered_map>
#include<Parameters.h>
#include <seqan/store.h>



//typedef  std::unordered_map<std::string, int> TAbundCoverMap;
//typedef  std::unordered_map<UInt, set<UInt> > TCOverSegmentMap;
typedef  std::map<std::string, int> TAbundCoverMap;
typedef  std::map<UInt, std::set<UInt> > TCoverSegmentMap;

template<typename T1, typename T2>
double rSquared(const T1& observed, const T2& pred)
{
  if (observed.size() != pred.size())
  {
    std::cerr << "calling rSquared with unequal length fields -- terminating" << std::endl;
    exit(1);
  }

  double ssTot = 0;
  double ssRes = 0;
  double obsMean = 0;

  size_t len = observed.size();

  typename T1::const_iterator itObs;
  typename T2::const_iterator itPred;
  for (itObs = observed.begin(), itPred = pred.begin(); itObs != observed.end(); ++itObs, ++itPred)
  {
    obsMean += *itObs;
    ssRes += (*itObs - *itPred) * (*itObs - *itPred);
  }
  obsMean /= len;

  for (itObs = observed.begin(); itObs != observed.end(); ++itObs)
  {
    ssTot += (*itObs - obsMean) * (*itObs - obsMean);
  }

  if (ssTot < 0.000001)
	  return 0;

  return 1. - (ssRes/ssTot);
}

/*!
 * \brief  Normalization function
 *
 *         This function computes the number of possible start positions in a transcript for a read (or mate pair reads) to fall into the given cover area
 *
 *
 * \param	node_cover    The cover for which the number of starting positions for a read are computed.
 * \param	node_cover    The corresponding transcript.
 * \param node_lengths  The lengths for each node id.
 * \param read_length   The read length.
 * \param insert_size   The insert size for mate pair reads.
 * \param single_paired_reads   Flag: If true and a single end cover is passed, then we sum up the number of positions for left and right ends of a mate
 *
 * \return The number of possible starting positions for a read or mate-pair in a given cover.
 */
float getNormFactor(const SegmentCover& node_cover,
                    const Transcript& transcript,
                    const TAllSegmentsMap& segments,
                    int read_length = Parameters::read_length_global,
                    int insert_size = Parameters::insert_size_global,
                    double s_dev = Parameters::s_dev,
                    bool single_paired_reads = false,
                    double cutoff = Parameters::normCutoff
                    );


double getNormalizedTranscriptLength (const Transcript& transcript,
                                      const TAllSegmentsMap& segments,
                                      int read_length = Parameters::read_length_global,
                                      int fragment_length = Parameters::insert_size_global,
                                      double s_dev = Parameters::s_dev,
                                      double cutoff = Parameters::normCutoff);


//float getNormFactor(const SegmentCover & node_cover, const Transcript &transcript, const std::map<UInt, Segment> &segments, int read_length, int insert_size = 0);
float getEmpiricalNormFactor(const SegmentCover & node_cover, const Transcript &transcript, const std::map<UInt, Segment> &segments, int read_length, int insert_size);

/*!
 * \brief  Check adjacency
 *
 *         This function checks whether the two segments i and j are contained in a locus and whether they are adjacent in the locus
 *
 *
 * \param	locus         The locus
 * \param	i,j           Segment ids
 * \return True if contained and adjacent, false otherwise
 */
inline bool isAdjacent(const std::vector<UInt>&locus,UInt i,UInt j)
{
  UInt tmp[2]={i,j};
  return (std::search(locus.begin(),locus.end(), tmp, tmp+2)!=locus.end());
}

/*!
 * \brief  computes probability of no mapping paired reads
 *
 *         Compute the probability, that for the given lengths w1, w2 and w2 of three segments s1 s2 s3,
 *         there are no paired reads such that the left starts in s1 and the right in segment 3 (From Feng, Li, Jiang RECOMB-Paper).
 *         This value is computed under the assumption that M million reads were mapped and the RPKM level of a transcript supporting the cover of s1, s2, s3 is alpha.
 *
 * \param	M             The number of mapped reads in million.
 * \param alpha         The assumed RPKM level.
 * \param w1, w2, w3    The lengths of the segments s1, s2 and s3.
 * \param mean_is       The mean insert size.
 * \param sd            The standard deviation of the insert size.
 * \return	            Probability of no reads mapping.
 */
float getNoHitProb(int M, float alpha, Size w1, Size w2, Size w3, float mean_is=200, float sd=20);


/*!
 * \brief  computes probability of no mapping single end reads
 *
 *         Compute the probability, that paired read map into one of the pos positions (From Feng, Li, Jiang RECOMB-Paper).
 *         This value is computed under the assumption that M million reads were mapped and the RPKM level of a corresponding transcript is alpha.
 *
 * \param	M             The number of mapped reads in million.
 * \param alpha         The assumed RPKM level.
 * \param pos           The number of possible mapping positions
 * \return	            Probability of no reads mapping. 
 */
float getNoHitProb(int M, float alpha, Size pos);


/*!
 * \brief  find all segment covers
 *
 *         For a given set of nodes (they actually define one locus) find all segment covers that belong to that locus and are supported by
 *         at least one transcript
 *
 * \param associated_covers      Holds the computed associated covers.
 * \param	all_covers             The set of all covers.
 * \param index                  Maps covers to segments (key: segment id. value: set of covers containing the segment).
 * \param transcripts            Vector of transcripts.
 * \param segments               The segments defining the locus (in topological order) 
 */
void findAssociatedCovers(std::vector<SegmentCover> &associated_covers, const std::vector<SegmentCover> &all_covers, const std::map<UInt, std::set<UInt> > &index, const std::vector<Transcript>&transcripts, const std::vector<UInt> &segments);

/*!
 * \brief  find all segment covers
 *
 *         For a given set of nodes (they actually define one locus) find all segment covers that belong to that locus.
 *
 * \param associated_covers      Holds the computed associated covers.
 * \param	all_covers             The set of all covers.
 * \param index                  Maps covers to segments (key: segment id. value: set of covers containing the segment).
 * \param segments               The segments defining the locus (in topological order)
 */
void findAssociatedCovers(std::vector<SegmentCover> &associated_covers, const std::vector<SegmentCover> &all_covers, const std::map<UInt, std::set<UInt> > &index, const std::vector<UInt> &segments);

/*!
 * \brief  check transcript support of segment cover
 *
 *         Check whether a given segment cover is supported by at least one transcript.
 *
 * \param cover                  The segment cover.
 * \param	transcripts            The set of transcripts. 
 * \return boolean value, true if supported false otherwise
 */
bool isCoverSupported(const SegmentCover &cover, const std::vector<Transcript>&transcripts);

/*!
 * \brief  check locus support of segment cover
 *
 *         Check whether a given segment cover is supported by a given locus.
 *         The cover is supported if all of its contained segments also appear in the locus and their order is preserved
 *
 * \param cover                  The segment cover.
 * \param	segments               The set of segments defining the locus.
 * \return boolean value, true if supported false otherwise
 */
bool isCoverSupported(const SegmentCover &cover, const std::vector<UInt>&segments);

///*!
// * \brief  check segment cover support of transcript
// *
// *         Check whether a given transcript is supported by a set of segment covers.
// *         The transcript is supported if all its segments appear in some node cover, and all non-adjacent junctions
// *         are also covered or have a noHitProb value above 5%.
// *
// * \param all_covers            The global set of segment covers.
// * \param selected_cover_ids    The indices of the covers that are selected for checking the support
// * \param	transcript            The transcript.
// * \return boolean value, true if supported false otherwise
// */
bool isTranscriptSupported(const std::vector<SegmentCover> &all_covers,
                           const std::vector<UInt> &selected_cover_ids,
                           const Transcript &transcript
                           /*const std::map<std::pair<UInt, UInt>, bool>&adjacent*/);

/*!
 * \brief  generate all possible transcripts (DEPRECATED -- USE appendValidTranscripts INSTEAD!)
 *
 *         Generate the complete set of possible transcripts for a set of segments. Does not care about biological sense.
 *
 * \param segments               Vector of segment ids.
 * \param transcripts            Vector of generated transcripts.
 */
void generate_all_transcripts(const std::vector<UInt>&locus, std::vector<Transcript> &transcripts);

/*!
 * \brief  Generate all valid transcripts for a certain locus. Validity according to observed segment junctions in the segment covers
 *
 *         Generate all valid transcripts for a certain locus. Validity according to observed segment junctions in the segment covers.
 *
 * \param locus                 Locus of interes
 * \param all_covers            All segment covers
 * \param selected_cover_ids    Id's of the selected segment covers
 * \param transcripts           Vector of generated transcripts.
 */
void appendValidTranscripts(const std::vector<UInt> &locus,
                            const std::vector<SegmentCover> &all_covers,
                            const std::vector<UInt> &selected_cover_ids,
                            std::vector<Transcript> &transcripts);

void generateFakeCovers(const std::vector<SegmentCover>& all_covers,
                        const std::vector<UInt>& cover_ids,
                        const std::vector<Transcript>&transcripts,
                        std::vector<SegmentCover>& fake_covers,
                        Parameters::FakeMode mode);

void generateFakeCovers(const std::vector<UInt> &locus,
                        const std::vector<SegmentCover>& all_covers,
                        const std::vector<UInt>& cover_ids,
                        const TAllSegmentsMap & all_segments,
                        std::vector<SegmentCover>& fake_covers,
                        const SpliceGraph& sg,
                        Parameters::FakeMode mode,
                        const std::pair<unsigned, unsigned> &fragLenBounds,
                        UInt end_space = 1,                        
                        UInt maxUncovered = -1u,
                        UInt maxExonSkip = -1u,
                        const SpliceGraph::TLegalExonBoundMap &legalBlockBounds = SpliceGraph::TLegalExonBoundMap()); //if nonEmpty is passed function is exonAware


void computeSingleRange(const std::vector<UInt>& nodes_cover,
                        const Transcript& transcript,
                        const TAllSegmentsMap& segments,
                        int read_length,
                        std::pair<int, int>& range);

bool isTranscriptSupported(const std::vector<SegmentCover> &all_covers,
                           const std::vector<UInt> &selected_cover_ids,
                           const Transcript& transcript);

bool isTranscriptSupported(const std::vector<SegmentCover>& all_covers,
                           const std::vector<UInt>& selected_cover_ids,
                           const Transcript& transcript,
                           const TAllSegmentsMap& allSegments,
                           const std::vector<UInt>& locus);


float getCoverLength(const SegmentCover& node_cover, const TAllSegmentsMap& segments, int read_length);

void addAssociatedCoverIds(std::set<UInt>& cover_ids, const std::vector<SegmentCover> &all_covers, const std::map<UInt, std::set<UInt> > &index, const std::vector<Transcript>&transcripts, const std::vector<UInt> &node_ids);

void findAssociatedCoversIds(std::set<UInt>& cover_ids, const std::vector<SegmentCover> &all_covers, const std::map<UInt, std::set<UInt> > &index, const std::vector<UInt> &node_ids);

std::ostream& operator<< (std::ostream &out, const SegmentCover &n);

void postproc_inf_trans(std::vector<Transcript>& input_trans,
                        std::map<UInt, double>& transcript_intens,
                        std::vector<Transcript>& output_trans,
                        std::vector<bool>& is_mandatory,
                        float thresh);

bool enumGapExonSets(const SegmentCover&, const TAllSegmentsMap&, const std::vector<UInt>&, std::vector<std::vector<bool> >&,
					 const std::map<UInt, std::set<UInt> > &legalBlockBounds = std::map<UInt, std::set<UInt> >(), int read_length = Parameters::read_length_global,
					 int fragment_length = Parameters::insert_size_global, double s_dev = Parameters::s_dev);

void addRecursiveGapExons(int, int, std::vector<std::pair<UInt,UInt> >&, std::vector<SegmentType>&, SegmentType, SegmentType, UInt, UInt, std::vector<bool>&, std::set<std::vector<bool> >&,
						  std::vector<UInt>::const_iterator low, const std::map<UInt, std::set<UInt> > &legalBlockBounds = std::map<UInt, std::set<UInt> >(), const std::set<UInt> &legal_ends = std::set<UInt>());

void stringToCover(SegmentCover &cover, const std::string &cover_string, double count);

void hasUnsupportedJunction(const std::vector<UInt>& locus,
                            const std::vector<SegmentCover>& allCovers,
                            const std::vector<UInt>& selectedCoverIds,
                            const TAllSegmentsMap& allSegments,
                            const std::vector<Transcript>& transcripts,
                            std::vector<UInt>& numUnsuppJunc,
                            UInt minGapLen,
                            bool inferNeighborEdges = false,
                            bool joinOnlyCoveredSubs = true);

//std::pair<int, int> computeSingleRange(const std::vector<UInt> & nodes_cover, const Transcript &transcript, const std::map<UInt, Segment> &segments, int read_length = Parameters::read_length_global, int insert_size = Parameters::insert_size_global);

//get for a candidate transcript the number of readcounts that can possibly be explained by that transcript respecting fragment length boundaries
void computeRefinedCounts (const SegmentCover& node_cover,
                       const std::vector<Transcript>& transcripts,
                       const TAllSegmentsMap& segments,
                       const std::pair<double,double>& fragLenBounds,
                       std::vector<std::pair<std::vector<UInt>, double> >& clusters,
                       int read_length);


void getTranscriptSupports (std::vector<std::pair<double, UInt> >& supports,
                              const std::vector<Transcript>& transcripts,
                              const TAllSegmentsMap& allSegments,
                              const std::vector<SegmentCover>& allCovers,
                              const std::vector<UInt>& selectedCoverIds,
                              int read_length = Parameters::read_length_global,
                              int fragment_length = Parameters::insert_size_global,
                              double s_dev = Parameters::s_dev);

void markMandatoryCandidates(const std::vector<Transcript>& infTranscripts,
                             const TAllSegmentsMap& allSegments,
                             std::vector<bool>& isMandatory);

#endif /*FUNCTIONS_H*/
