/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ILP_H_INCLUDED
#define ILP_H_INCLUDED

#include <Structures.h>
#include <map>

#include <seqan/graph_align.h>

#include <ilcplex/ilocplex.h>

/** @defgroup LP LP and ILP formulations for the transcript inference problem *
 *  @{
 */

/*!
 * \brief  ILP formulation
 *
 *         The ILP formulation with error variables
 *
 *
 * \param	transcripts   All possible transcripts
 * \param	covers        Segment covers
 * \param node_lengths  The lengths for each node id.
 * \param read_length   The read length.
 * \param insert_size   The insert size for mate pair reads.
 * \param de_novo Whether de novo isoform prediction is performed
 *
 */
int buildILP(const std::vector<Transcript>& transcripts,
             const std::vector<SegmentCover>& covers,
             const TAllSegmentsMap& segments,
             UInt read_length,
             UInt insert_size,
             bool de_novo);

int buildILP2(const std::vector<Transcript>&transcripts, const std::vector<UInt>&locus, const std::vector<SegmentCover>& covers_in, const std::map<UInt, Segment>&segments, UInt read_length, UInt insert_size, bool de_novo);

/** @defgroup
 *  @}
 */

//int buildILP(const std::vector<Transcript>&transcripts,
//             std::map<UInt, double>&,
//             const std::vector<SegmentCover>& covers,
//             const std::map<UInt, Segment> &segments,
//             UInt read_length,
//             UInt insert_size,
//             bool de_novo);

float buildILP(const std::vector<Transcript>&transcripts,
             std::map<UInt, double>& computed_intens,
             const std::vector<SegmentCover>& all_covers,
             const std::vector<UInt>& selected_cover_ids,
             const std::vector<SegmentCover>& fake_covers,
             const TAllSegmentsMap &segments,
             bool de_novo);

int doGlmnet(const std::vector<Transcript>&transcripts,
			 const std::vector<UInt>& locus,
        	 std::map<UInt, double>& computed_intens,
        	 const std::vector<SegmentCover>& all_covers,
        	 const std::vector<UInt>& selected_cover_ids,
        	 std::vector<bool>& cov_expl,
        	 const std::vector<SegmentCover>& fake_covers,
             const TAllSegmentsMap &segments,
        	 float& lambda,
        	 float& total_error,
        	 const std::vector<Transcript>& known_transcripts = std::vector<Transcript>());


int buildILP(const std::vector<Transcript>&transcripts,
             std::map<UInt, double>& computed_intens,
             const std::vector<SegmentCover>& all_covers,
             const std::vector<UInt>& selected_cover_ids,
             const std::vector<SuperSegmentCover>& all_super_covers,
             const std::map<UInt, std::set<UInt> > &supercovers_to_covers,
             const TAllSegmentsMap &segments,
             UInt read_length,
             UInt insert_size,
             bool de_novo);


int buildfullILP(std::map<UInt, std::vector<UInt> >& all_loci,
                 const std::map<UInt, std::vector<Transcript> >& known_transcripts,
                 std::map<UInt, double>& computed_intens,
                 const std::vector<SegmentCover>& covers_in,
                 const std::map<UInt, Segment> &segments,
                 UInt read_length,
                 UInt insert_size,
                 bool de_novo);

void colgen_se(std::vector<Transcript>& transcripts,
           const std::vector<UInt>& locus,
		       std::map<UInt, double>& computed_intens,
		       const std::vector<SegmentCover>& covers_in,
           const TAllSegmentsMap& segments,
		       UInt read_length);
		      
void colgen_pe(std::vector<Transcript>& transcripts,
             const std::vector<UInt>& locus,
	           std::map<UInt, double>& computed_intens,
	           const std::vector<SegmentCover>& covers_in,
             const TAllSegmentsMap& segments,
	           UInt read_length,
	           UInt insert_size,
             std::vector<Transcript>& init_sol);


int colgen_pe_pwlin(std::vector<Transcript>&transcripts,
				std::vector<Transcript>& new_transcripts,
	            const std::vector<UInt>&locus,
	            std::map<UInt, double>& computed_intens,
	            const std::vector<SegmentCover>& all_covers,
	            const std::vector<UInt>& selected_cover_ids,
	            std::vector<bool>& cov_expl,
	            const std::vector<SegmentCover>& fake_covers,
	            const TAllSegmentsMap &segments,
		    	std::vector<Transcript>&init_sol,
	            bool tss_tes,
	            SpliceGraph& sg,
	            const std::vector<pwQuadCoeff>& coeff_bp,
	            const float lp_error,
	            const std::map<UInt, std::set<UInt> > &legalBlockBounds = std::map<UInt, std::set<UInt> >() );

typedef seqan::IntervalAndCargo<UInt, UInt> TInterval;
typedef seqan::IntervalTree<UInt, UInt> TIntervalTree;

void get_locus_structure(const std::vector<UInt>& locus,
             	 	 	 const TAllSegmentsMap& segments,
						 const std::map<UInt, std::set<UInt> > &legalBlockBounds,
						 std::vector<UInt>& tss,
						 std::vector<UInt>& tes,
						 std::vector<std::vector<std::pair<UInt,UInt> > >& exon_cluster,
						 std::map<UInt, UInt>& locid2exidx,
						 std::vector<std::pair<UInt,UInt> >& intron_retentions);
						 //std::vector<std::vector<std::pair<UInt,UInt> > >& intret_exon_cluster,
						 //std::map<UInt, UInt>& locid2intretexidx);

void prepareGreedyPricing(const std::vector<UInt>& locus,
						  const std::vector<SegmentCover>& selected_covers,
						  const std::map<UInt, std::set<UInt> > &legalBlockBounds,
						  seqan::String<TInterval>& mate_intervals,							//from here result parameters
						  seqan::String<TInterval>& insert_intervals,
						  std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
						  std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate);

double greedyPricing(const std::vector<UInt>& locus,
		    	     const std::vector<SegmentCover>& selected_covers,
             const TAllSegmentsMap& segments,
				     const IloCplex& approxSolver,
				     IloRangeArray& covered,
				     const std::map<UInt, std::set<UInt> > &legalBlockBounds,
				     const std::vector<UInt>& tss,
				     const std::vector<UInt>& tes,
				     const std::vector<std::vector<UInt> >& exon_cluster,
				     const std::map<UInt,UInt>& locid2exidx,
				     const TIntervalTree& mate_tree,
				     const TIntervalTree& insert_tree,
				     Transcript& result_trans,
				     const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
				     const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate);


void find_best_exskip(const std::vector<UInt>& locus,
					  const std::vector<SegmentCover>& selected_covers,
            const TAllSegmentsMap& segments,
					  const IloCplex& approxSolver,
					  IloRangeArray& covered,
					  const std::vector<std::pair<UInt,UInt> >& meta_trans,
					  const std::vector<bool>& ex_skipped,
					  const std::map<UInt, double>& normfs,
					  const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
					  const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate,
					  double& highest_imprv,
					  UInt& best_exskip,
					  std::map<UInt, double>& new_normfs_4best);

void find_best_ass(const std::vector<UInt>& locus,
				   const std::vector<SegmentCover>& selected_covers,
           const TAllSegmentsMap& segments,
				   const std::map<UInt, UInt>& locid2exidx,
				   const std::vector<bool>& ex_skipped,
				   const std::vector<std::vector<UInt> >& exon_cluster,
				   const std::vector<UInt>::const_iterator tss_it,
				   const std::vector<UInt>::const_iterator tes_it,
				   const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
				   const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate,
				   const std::map<UInt,std::set<UInt> > &legalBlockBounds,
				   const std::vector<std::pair<UInt,UInt> >& meta_trans,
				   const std::map<UInt, double>& normfs,
				   const IloCplex& approxSolver,
				   IloRangeArray& covered,
				   double& highest_exswap_imprv,
				   UInt& best_exswap_cluster,
				   UInt& best_exswap_entry_idx,
				   UInt& best_exswap_exit,
				   std::map<UInt, double>& new_normfs_4best);

void get_relevant_covers_exskip(std::set<UInt>& relevant_covers,
								const std::vector<UInt>& locus,
								const std::vector<std::pair<UInt, UInt> >& meta_trans,
								UInt i,
								const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
								const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate,
                const TAllSegmentsMap& segments,
								const std::vector<bool>& ex_skipped);

void get_relevant_covers_ass(std::set<UInt>& relevant_covers,
							 const std::vector<UInt>& locus,
							 const std::vector<std::pair<UInt, UInt> >& meta_trans,
							 int rel_cluster_idx,
							 UInt new_left_boundary_idx,
							 UInt new_right_boundary,
							 const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& lmate,
							 const std::vector<std::vector<std::vector<std::pair<UInt,UInt> > > >& rmate,
               const TAllSegmentsMap& segments,
							 const std::vector<bool>& ex_skipped);

bool is_mate_compatible(const std::vector<UInt>& mate,
						UInt left_boundary,
						UInt right_boundary,
						UInt current_cluster,
						const std::map<UInt,UInt>& locid2exidx,
						const std::vector<UInt>& locus);

void createGraph(SpliceGraph &sg,
                 const std::vector<SegmentCover>& selectedCovers,
                 TAllSegmentsMap allSegments,
                 const std::pair<unsigned, unsigned> &fragLenBounds,
                 UInt maxUncovered,
                 UInt maxSkip,
                 bool exonAware);

	      

#endif // ILP_H_INCLUDED
