/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstdio>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <algorithm>
#include <Functions.h>
#include <FunctionsIO.h>
#include <FunctionsPreproc.h>
#include <FunctionsGraph.h>
#include <Ilp.h>
#include <Parameters.h>
#include <seqan/arg_parse.h>
#include <gsl/gsl_cdf.h>

using namespace std;

#define DO_REOPTIMIZATION
#define DO_REDISTRIBUTE_READS
#define REMOVE_NON_MAXIMAL
#define FORCE_MIN_SUPPORT


struct Options
{
  string cntFile;
  string gtfFile;
  //string knownExpressionFile;
  UInt locusId;
  UInt endspace;
  string outDir;
  Parameters::FakeMode fake;
  float theta;
  float epsilon;
  float pseudoCount;
  float expThreshold;
  UInt tsstes_thr;
  UInt max_locsize;
  UInt min_locsize;
  UInt max_nof_exskip;
  UInt maxUncovered;
  //string quadApproxCoeffFile;
  int max_nof_relaxed_edges;
  
  bool doColGen;
  //bool weights;
  bool singleLocus;
  //bool noSingle;
  //bool lp;
  bool supTrans;
  bool strongSupTrans;
  //bool pwLin;
  bool quadError;
  bool tss_tes;
  bool maxLoc;
  bool minLoc;
  bool eaw;
  bool absThreshold;
  bool useKnownTrans;
  bool useRefinedCounts;
  bool forceKnownTerminals;
  bool regroup;
};

seqan::ArgumentParser::ParseResult runParser(int argc, char const** argv, Options & option)
{
  using namespace seqan;
  // Setup ArgumentParser.
  ArgumentParser parser("CIDANE");
  
  addArgument(parser, seqan::ArgParseArgument(
                                              seqan::ArgParseArgument::STRING, ".gtf"));
  
  addArgument(parser, seqan::ArgParseArgument(
                                              seqan::ArgParseArgument::STRING, ".cnt"));

/*
  addOption(parser, seqan::ArgParseOption(
                                          "l", "locus", "locus id if not for all.", ArgParseArgument::INTEGER, "INT"));
*/
  
/*
  addOption(parser, seqan::ArgParseOption(
                                          "s", "supTrans", "only consider supported transcripts in LP."));
*/
/*
  addOption(parser, seqan::ArgParseOption(
                                          "ss", "strongSupTrans", "only consider supported transcripts in LP with the stronger criterion like in isoLasso."));
*/
  
/*
  addOption(parser, seqan::ArgParseOption(
                                          "w", "weights", "use weights."));
*/
  
  addSection(parser, "Required library options (will be estimated automatically in future versions)");
  
  addOption(parser, seqan::ArgParseOption(
                                          "r", "readlength", "read length", seqan::ArgParseArgument::INTEGER, "INT"));
  setRequired(parser, "readlength", true);
  
  addOption(parser, seqan::ArgParseOption(
                                          "fl", "fraglength", "fragment length", seqan::ArgParseArgument::INTEGER, "INT"));
  setRequired(parser, "fraglength", true);
  
  addOption(parser, seqan::ArgParseOption(
                                          "sd", "sdev", "standard deviation", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  setRequired(parser, "sdev", true);
  
  addOption(parser, seqan::ArgParseOption(
                                          "lib", "librarysize", "number of mapped fragments in mio (for fpkm calculation)", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  setRequired(parser, "librarysize", true);
 
/*
  addOption(parser, seqan::ArgParseOption(
                                          "z", "endspace", "minimal end space for fake covers", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "endspace", 1);
  setMinValue(parser, "endspace", "1");
*/
/*
  addOption(parser, seqan::ArgParseOption(
                                          "q", "quadError", "minimize quadratic error instead of absolut error."));
*/
/*
  addOption(parser, seqan::ArgParseOption(
                                          "p", "pwLinApprox", "approximate quadratic cost by piecwise linear cost."));
*/
/*
  addOption(parser, seqan::ArgParseOption(
                                          "a", "theta", "scalar to penalize transcript intensity.", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "theta", 50.0);
*/
  addSection(parser, "Miscellaneous options");
  
  addOption(parser, seqan::ArgParseOption(
                                          "gl", "groupLoci", "infer loci from mapping instead of gene_id in provided gtf. Attention, this option currently deactivates option useTranscriptInformation"));
  
  addOption(parser, seqan::ArgParseOption(
                                          "m", "max", "maximal size of loci considered.", seqan::ArgParseArgument::INTEGER, "INT"));
  
  addOption(parser, seqan::ArgParseOption(
                                          "th", "expThreshold", "threshold of predicted transcripts to be reported."
                                          "Either min fraction per gene or FPKM depending on parameter absThreshold", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "expThreshold", 0.1);
  setMinValue(parser, "expThreshold", "0");
  
  addOption(parser, seqan::ArgParseOption(
                                          "ta", "absThreshold", "use FPKM threshold instead of isoform fraction per gene"));
  
  addOption(parser, seqan::ArgParseOption(
                                          "nea", "notExonAware", "deactivate exon awareness"));
  
  addOption(parser, seqan::ArgParseOption(
                                          "uti", "useTranscriptInformation", "incorporate annotated transcripts into model"));


  //--------------- Options for faux segments ---------------
  
  addSection(parser, "Faux segment cover options");
  //addOption(parser, seqan::ArgParseOption(
  //                                        "f", "fakes", "use fake covers.", seqan::ArgParseArgument::STRING, "NONE/PAIRED"));
  //addDefaultValue(parser, "fakes", "PAIRED");
  
  addOption(parser, seqan::ArgParseOption("f", "noFaux", "deactivate faux covers."));
  
  addOption(parser, seqan::ArgParseOption(
                                          "e", "pseudo", "pseudo count for error normalization", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "pseudo", 1.0);
  
  addOption(parser, seqan::ArgParseOption(
                                          "o", "outdir", "output directory.", seqan::ArgParseArgument::STRING, "STRING"));
  addDefaultValue(parser, "outdir", "./");
  
  addOption(parser, seqan::ArgParseOption(
                                          "t", "tsstes", "threshold of spliced alignments to exclude TSS/TES.", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "tsstes", 0);
            
  
  /*
  addOption(parser, seqan::ArgParseOption(
                                          "i", "min", "minimal size of loci considered.", seqan::ArgParseArgument::INTEGER, "INT")); */

  /*
  addOption(parser, seqan::ArgParseOption(
                                          "k", "maxSkip", "maximum number of exons skipped.", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "maxSkip", -1);*/
  
  /*
  addOption(parser, seqan::ArgParseOption(
                                          "nc", "normCut", "minimum value of getNormFactor not interpreted as zero.", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "normCut", 0);*/
  
  
  // ----------- Options for CG step ------------
  
  addSection(parser, "Column Generation Options");
  
  addOption(parser, seqan::ArgParseOption(
                                          "cg", "useColGen", "apply optional column generation step"));

  addOption(parser, seqan::ArgParseOption(
                                          "mr", "maxNofRelaxedEdges", "maximum number of edges relaxed in SG during column generation.", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "maxNofRelaxedEdges", 1);
  
  /*
   addOption(parser, seqan::ArgParseOption(
   "lb", "unexplCovBound", "lower bound on intensity of covers considered unexplained.", seqan::ArgParseArgument::INTEGER, "INT"));
   addDefaultValue(parser, "unexplCovBound", 0);*/
  
  addOption(parser, seqan::ArgParseOption(
                                          "u", "epsilon", "(upper) bound on error by piecewise linear approximation.", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "epsilon", 0.001);
  
  addOption(parser, seqan::ArgParseOption(
                                          "slb", "unexplCovSumBound", "lower bound on sum of intensity of covers considered unexplained.", seqan::ArgParseArgument::INTEGER, "INT"));
  addDefaultValue(parser, "unexplCovSumBound", 1);
  
  addOption(parser, seqan::ArgParseOption(
                                          "rw", "cgRegWeight", "scale lambda for transcripts built by CG in second glmnet run.", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "cgRegWeight", 1.3);
  
  addOption(parser, seqan::ArgParseOption(
                                          "rl", "cgRedLambda", "reduce lambda in CG.", seqan::ArgParseArgument::DOUBLE, "FLOAT"));
  addDefaultValue(parser, "cgRedLambda", 0.9);
  
/*
  addOption(parser, seqan::ArgParseOption(
                                          "ec", "explByCand", "covers explained with respect to candidate set instead of optimal glmnet solution."));
*/
  addOption(parser, seqan::ArgParseOption(
                                          "g", "greedyExplUncov", "update set of unexplained covers after each iteration of column generation."));
/*
  addOption(parser, seqan::ArgParseOption(
                                          "ac", "explainAnyCover", "column generation does not have to explain new cover in each iteration."));
*/
//  addOption(parser, seqan::ArgParseOption(
//                                          "y", "maxUncov", "maximum number of uncovered junctions in predicted transcripts (-1 for unlimited)", ArgParseArgument::INTEGER, "INT"));
//  addDefaultValue(parser, "maxUncov", 0);
//  setMinValue(parser, "maxUncov", "-1");

  
/*
  addOption(parser, seqan::ArgParseOption(
                                          "qa", "quadApprox", "File containing coefficients of quadratic approximation.", ArgParseArgument::STRING, "String"));
*/
/*
  addOption(parser, seqan::ArgParseOption(
                                          "fkt", "forceKnownTerminals", "Prevents all known TSS/TES in the candidate set during filtering."));
*/
  
  addUsageLine(parser,
               "[\\fIOPTIONS\\fP] \"\\fI.gtf\\fP\" \"\\fI.cnt\\fP\"");
  
  
  // Parse command line.
  seqan::ArgumentParser::ParseResult res = seqan::parse(parser, argc, argv);
  
  // If parsing was not successful then exit with code 1 if there were errors.
  // Otherwise, exit with code 0 (e.g. help was printed).
  if (res != seqan::ArgumentParser::PARSE_OK)
    return res;
  
  // Extract option values and print them.
  getArgumentValue(option.cntFile, parser, 1);
  getArgumentValue(option.gtfFile, parser, 0);
  
  getOptionValue(option.outDir, parser, "outdir");
  int tmp;
  //option.tss_tes = getOptionValue(tmp, parser, "tsstes");
  option.tss_tes = true;
  getOptionValue(option.tsstes_thr, parser, "tsstes");
  
  option.maxLoc = getOptionValue(tmp, parser, "max");
  if(option.maxLoc)
	  option.max_locsize = static_cast<size_t>(tmp);
  
  option.minLoc = false;
  
  //getOptionValue(tmp, parser, "min");
  //if(option.minLoc)
  //  option.min_locsize = static_cast<size_t>(tmp);
  
  getOptionValue(Parameters::insert_size_global, parser, "fraglength");
  
  getOptionValue(Parameters::read_length_global, parser, "readlength");
  
  getOptionValue(Parameters::s_dev, parser, "sdev");
  
  //Candidate Generation Options
  option.supTrans = true;
  option.strongSupTrans = false;
  
  option.eaw = !isSet(parser, "notExonAware");

  option.max_nof_exskip = -1u;
  
  option.regroup = isSet(parser, "groupLoci");
  if (option.regroup)
    option.useKnownTrans = false; //@Todo allow for regrouping AND known transcripts in the future
  else
    option.useKnownTrans = isSet(parser, "useTranscriptInformation");
  
  option.forceKnownTerminals = false;
  
  //Column Generation
  option.doColGen = isSet(parser, "useColGen");
  Parameters::greedy_expl_uncov = isSet(parser, "greedyExplUncov");
  //Parameters::cov_expl_by_opt = !isSet(parser, "explByCand");
  Parameters::cov_expl_by_opt = false;
  //Parameters::explainNewCover = !isSet(parser, "explainAnyCover");
  Parameters::explainNewCover = true;
  getOptionValue(Parameters::max_nof_relaxed_edges, parser, "maxNofRelaxedEdges");
  //getOptionValue(Parameters::unexpl_cov_bound, parser, "unexplCovBound");
  Parameters::unexpl_cov_bound = 0;
  
  
  getOptionValue(Parameters::unexpl_cov_sum_bound, parser, "unexplCovSumBound");
  getOptionValue(Parameters::cg_regWeight, parser, "cgRegWeight");
  getOptionValue(Parameters::scale_lambda_cg, parser, "cgRedLambda");
  
 
  option.maxUncovered = 0;
  option.singleLocus = false;
  
  //model Coefficients
  getOptionValue(option.epsilon, parser, "epsilon");
  
  option.theta = -1u;
  Parameters::normCutoff = 0;
  
  //Threshold filtering
  getOptionValue(option.expThreshold, parser, "expThreshold");
  option.absThreshold = isSet(parser, "absThreshold");
  
  //Algorithm selection
  //option.quadError = isSet(parser, "quadError");
  //option.pwLin = isSet(parser, "pwLinApprox");
  option.quadError = true;
  
  //Faux Cover Stuff
  getOptionValue(option.pseudoCount, parser, "pseudo");
  //getOptionValue(option.endspace, parser, "endspace");
  option.endspace = 1;
  
  if (isSet(parser, "noFaux"))
    option.fake = Parameters::FAKE_NONE;
  else
    option.fake = Parameters::FAKE_PAIRED_END;

  /*
  std::cerr << "cnt\t" << option.cntFile << '\n'
  << "fake\t" << fake << '\n'
//  << "weights\t" << option.weights << '\n'
  << "theta\t" << option.theta << '\n'
  << "normCut\t" << Parameters::normCutoff << '\n';
  if(option.singleLocus)
    std::cerr << "locus\t" << option.locusId << '\n';
  else
    std::cerr << "all loci" << '\n';
  
  if(option.quadError)
    std::cerr << "minimizing quadratic error using glmnet." << std::endl;
  
  std::cerr << "maximal uncovered junctions: " << option.maxUncovered << std::endl;
   */
  
  
  return seqan::ArgumentParser::PARSE_OK;
  
}

int main (int argc, const char** argv)
{
  
  //--------------------------------------
  //-------------Parsing Parameters-------
  //--------------------------------------
  Options opt;
  seqan::ArgumentParser::ParseResult res = runParser(argc, argv, opt);
  
  if (res != seqan::ArgumentParser::PARSE_OK)
    return res == seqan::ArgumentParser::PARSE_ERROR;
  
  Parameters::fake_mode = opt.fake;
//  Parameters::weights = opt.weights;
  Parameters::theta = opt.theta;
  Parameters::quad_error = opt.quadError;
//  Parameters::pw_lin = opt.pwLin;
  Parameters::epsilon = opt.epsilon;
  Parameters::max_nof_exskip = opt.max_nof_exskip;
  Parameters::pseudoCount = opt.pseudoCount;
  Parameters::endSpace = opt.endspace;
  Parameters::useKnownTrans = opt.useKnownTrans;
  
  bool cutterhead = opt.doColGen;
  
  //--------------------------------------
  //-------------Read data from files-----
  //--------------------------------------
  
  //all known transcripts mapped to their locus (key = locus_id)
  TKnownTranscriptsMap known_transcripts;
  
  // all loci (key = locus_id, value = segments (id) in locus)
  TAllLociMap all_loci;
  
  // the segment information. Key = segment_id
  TAllSegmentsMap all_segments;
  
  // all segment covers (counts)
  std::vector<SegmentCover> all_segment_covers;
  
  // all segment supercovers
  std::vector<SuperSegmentCover> all_segment_super_covers;
  
  //what do we use this for again?
  std::map<UInt, std::set<UInt> > covers_to_segments;
  
  // map to eachcover all supercovers containing it
  std::map<UInt, std::set<UInt> > covers_to_supercovers;
  
  //if only a single locus is processed the covers for this locus are distilled (deprecated?)
  std::vector<SegmentCover> selected_segment_covers;
  
  //the ground truth intensities for each transcript
  std::map<UInt, double> transcript_trueintens;
  
  //the intensities predicted
  std::map<UInt, double> transcript_intens;
  
  //the intensities predicted by the ILP
  std::map<UInt, std::string> trans_id_to_name;
  
  //contains for all real subexons the possible exit points (only matching donor - acceptor sites for exons)
  std::map<UInt, std::set<UInt> > validSubExits;
  
  //read the annotation File
  readAnnotationFile(opt.gtfFile, "", all_loci, all_segments, known_transcripts, trans_id_to_name, validSubExits);
  
  //read the covers (counts) from .cnt file
  readSegmentCovers(opt.cntFile, all_segment_covers, covers_to_segments);
  
  //inferred transcripts if denovo
  TKnownTranscriptsMap inferred_transcripts(all_loci.size());
  
  //vector<double> quad_apx_bp;
  //vector<pair<double,double> > left_quad_apx_coeff;
  //vector<pair<double,double> > right_quad_apx_coeff;
  vector<pwQuadCoeff> coeff_bp;
  
//  if(opt.quadApproxCoeffFile != "")
//	  readQuadApproxCoeff(opt.quadApproxCoeffFile, quad_apx_bp, left_quad_apx_coeff, right_quad_apx_coeff, coeff_bp);
  
  //cluster loci that share segments to reduce side effects
  std::map<UInt, std::vector<UInt> > loci_clusters;

  if (opt.regroup)
  {
    regroupLoci(all_segment_covers, all_loci, all_segments);
    known_transcripts.clear();
    known_transcripts.resize(all_loci.size());
    inferred_transcripts.resize(all_loci.size());
  }
  
  for(size_t i = 0; i < all_loci.size(); ++i)
  {
    loci_clusters[i] = std::vector<UInt>(1,i);
  }
  
  cerr<<"Input files read."<<endl;
  
  if(Parameters::de_novo_inference)
  {
    int count = 0;
    double sum_success_rate = 0.;
    double sum_specificity = 0.;
    int nof_contrib = 0;
    int tot_nof_inf = 0;
    int tot_nof_correct = 0;
    int tot_nof_exp_known = 0;
    std::map<UInt, std::vector<Transcript> > ground_truth;
    
    int nof_intronretention_loci = 0;
    
    if(opt.tss_tes)
    {
      identifyTssTes(all_segment_covers, all_segments, opt.tsstes_thr, true);
    }
    else if (opt.useKnownTrans)
    {
      getTssTesFromAnnotation(all_segments, known_transcripts);
    }
    
    //Loop over all Loci or work only selected gene
    //-------------------------------------------------------------------------------------------
    std::map<UInt, std::vector<UInt> >::const_iterator it = (opt.singleLocus) ? loci_clusters.find(opt.locusId) : loci_clusters.begin();
 
    unsigned loci_done = 0, prev_perc = 0;
    
    for(; it!=loci_clusters.end() && (!opt.singleLocus || it->first == opt.locusId); it++)
    {
      // ---- printing status bar
      unsigned perc = (100. * loci_done) / loci_clusters.size();
      ++loci_done;

      if (perc != prev_perc && 0 == perc % 10)
      {
        cout << perc << '%';
        cout.flush();
      }
      else if (perc != prev_perc && 0 == perc % 2)
      {
        cout << '.';
        cout.flush();
      }
      if (loci_done == loci_clusters.size())
      {
        cout << "100%" << endl;
      }

      prev_perc = perc;
      // ----- endof printing status bar
      
      unsigned numUncovered = 0;
      
      //std::cerr<<"cluster: " << it->first <<" of total: " << loci_clusters.size() << std::endl;
      
      std::set<UInt> cover_ids;
      int cluster_size = 0;
      std::set<vector<UInt> > unique_known;
      size_t nof_unsup = 0;
      for (size_t i = 0; i<loci_clusters[it->first].size(); i++)
      {
        int loc_id = loci_clusters[it->first][i];
        cluster_size += all_loci[loci_clusters[it->first][i]].size();
        
        //which covers are relevant for this locus? covers are added to current set in each iteration
        findAssociatedCoversIds(cover_ids, all_segment_covers, covers_to_segments, all_loci[loc_id]);
      }
      
//      cout << "Locus: " << it->first << endl;
//      for(size_t jj = 0; jj<known_transcripts[loci_clusters[it->first][0]].size(); ++jj)
//      {
//    	  cout << known_transcripts[loci_clusters[it->first][0]][jj].name << endl;
//    	  for(size_t ii = 0; ii < known_transcripts[loci_clusters[it->first][0]][jj].nodes.size(); ii++)
//    		  cout << known_transcripts[loci_clusters[it->first][0]][jj].nodes[ii] << " ";
//    	  cout << endl;
//      }

      if (Parameters::verbose)
        std::cout << "cluster has " << cluster_size << " segments" << std::endl;

      if(opt.maxLoc && cluster_size > opt.max_locsize)
      {
        std::cout << "too large cluster. skipping." << std::endl;
        continue;
      }
      if(opt.minLoc && cluster_size < opt.min_locsize)
      {
    	  std::cout << "too small cluster. skipping." << std::endl;
    	  continue;
      }
      
      
      std::vector<Transcript> cand_trans;
      SpliceGraph sg(all_loci[loci_clusters[it->first][0]], all_segments);
      
      if(!opt.supTrans && !cutterhead)
      {
        //enumerate all possible (2^n) transcripts
        for(size_t i=0; i<loci_clusters[it->first].size(); i++) {
          //generate_all_transcripts(all_loci[loci_clusters[it->first][i]], cand_trans);
          sg = SpliceGraph(all_loci[loci_clusters[it->first][i]], all_segments);
          sg.appendValidTranscripts(cand_trans);
        }
        if(cand_trans.empty())
        {
        	continue;
        }
      }
      else //if(((opt.lp || opt.quadError) && opt.supTrans) || cutterhead)
      {
        //generate all candidate transcripts
        std::vector<Transcript>tmp_enum_filtered_trans;
        for (size_t i = 0; i < loci_clusters[it->first].size(); i++)
        {
          sg = SpliceGraph(all_loci[loci_clusters[it->first][i]], all_segment_covers, std::vector<UInt>(cover_ids.begin(), cover_ids.end()), all_segments, true, false, validSubExits);
          sg.appendValidTranscripts(tmp_enum_filtered_trans, opt.maxUncovered, -1, false, opt.eaw);
          //cout << "number of candidate transcripts: " << tmp_enum_filtered_trans.size() << endl;
        }
        
        //filter multiple copies of same transcript
        std::set<vector<UInt> > unique_enum;
        for (size_t i = 0; i < tmp_enum_filtered_trans.size(); i++)
        {
          pair<std::set<vector<UInt> >::iterator, bool> ret = unique_enum.insert(tmp_enum_filtered_trans[i].nodes);
          if(ret.second==true)
            cand_trans.push_back(tmp_enum_filtered_trans[i]);
        }
        
        tmp_enum_filtered_trans.clear();
        unique_enum.clear();
        
        if(cand_trans.empty())
        	continue;

        if (cand_trans.size() > 100000)
          continue;
      }
      
      if(Parameters::useKnownTrans)
      {
        for (size_t i = 0; i < cand_trans.size(); ++i)
        {
          //check whether transcript is known
          bool t_known = false;
          const vector<Transcript>& known_trans = known_transcripts[loci_clusters[it->first][0]];
          vector<Transcript>::const_iterator t_it;
          for(t_it = known_trans.begin(); t_it != known_trans.end(); ++t_it)
          {
            if(t_it->nodes != cand_trans[i].nodes)
            {
              continue;
            }
            else
            {
              t_known = true;
              break;
            }
          }
          cand_trans[i].isKnown = t_known;
        }
      }
      
      
#ifdef REMOVE_NON_MAXIMAL
      //remove non maximal candidates
      std::vector<Transcript> tmp_cand;
      for (size_t i = 0; i < cand_trans.size(); ++i)
      {
        bool remove = false;
        if (!cand_trans[i].isKnown)
        {
          for (size_t j = 0; j < cand_trans.size(); ++j)
          {
            if (i == j)
              continue;
            
            if (cand_trans[i].nodes.size() < cand_trans[j].nodes.size() && 
                std::search(cand_trans[j].nodes.begin(), cand_trans[j].nodes.end(), cand_trans[i].nodes.begin(), cand_trans[i].nodes.end())!= cand_trans[j].nodes.end())
            {
              remove = true;
              
              if (Parameters::verbose)
              {
                std::cout << "Remove: ";
                for (size_t kk = 0; kk < cand_trans[i].nodes.size(); ++kk)
                  std::cout << cand_trans[i].nodes[kk] << '\t';
                std::cout << endl;
              }
              
                
              //cand_trans[i].regWeight = 100;
              break;
            }
          }
        }
        if (!remove)
          tmp_cand.push_back(cand_trans[i]);
      }
      tmp_cand.swap(cand_trans);
#endif
      
//      cout << "number of candidate transcripts: " << cand_trans.size() << endl;
//      for (size_t j = 0; j < cand_trans.size(); ++j)
//      {
//      	for (size_t z = 0; z < cand_trans[j].nodes.size(); z++)
//       		cout << cand_trans[j].nodes[z] << " ";
//       	cout << endl;
//      }
      
      //prepare covers here
      //-------------------
      vector<UInt> selected_cover_ids;
      static UInt count = 0;
      for (std::set<UInt>::iterator  id_it = cover_ids.begin(); id_it != cover_ids.end(); id_it++)
      {
    	  if(all_segment_covers[*id_it].exons_2.front() < all_segment_covers[*id_it].exons_1.front())
        {
          continue;
        }
    	  if(all_segment_covers[*id_it].exons_1.back() > all_segment_covers[*id_it].exons_2.back())
        {
          continue;
        }
        
    	  if(all_segment_covers[*id_it].exons_2.front() <= all_segment_covers[*id_it].exons_1.back())
        {
    		  vector<UInt>::const_iterator intersect_it = find(all_segment_covers[*id_it].exons_1.begin(), all_segment_covers[*id_it].exons_1.end(), all_segment_covers[*id_it].exons_2.front());
    		  if(intersect_it == all_segment_covers[*id_it].exons_1.end() )
    			  continue;
          
    		  bool invalid = false;
    		  for(int j = 0; intersect_it != all_segment_covers[*id_it].exons_1.end(); ++intersect_it, ++j)
    		  {
            if(j > all_segment_covers[*id_it].exons_2.size()-1 || all_segment_covers[*id_it].exons_2[j] != *intersect_it)
            {
    				  invalid = true;
    				  break;
    			  }
    		  }
    		  if(invalid)
    			  continue;
    	  }
    	  selected_cover_ids.push_back(*id_it);
      }
      
      if (Parameters::verbose)
        cout << "selected " << selected_cover_ids.size() << " covers." << endl;
      
      std::vector<SegmentCover> sup_fake_covers;
      std::vector<SegmentCover> all_fake_covers;
      std::pair<unsigned, unsigned> bounds;
      if(Parameters::fake_mode)
      {
        bounds.first = max((2 * Parameters::read_length_global), (int)floor(gsl_cdf_gaussian_Pinv(Parameters::lower_percentile, Parameters::s_dev)) + Parameters::insert_size_global);
        bounds.second = ceil(gsl_cdf_gaussian_Pinv(Parameters::upper_percentile, Parameters::s_dev)) + Parameters::insert_size_global;
        if(!opt.supTrans)
          generateFakeCovers(all_loci[loci_clusters[it->first][0]], all_segment_covers, selected_cover_ids, all_segments, all_fake_covers, sg, Parameters::fake_mode, bounds, Parameters::endSpace, opt.maxUncovered);
        if(cutterhead || opt.supTrans)
        {
          generateFakeCovers(all_loci[loci_clusters[it->first][0]], all_segment_covers, selected_cover_ids, all_segments, sup_fake_covers, sg, Parameters::fake_mode, bounds, Parameters::endSpace, opt.maxUncovered);
        }
        
        if (Parameters::verbose)
          cerr << "number of fakes: " << sup_fake_covers.size() << " / " << all_fake_covers.size() << endl;
      }
      
      clock_t _start = clock();
      
      if (opt.strongSupTrans)
      {
        //filter unsupported transcripts as in IsoLasso paper
        for (size_t i = 0; i < selected_cover_ids.size(); ++i)
          std::cout << all_segment_covers[selected_cover_ids[i]] << std::endl;
        
        std::vector<Transcript> cand_trans_tmp;
        
        for (size_t i = 0; i < cand_trans.size(); ++i)
        {
          if (isTranscriptSupported(all_segment_covers, selected_cover_ids, cand_trans[i], all_segments, all_loci[loci_clusters[it->first][0]]))
            cand_trans_tmp.push_back(cand_trans[i]);
        }
        cand_trans.swap(cand_trans_tmp);
        
        std::cout << "Transcripts before / after filtering: " << cand_trans_tmp.size() << " / " << cand_trans.size() << std::endl;
        
        if (cand_trans.empty())
          continue;
      }
      
      if (Parameters::penaltyUncov)
      {
        for (size_t i = 0; i < cand_trans.size(); ++i)
        {
          if (cand_trans[i].uncovered != 0)
            cand_trans[i].regWeight = 100000000.;
        }
      }
      
      if(Parameters::useKnownTrans)
      {
        for (size_t i = 0; i < cand_trans.size(); ++i)
        {
          if(!cand_trans[i].isKnown)
            cand_trans[i].regWeight *= 2;
        }
      }
      
      if (Parameters::lengthNormRegularized)
      {
        //as proposed in SLIDE paper
        for (size_t i = 0; i < cand_trans.size(); ++i)
        {
          cand_trans[i].regWeight *= 1./cand_trans[i].nodes.size();
        }
      }
      
      if (Parameters::verbose)
        std::cout << "Number candidates: " << cand_trans.size() << std::endl;
      
      
      std::vector<Transcript> inf_transcripts; //positively expressed inferred transcripts
      transcript_intens.clear();
      float lambda;
      float lp_error = -1;
      int success = 0;

      std::vector<bool> cov_expl (selected_cover_ids.size(), false);
//      if(opt.quadError || cutterhead)
//      {
        //if(false) {
      //*********************************
      //denovo LP, fit Lasso, no column generation
      //*********************************
      std::vector<SegmentCover>& fake_covers = (opt.supTrans || cutterhead) ? sup_fake_covers : all_fake_covers;
      success = doGlmnet(cand_trans,
                         all_loci[loci_clusters[it->first][0]],
                         transcript_intens,
                         all_segment_covers,
                         //std::vector<UInt>(cover_ids.begin(), cover_ids.end()),
                         selected_cover_ids,
                         cov_expl,
                         fake_covers,
                         all_segments,
                         lambda,
                         lp_error, //ATTENTION not defined if ran for all transcripts (= no option -s)
                         known_transcripts[loci_clusters[it->first][0]]);
      if(success == 7777)
      {
        if (Parameters::verbose)
          cout << "glmnet reports zero variance on predictor variables. resorting to cplex." << endl;
        
        float theta = Parameters::theta;
        Parameters::theta = 0.;
        //Parameters::theta = lambda;
        bool quad_error = Parameters::quad_error;
        Parameters::quad_error = true;
        float score = buildILP(cand_trans,
                               transcript_intens,
                               all_segment_covers,
                               //std::vector<UInt>(cover_ids.begin(), cover_ids.end()),
                               selected_cover_ids,
                               fake_covers,
                               all_segments,
                               true);
        lp_error = score; //since theta=0, score = error
        lambda = 0; //lambda not set by doglmnet, but used by column generation
        Parameters::theta = theta;
        Parameters::quad_error = quad_error;
        success = 0;
      }
//      }
//      else if(opt.lp)
//      {
//    	  //success = 0;
//    	  std::vector<SegmentCover>& fake_covers = (opt.supTrans) ? sup_fake_covers : all_fake_covers;
//    	  buildILP(cand_trans,
//                 transcript_intens,
//                 all_segment_covers,
//                 //std::vector<UInt>(cover_ids.begin(), cover_ids.end()),
//                 selected_cover_ids,
//                 fake_covers,
//                 all_segments,
//                 true);
//      }
//      if(opt.lp || opt.quadError || cutterhead)
//      {
      if (Parameters::verbose)
        cout << "initial prediction: " << endl;
      
//      if(cand_trans.empty())
//      {
//        const vector<UInt>& locus = all_loci[loci_clusters[it->first][0]];
//        for(map<UInt, double>::iterator trans_it = transcript_intens.begin(); trans_it != transcript_intens.end(); trans_it++)
//          if(trans_it->second > 0) {
//            Transcript tmp;
//            tmp.nodes.reserve(locus.size());
//            for(UInt pos=0; pos <locus.size(); ++pos)
//            {
//              if((trans_it->first + 1) & (1<<pos))
//              {
//                tmp.nodes.push_back(locus[pos]);
//                if (Parameters::verbose)
//                  cout << locus[pos] << " ";
//              }
//            }
//            
//            if (Parameters::verbose)
//              cout << endl;
//            
//            tmp.id = trans_it->first;
//            inf_transcripts.push_back(tmp);
//          }
//      }

      //check tss/tes coverage
      //int a, *b;
      
      const vector<UInt>& locus = all_loci[loci_clusters[it->first][0]];
      vector<bool> tss_cov(locus.size(), false);
      vector<bool> tes_cov(locus.size(), false);
      
      for(UInt pos=0; pos<locus.size(); ++pos)
      {
        if(all_segments[locus[pos]].tss)
        {
          tss_cov[pos]=true;
          //cout << "found tss: " << locus[pos] << endl;
        }
        if(all_segments[locus[pos]].tes)
        {
          tes_cov[pos]=true;
          //cout << "found tes: " << locus[pos] << endl;
        }
      }
      
      for (size_t i = 0; success<=0 && i < cand_trans.size(); i++)
      {
        std::map<UInt, double>::const_iterator intens_it = transcript_intens.find(cand_trans[i].id);
        
        if(intens_it == transcript_intens.end())
          continue;
        
        //if(transcript_intens[intens_it->first]>0.0000001) //had trouble with qp (lots of e-16 values)
        if(transcript_intens[intens_it->first] > 0)
        {
          inf_transcripts.push_back(cand_trans[i]);
          inf_transcripts.back().predictedExpression = intens_it->second;
          
          if (Parameters::verbose)
          {
            for(int z = 0; z < cand_trans[i].nodes.size(); z++)
              cout << cand_trans[i].nodes[z] << " ";
            cout << endl;
            cout << "intens:" << intens_it->second << endl;
          }
        }
      }
      
      if(cluster_size <= 50 && cutterhead && fabs(lp_error-0)>0.001) //column generation
      {
        //prepare arguments for column generation method
        std::vector<Transcript> initial_trans;
        
        //paired-end column generation
        float theta;
        if(cutterhead)
        {
        	initial_trans = inf_transcripts;
        	inf_transcripts.clear();
        	theta = Parameters::theta;
        	Parameters::theta = Parameters::scale_lambda_cg*lambda;
        	//Parameters::theta = lambda;
        }
        
        //cout << "theta for cutterhead: " << lambda << endl;
        std::vector<Transcript> new_transcripts;
        int nof_iter = colgen_pe_pwlin(inf_transcripts,
                                       new_transcripts,
                  	  	  	  	  	   all_loci[loci_clusters[it->first][0]],
                  	  	  	  	  	   transcript_intens,
                  	  	  	  	  	   all_segment_covers,
                  	  	  	  	  	   //std::vector<UInt>(cover_ids.begin(), cover_ids.end()),
                  	  	  	  	  	   selected_cover_ids,
                  	  	  	  	  	   cov_expl,
                  	  	  	  	  	   all_fake_covers,
                  	  	  	  	  	   all_segments,
                  	  	  	  	  	   initial_trans,
                  	  	  	  	  	   opt.tss_tes,
                  	  	  	  	  	   sg,
                  	  	  	  	  	   coeff_bp,
                  	  	  	  	  	   lp_error,
                  	  	  	  	  	   validSubExits);
        if(cutterhead)
        {
        	Parameters::theta = theta;
        	if(nof_iter<2) //if no improving transcript was found (nof_iter==1) keep initial solution (transcript_intens not overwritten)
        	{
        		inf_transcripts = initial_trans;
        	}
          else //do glmnet again with extended candidate set
        	{
        		assert(new_transcripts.size()>0);
            if (Parameters::verbose)
              cout << "adding following transcripts to candidate set: " << endl;
            
            std::vector<UInt> has_junc_new(new_transcripts.size());
            hasUnsupportedJunction(all_loci[loci_clusters[it->first][0]], all_segment_covers, std::vector<UInt>(selected_cover_ids.begin(), selected_cover_ids.end()), all_segments, new_transcripts, has_junc_new, 1, true);
            
        		for(size_t z=0; z<new_transcripts.size(); z++)
            {
              if(!has_junc_new[z])
                continue;
              
              if (Parameters::verbose)
              {
                for(int t = 0; t < new_transcripts[z].nodes.size(); t++)
                  cout << new_transcripts[z].nodes[t] << " ";
                cout << endl;
              }
        			cand_trans.push_back(new_transcripts[z]);
        		}
        		transcript_intens.clear();
        		cov_expl.assign(selected_cover_ids.size(), false);
        		//generate more fakes to account for additional transcripts
        		//sup_fake_covers.clear();
        		//generateFakeCovers(all_loci[loci_clusters[it->first][0]], all_segment_covers, selected_cover_ids, all_segments, sup_fake_covers, sg, Parameters::fake_mode, bounds, Parameters::endSpace, Parameters::max_nof_relaxed_edges);
        		std::vector<SegmentCover>& fake_covers = (opt.supTrans || cutterhead) ? sup_fake_covers : all_fake_covers;
        		success = doGlmnet(cand_trans,
                               all_loci[loci_clusters[it->first][0]],
                               transcript_intens,
                               all_segment_covers,
                               //std::vector<UInt>(cover_ids.begin(), cover_ids.end()),
                               selected_cover_ids,
                               cov_expl,
                               fake_covers,
                               all_segments,
                               lambda,
                               lp_error, //ATTENTION not defined if ran for all transcripts (= no option -s)
                               known_transcripts[loci_clusters[it->first][0]]);
        		assert(success != 7777); //right?
        		//assuming cand_trans != empty here... sloppy
            
            if (Parameters::verbose)
              cout << "final prediction after glmnet + cg + glmnet: " << endl;
        		
            inf_transcripts.clear();
        		for (size_t i = 0; success<=0 && i < cand_trans.size(); i++)
        		{
        			std::map<UInt, double>::const_iterator intens_it = transcript_intens.find(cand_trans[i].id);
              
        			if(intens_it == transcript_intens.end())
        				continue;
              
        			//if(transcript_intens[intens_it->first]>0.0000001) //had trouble with qp (lots of e-16 values)
        			if(transcript_intens[intens_it->first] > 0)
        			{
        				inf_transcripts.push_back(cand_trans[i]);
        				inf_transcripts.back().predictedExpression = intens_it->second;
                
                if (Parameters::verbose)
                {
                  for(int z = 0; z < cand_trans[i].nodes.size(); z++)
                    cout << cand_trans[i].nodes[z] << " ";
                  cout << endl;
                  cout << "intens:" << intens_it->second << endl;
                }
        			}
        		}
        	}
        }
      }
      clock_t _end = clock();
      
      if (Parameters::verbose)
        cout << "solving locus " << it->first << " took: " << (double)(_end-_start) / CLOCKS_PER_SEC << endl;
      
      
      if(success>0)
      {
        cout << "not successful." << endl;
        continue;
      }
      
#ifdef DO_REOPTIMIZATION
      Parameters::quad_error = true;
      std::vector<SegmentCover>& fake_covers_tmp = (opt.supTrans) ? sup_fake_covers : all_fake_covers;
      transcript_intens.clear();
      
      buildILP(inf_transcripts,
               transcript_intens,
               all_segment_covers,
               selected_cover_ids,
               fake_covers_tmp,
               all_segments,
               false);
      
      for (size_t i = 0; i < inf_transcripts.size(); ++i)
      {
        std::map<UInt, double>::const_iterator intens_it = transcript_intens.find(inf_transcripts[i].id);
        
        if(intens_it == transcript_intens.end())
          continue;
        
        inf_transcripts[i].predictedExpression = intens_it->second;
      }
#endif
      
      //mark the higest expressed transcript for each previously unexplained known TSS/TES as mandatory
      std::vector<bool> is_mandatory;
      
      
#ifdef DO_REDISTRIBUTE_READS
      
#ifdef FORCE_MIN_SUPPORT
      
      is_mandatory.assign(inf_transcripts.size(), false);
      if (opt.forceKnownTerminals)
      {
        markMandatoryCandidates(inf_transcripts, all_segments, is_mandatory);
      }
      
      std::map<UInt, double> trans_support;
      for(auto cov_id_iter = selected_cover_ids.begin(); cov_id_iter != selected_cover_ids.end(); ++cov_id_iter)
      {
        for(int j = 0; j < inf_transcripts.size(); ++j)
        {
          if(isCoverSupported(all_segment_covers[*cov_id_iter], vector<Transcript>(1,inf_transcripts[j])))
          {
            trans_support[j] += all_segment_covers[*cov_id_iter].intensity;
          }
        }
      }
      std::vector<Transcript> tmp;
      for (size_t i = 0; i < inf_transcripts.size(); ++i)
      {
        if (trans_support[i]< 10 && !is_mandatory[i])
          continue;
        
        tmp.push_back(inf_transcripts[i]);
      }
      inf_transcripts.swap(tmp);
#endif
      
      std::vector<double> reassigned_expression(inf_transcripts.size(),0);
      for(auto cov_id_iter = selected_cover_ids.begin(); cov_id_iter != selected_cover_ids.end(); ++cov_id_iter)
      {
        std::vector<double> trans_pred_local(inf_transcripts.size(),0);
        double local_pred = 0.;
        for(int j = 0; j < inf_transcripts.size(); ++j)
        {
          if(isCoverSupported(all_segment_covers[*cov_id_iter], vector<Transcript>(1,inf_transcripts[j])))
          {
            //float normf = getNormFactor(all_covers[*cov_id_iter], transcripts[j], segments, read_length, insert_size);
            double normf = getNormFactor(all_segment_covers[*cov_id_iter], inf_transcripts[j], all_segments);
            trans_pred_local[j] = normf * inf_transcripts[j].predictedExpression;
            local_pred += trans_pred_local[j];
          }
        }
        
        if(local_pred > 0)
        {
          //std::cout << "pred: " << local_pred << " " << all_segment_covers[*cov_id_iter].intensity << " " << all_segment_covers[*cov_id_iter] << std::endl;
          double factor = all_segment_covers[*cov_id_iter].intensity / local_pred;
          for(int j = 0; j < inf_transcripts.size(); ++j)
          {
            if(isCoverSupported(all_segment_covers[*cov_id_iter], vector<Transcript>(1,inf_transcripts[j])))
            {
              reassigned_expression[j] += trans_pred_local[j] * factor;
            }
          }
        }
      }
      
      for (size_t i = 0; i < inf_transcripts.size(); ++i)
      {
        //get size of transcript
        unsigned len = 0;
        for (size_t j = 0; j < inf_transcripts[i].nodes.size(); ++j)
        {
          len += all_segments[inf_transcripts[i].nodes[j]].length;
        }
        //make FPKM value
        inf_transcripts[i].predictedExpression = reassigned_expression[i] / len;
      }
#endif
      
      //transform into FPKM
      double km_factor = 1000. / Parameters::M_global;
      for (size_t i = 0; i < inf_transcripts.size(); ++i)
      {
        inf_transcripts[i].predictedExpression *= km_factor;
      }
      
      //mark the higest expressed transcript for each previously unexplained known TSS/TES as mandatory
      is_mandatory.assign(inf_transcripts.size(), false);
      if (opt.forceKnownTerminals)
      {
        markMandatoryCandidates(inf_transcripts, all_segments, is_mandatory);
      }
      
      //filter low expressed
      if (opt.absThreshold)
      {
        for (size_t i = 0; i < inf_transcripts.size(); i++)
        {
          if (inf_transcripts[i].predictedExpression >= opt.expThreshold || is_mandatory[i])
            inferred_transcripts[it->first].push_back(inf_transcripts[i]);
        }
      }
      else
      {
        postproc_inf_trans(inf_transcripts, transcript_intens, inferred_transcripts[it->first], is_mandatory, opt.expThreshold);
      }
    }

    writeAnnotationFile(inferred_transcripts, string(opt.gtfFile), string(opt.outDir + "/transcripts.gff"));
  }
  
  return 0;
}
