/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Parameters.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <algorithm>
#include <Functions.h>
#include <FunctionsGraph.h>
#include <cmath>
#include <gsl/gsl_cdf.h>
#include <assert.h>
//#include <unordered_map>
#include <cstdio>
#include <string.h>

//#include <seqan/PathEnumeration.h>
#include <seqan/graph_algorithms.h>
#include <seqan/graph_types.h>
//#include <TR1/unordered_map>
//#include<hash_map.h>
using namespace std;

#define USEREACH //this is for test reasons in generateFakeCovers.


/*
Compute the probability, that for the given lengths w1, w2 and w2 of three segments s1 s2 s3, there are no paired reads that such that the left starts in s1 and the right in segment 3 (From Feng, Li, Jiang RECOMB-Paper)
*/
float getNoHitProb(int M, float alpha, Size w1, Size w2, Size w3, float mean, float sd)
{
  //compute term P0 as defined in paper assuming a normal distribution of insert sizes with mean and standard deviation sd
  const static float scale = pow(10.0,-9.0);
  float Po=0;
  Size l_i=w1+w2;
  Size u_i=w1+w2+w3;

  for(Size i=0; i<w1; ++i)
  {
    Po+= (gsl_cdf_gaussian_P(u_i-mean , sd) - gsl_cdf_gaussian_P(l_i-mean, sd));
    --l_i;
    --u_i;
  }
  return exp(-1 * M * Po * alpha*scale);

}



const float billion = 1000000000;
float getNoHitProb(int M, float alpha, Size pos)
{    
  return exp(-1 * alpha * pos * M * (1/billion));
}

float getNoHitProbBeta(int M, float rpkm, UInt valid_positions, UInt transcript_length)
{
  float expected_num_reads=M * rpkm * (float)transcript_length/1000;
  return pow((1-valid_positions/transcript_length), expected_num_reads);
}



void stringToCover(SegmentCover& cover, const string& cover_string, double count)
{
  char buffer [500];
  strcpy (buffer, cover_string.c_str());

  cover.exons_1.clear();
  cover.exons_2.clear();
  cover.mapped_reads.clear();

  cover.intensity = count;

  char delim [] = "^";
  char * left = strtok(buffer, delim);
  char * right = strtok(NULL, delim);

  delim[0] = '-';

  if(left != NULL)
  {
    char * node_id = strtok(left, delim);

    while(node_id != NULL)
    {
      cover.exons_1.push_back(atoi(node_id));
      node_id = strtok(NULL, delim);
    }
  }

  if(right != NULL)
  {
    char * node_id = strtok(right, delim);

    while(node_id != NULL)
    {
      cover.exons_2.push_back(atoi(node_id));
      node_id = strtok(NULL, delim);
    }
  }
}



void getSegmentSuperCovers(const std::vector<SegmentCover>& segment_covers,
                           std::vector<SuperSegmentCover>& segment_super_covers)
{
  std::map<Size, std::vector<std::pair<unsigned, double> > >reads;
  for(Size i = 0; i < segment_covers.size(); ++i)
  {
    double max_count = 0;
    for(Size read = 0; read < segment_covers[i].mapped_reads.size(); ++read)
    {
      max_count = std::max(max_count, segment_covers[i].mapped_reads[read].count);
    }
    for(Size read = 0; read < segment_covers[i].mapped_reads.size(); ++read)
    {
      if(segment_covers[i].mapped_reads[read].count > 0.05 && segment_covers[i].mapped_reads[read].count / max_count > 0.1)
      {
        reads[segment_covers[i].mapped_reads[read].id].push_back(std::make_pair(i, segment_covers[i].mapped_reads[read].count));
      }
    }
  }

  std::map<std::set<unsigned>, Size> combs;
  std::map<std::set<unsigned>, Size>::const_iterator combs_it;
  for(Size i= 0; i < reads.size(); ++i)
  {
    std::set<UInt>check;
    for(Size j = 0; j < reads[i].size(); ++j)
    {
      if(!check.insert(reads[i][j].first).second)
      {
        //invalid read
        reads[i].clear();
        break;
      }
    }
    if(reads[i].size() > 1)
    {
      combs[check] += 1;
    }
  }

  std::cerr<<"Number of combs: " << combs.size() << std::endl;
  for(combs_it = combs.begin(); combs_it != combs.end(); ++ combs_it)
  {
    SuperSegmentCover cov;
    cov.intensity = combs_it->second;

    std::set<unsigned>::const_iterator seg_it;
    for(seg_it = combs_it->first.begin(); seg_it != combs_it->first.end(); ++seg_it)
    {
      cov.covers.push_back(*seg_it);
    }
    segment_super_covers.push_back(cov);
  }
}



//check whether a NodeCover is supported by some transcript
bool isCoverSupported(const SegmentCover& cand_cover, const vector<Transcript>& transcripts)
{
  vector<Transcript>::const_iterator it;
  bool supported=false;

  for(it=transcripts.begin(); it!=transcripts.end() && !supported; ++it)
  {
    vector<UInt>::const_iterator it_l, it_r;
    it_l = search(it->nodes.begin(),it->nodes.end(), cand_cover.exons_1.begin(), cand_cover.exons_1.end());

  if(!cand_cover.exons_2.empty())
  {
      it_r = search(it->nodes.begin(),it->nodes.end(), cand_cover.exons_2.begin(), cand_cover.exons_2.end());
      if (it_l != it->nodes.end() && it_r != it->nodes.end()&& it_l<=it_r)
    {
        supported=true;
    }
  }
    else if(it_l != it->nodes.end())
  {
      supported=true;
  }
}
  return supported;
}



//check whether a NodeCover is supported by some Locus
//it is supported if all nodes are contained in the locus and ordering is preserved
bool isCoverSupported(const SegmentCover& cand_cover, const vector<UInt>& nodes)
{
  vector<UInt>::const_iterator nodes_start = nodes.begin();

  for(UInt i=0; i< cand_cover.exons_1.size(); ++i)
  {
    nodes_start = find(nodes_start, nodes.end(), cand_cover.exons_1[i]);
    if(nodes_start == nodes.end())
    {
      return false;
    }
  }
  nodes_start = nodes.begin();
  for(UInt i=0; i< cand_cover.exons_2.size(); ++i)
  {
    nodes_start = find(nodes_start, nodes.end(), cand_cover.exons_2[i]);
    if(nodes_start == nodes.end())
    {
      return false;
    }
  }
  return true;
}



bool isTranscriptSupported(const std::vector<SegmentCover>& all_covers,
                           const std::vector<UInt>& selected_cover_ids,
                           const Transcript& transcript)
{
  std::vector<UInt>::const_iterator sel_covers_it;
  std::vector<UInt>::const_iterator it_l, it_r;

  //check for single segments
  for(size_t j = 0; j < transcript.nodes.size(); ++j)
  {
    bool is_supported = false;


    for(sel_covers_it = selected_cover_ids.begin(); sel_covers_it != selected_cover_ids.end(); ++sel_covers_it)
    {
      it_l = find(all_covers[*sel_covers_it].exons_1.begin(), all_covers[*sel_covers_it].exons_1.end(), transcript.nodes[j]);
      it_r = find(all_covers[*sel_covers_it].exons_2.begin(),all_covers[*sel_covers_it].exons_2.end(), transcript.nodes[j]);
      if (it_l != all_covers[*sel_covers_it].exons_1.end() || it_r != all_covers[*sel_covers_it].exons_2.end())
      {
        is_supported = true;
        break;
      }
    }
    if(!is_supported)
    {
      return false;
    }

    //check for junction
    if(j+1 < transcript.nodes.size())
    {
      is_supported = false;
      for(sel_covers_it = selected_cover_ids.begin(); sel_covers_it != selected_cover_ids.end(); ++sel_covers_it)
      {
        it_l = search(all_covers[*sel_covers_it].exons_1.begin(), all_covers[*sel_covers_it].exons_1.end(), &transcript.nodes[j], &transcript.nodes[j+2]);
        it_r = search(all_covers[*sel_covers_it].exons_2.begin(),all_covers[*sel_covers_it].exons_2.end(), &transcript.nodes[j], &transcript.nodes[j+2]);
        if (it_l != all_covers[*sel_covers_it].exons_1.end() || it_r != all_covers[*sel_covers_it].exons_2.end())
        {
          is_supported = true;
          break;
        }
      }
    }
    if(!is_supported)
    {
      return false;
    }
  }
  return true;
}


bool isTranscriptSupported(const std::vector<SegmentCover>& all_covers,
                            const std::vector<UInt>& selected_cover_ids,
                            const Transcript& transcript,
                            const TAllSegmentsMap& allSegments,
                            const std::vector<UInt>& locus)
{
  std::vector<UInt>::const_iterator sel_covers_it;
  std::map<size_t, bool> covered_single, covered_junc;

  std::map<size_t, size_t> locus_invert;
  for (size_t i = 0; i < locus.size(); ++i)
    locus_invert[locus[i]] = i;

  for(sel_covers_it = selected_cover_ids.begin(); sel_covers_it != selected_cover_ids.end(); ++sel_covers_it)
  {
    double normf = getNormFactor(all_covers[*sel_covers_it], transcript, allSegments);
    std::cout << transcript.nodes.size() << " " << all_covers[*sel_covers_it] << "  ::  " << normf << std::endl;

    if (normf > 0.0001)
    {
      for (size_t i = 0; i < all_covers[*sel_covers_it].exons_1.size(); ++i)
      {
        covered_single[all_covers[*sel_covers_it].exons_1[i]] = true;

        if (i + 1 < all_covers[*sel_covers_it].exons_1.size())
        {
          covered_junc[all_covers[*sel_covers_it].exons_1[i]] = true;
        }
      }

      for (size_t i = 0; i < all_covers[*sel_covers_it].exons_2.size(); ++i)
      {
        covered_single[all_covers[*sel_covers_it].exons_2[i]] = true;

        if (i + 1 < all_covers[*sel_covers_it].exons_2.size())
        {
          covered_junc[all_covers[*sel_covers_it].exons_2[i]] = true;
        }
      }

      //save neighboring exons covered by pair
      if (locus_invert[all_covers[*sel_covers_it].exons_2.front()] - locus_invert[all_covers[*sel_covers_it].exons_1.back()] == 1)
        covered_junc[all_covers[*sel_covers_it].exons_1.back()] = true;
    }
  }

  bool is_supported = true;
  std::cout << "cov map size: " << covered_single.size() << "  " << covered_single.count(transcript.nodes[0]) << std::endl;
  for(size_t j = 0; j < transcript.nodes.size(); ++j)
  {
    if (!covered_single.count(transcript.nodes[j]))
    {
      is_supported = false;
      break;
    }

    if (j+1 < transcript.nodes.size())
    {
      if (!covered_junc.count(transcript.nodes[j]) &&
          allSegments[transcript.nodes[j]].endPos != allSegments[transcript.nodes[j+1]].beginPos)
      {
        is_supported = false;
        break;
      }
    }
  }

  std::cout << "returns: " << is_supported << std::endl;
  return is_supported;
}




void addAssociatedCoverIds(std::set<UInt>& cover_ids,
                           const std::vector<SegmentCover>& all_covers,
                           const std::map<UInt, set<UInt> > &index,
                           const std::vector<Transcript>& transcripts,
                           const std::vector<UInt>& node_ids)
{
  for(UInt i=0; i<node_ids.size(); ++i)
  {
    if(index.find(node_ids[i]) == index.end())
    {
      continue;
      //if no cover exists for given node we continue!
      //cerr<<"calling for node id: "<<node_ids[i]<<" which is not available --- terminating"<<endl;
      //throw -1;
    }
    for(set<UInt>::const_iterator it = index.find(node_ids[i])->second.begin(); it!=index.find(node_ids[i])->second.end(); ++it)
    {
      if(isCoverSupported(all_covers[*it], transcripts))
      {
        cover_ids.insert(*it);
      }
    }
  }
}

/*
For a given set of transcripts find all NodeCovers that support at least one transcript
*/
void findAssociatedCovers(vector<SegmentCover>& associated_covers,
                          const vector<SegmentCover>& all_covers,
                          const map<UInt, set<UInt> >& index,
                          const vector<Transcript>& transcripts,
                          const vector<UInt>& node_ids)
{
  set<UInt>associated_covers_cand;
  for(UInt i=0; i<node_ids.size(); ++i)
  {
    if(index.find(node_ids[i])==index.end())
    {
      //cerr<<"calling for node id: "<<node_ids[i]<<" which is not available --- but ok, not terminating"<<endl;
      //throw -1;
    }
    for(set<UInt>::const_iterator it = index.find(node_ids[i])->second.begin(); it!=index.find(node_ids[i])->second.end(); ++it)
    {
      if(isCoverSupported(all_covers[*it], transcripts))
      {
        associated_covers_cand.insert(*it);
      }
    }
  }

  associated_covers.clear();
  associated_covers.reserve(associated_covers_cand.size());
  for(set<UInt>::iterator it = associated_covers_cand.begin(); it!=associated_covers_cand.end(); ++it)
  {
    associated_covers.push_back(all_covers[*it]);
  }
}

/*
For a given set of nodes (they actually define one locus) find all NodeCovers that belong to that locus
*/
void findAssociatedCovers(vector<SegmentCover>& associated_covers,
                          const vector<SegmentCover>& all_covers,
                          const map<UInt, set<UInt> >& index,
                          const vector<UInt>& node_ids)
{
  set<UInt>associated_covers_cand;
  for(UInt i=0; i<node_ids.size(); ++i)
  {
    if(index.find(node_ids[i])==index.end())
    {
      //cerr<<"calling for node id: "<<node_ids[i]<<" which is not available --- but ok, not terminating"<<endl;
      //throw -1;
    }
    for(set<UInt>::const_iterator it = index.find(node_ids[i])->second.begin(); it!=index.find(node_ids[i])->second.end(); ++it)
    {
      if(isCoverSupported(all_covers[*it], node_ids))
      {
        associated_covers_cand.insert(*it);
      }
    }
  }
  associated_covers.clear();
  associated_covers.reserve(associated_covers_cand.size());
  for(set<UInt>::iterator it = associated_covers_cand.begin(); it!=associated_covers_cand.end(); ++it)
  {
    associated_covers.push_back(all_covers[*it]);
  }
}

void findAssociatedCoversIds(std::set<UInt>& cover_ids, const vector<SegmentCover> &all_covers, const map<UInt, set<UInt> > &index, const vector<UInt> &node_ids)
{
  for(UInt i=0; i<node_ids.size(); ++i)
  {
    if(index.find(node_ids[i])==index.end())
    {
      //cerr<<"calling for node id: "<<node_ids[i]<<" which is not available --- but ok, not terminating"<<endl;
      continue;
      //throw -1;
    }
    for(set<UInt>::const_iterator it = index.find(node_ids[i])->second.begin(); it!=index.find(node_ids[i])->second.end(); ++it)
    {
      //std::cerr << all_covers.size() << " !! " << *it << " !! " << node_ids.size() << std::endl;
      if(isCoverSupported(all_covers[*it], node_ids))
      {
        cover_ids.insert(*it);
      }
    }
  }
}



float getCoverLength(const SegmentCover& node_cover, const TAllSegmentsMap& segments, int read_length)
{  
  int left_max=0;

  //for the left mate coverage
  int total_l = 0;
  SegmentCover::TContainerType::const_iterator start_it = node_cover.exons_1.begin();
  SegmentCover::TContainerType::const_iterator end_it = node_cover.exons_1.end();
  while(start_it!=end_it)
  {
    total_l+=segments[*start_it].length;
    ++start_it;
  }

  //total_l = segments.find(node_cover.exons_1[0])->second.length + segments.find(node_cover.exons_1[1])->second.length;
  //the number of reads that can be placed is limited by the complete length of the covered nodes
  left_max=max(1+total_l-read_length, 0);

  //left_max = total_l;
  if(node_cover.exons_1.size()>1)
  {
    int front_l=segments[node_cover.exons_1.front()].length;
    int back_l = segments[node_cover.exons_1.back()].length;
    //the number of reads that can be placed is limited by lengths of the first and last node
    left_max=min(min(front_l, back_l), left_max);

    //the number of reads that can be placed is limited the number of reads that can span the interior nodes
    //left_max=min(max(read_length-(total_l-front_l-back_l)-1,0), left_max);

    int second_l=segments[node_cover.exons_1[1]].length;
    left_max=min(max(read_length-(total_l-front_l-second_l)-1,0), left_max);

    //left_max = min(2*read_length - 2, left_max);
  }
  return left_max;
}



void computeSingleRange(const vector<UInt>& nodes_cover,
                        const Transcript& transcript,
                        const TAllSegmentsMap& segments,
                        int read_length,
                        pair<int, int>& range)
{
  range.first = 0;
  range.second = 0;

  //case single exon:
  if(nodes_cover.size() == 1)
  {    
    for (size_t i = 0; transcript.nodes[i] != nodes_cover.front(); ++i)
    {
      range.first += segments[transcript.nodes[i]].length;
    }
    range.second = range.first + (int)segments[nodes_cover.front()].length - read_length;
  }

  //case exon junction
  else if(nodes_cover.size() == 2)
  {
    range.first = segments[nodes_cover.front()].length;
    for (size_t i = 0; transcript.nodes[i] != nodes_cover.front(); ++i)
    {
      range.first += segments[transcript.nodes[i]].length;
      //std::cerr << transcript.nodes[i] << " : : "<< segments.find(transcript.nodes[i])->second.beginPos << " " << segments.find(transcript.nodes[i])->second.endPos << std::endl;
    }    
    //std::cerr << range.first << " " << segments.find(nodes_cover.back())->second.length << " " << read_length << " = " << range.first + (int) segments.find(nodes_cover.back())->second.length - read_length << std::endl;
    range.second = min(range.first - 1, range.first + (int) segments[nodes_cover.back()].length - read_length);
    range.first = max(range.first - read_length + 1, range.first - (int) segments[nodes_cover.front()].length);
  }

  //case spanned exons
  else if(nodes_cover.size() > 2)
  {    
    for (size_t i = 0; transcript.nodes[i] != nodes_cover.front(); ++i)
    {
      range.first += segments[transcript.nodes[i]].length;
    }
    int total_last = range.first;
    for (size_t i = 0; i < nodes_cover.size(); ++i)
    {
      total_last += segments[nodes_cover[i]].length;
    }
    range.second = min(range.first + (int) segments[nodes_cover.front()].length - 1, total_last - read_length);
    range.first = max(range.first, total_last - (int) segments[nodes_cover.back()].length - read_length + 1);
  }

  //return make_pair(range.first, range.second);
}



/*
Compute for a given NodeCover and readlength the number of possible mapping positions
-we look first at the NodeCover for "left mates" and then at the NodeCover for "right" mates
-case destinction between the case of a single node, or spanning of several nodes
*/
float getNormFactor(const SegmentCover& node_cover,
                    const Transcript& transcript,
                    const TAllSegmentsMap& segments,
                    int read_length,
                    int fragment_length,
                    double s_dev,
                    bool single_paired_reads,
                    double cutoff
                    )
{  

  //take the 0.05 and 0.95 quantile as boundaries for considered fragment lengths
  int lower = max((2 * read_length), (int)floor(gsl_cdf_gaussian_Pinv(Parameters::lower_percentile, s_dev)) + fragment_length);
  int upper = ceil(gsl_cdf_gaussian_Pinv(Parameters::upper_percentile, s_dev)) + fragment_length;

  static vector<double> probs(upper + 1, 0);
  static int last_frag_len = -1;

  if(fragment_length != last_frag_len && s_dev > 0)
  {
    for(int i = lower; i <= upper; ++i)
    {
      probs[i] = (1./(Parameters::upper_percentile - Parameters::lower_percentile)) * (gsl_cdf_gaussian_P(i - fragment_length, s_dev) - gsl_cdf_gaussian_P(i - fragment_length - 1, s_dev));
    }
    last_frag_len = fragment_length;
  }

  if(!isCoverSupported(node_cover, vector<Transcript>(1,transcript)))
  {
    return 0;
  }

  //case single end reads
  if(node_cover.exons_2.empty() && !single_paired_reads)
  {
    pair<int, int> pos;
    computeSingleRange(node_cover.exons_1, transcript, segments, read_length, pos);
    double normf = pos.second - pos.first + 1;
    return normf > cutoff ? normf : 0;
  }

  //case single end of a paired end read (could be left or right end)
  if(node_cover.exons_2.empty() && single_paired_reads)
  {
    //compute transcript length
    int trans_len = 0;
    for(Size i = 0; i < transcript.nodes.size(); ++i)
    {
      trans_len += segments[transcript.nodes[i]].length;
    }

    pair<int, int> pos;
    computeSingleRange(node_cover.exons_1, transcript, segments, read_length, pos);


    int space_left = 0;
    int space_right = 0;
    if(s_dev > 0)
    {
      for(int frag_len_s = lower; frag_len_s <= upper; ++frag_len_s)
      {
        space_left += probs[frag_len_s] * max(0, 1 + pos.second - max(frag_len_s - read_length, pos.first));
        space_right += probs[frag_len_s] * max(0, max(trans_len - frag_len_s, pos.second) - pos.first);
      }
    }
    else
    {
      space_left += max(0, 1 + pos.second - max(fragment_length - read_length, pos.first));
      space_right += max(0, 1 + min(trans_len - fragment_length, pos.second) - pos.first);
    }
    double normf = space_left + space_right;
    return normf > cutoff ? normf : 0;
  }

  //case paired end reads
  else
  {
    double norm = 0;
    pair<int, int> pos_l, pos_r;
    computeSingleRange(node_cover.exons_1, transcript, segments, read_length, pos_l);
    computeSingleRange(node_cover.exons_2, transcript, segments, read_length, pos_r);
    int space_left = pos_l.second - pos_l.first + 1;
    int space_right = pos_r.second - pos_r.first + 1;
    int delta_inner = pos_r.first - pos_l.second - read_length;
    int delta_outer = pos_r.second - pos_l.first + read_length;
    int insert_size = fragment_length - 2 * read_length;
    //std::cerr << node_cover << " " << pos_l.first << " " << pos_l.second << std::endl;

    if(s_dev > 0)
    {
      for(int frag_len_s = lower; frag_len_s <= upper; ++frag_len_s)
      {
        insert_size = frag_len_s - 2 * read_length;
        int norm_local = min(space_left, space_right);
        norm_local = min(norm_local, delta_outer - frag_len_s + 1);
        norm_local = min(norm_local, insert_size - delta_inner + 1);

        norm += float(max(0, norm_local)) * probs[frag_len_s];
      }
    }
    else
    {
      norm = min(min(space_left, space_right), min(delta_outer - fragment_length + 1, insert_size - delta_inner + 1));
      norm = max (norm, 0.);
    }    
    return norm > cutoff ? norm : 0;
  }
}


void computeRefinedCounts (const SegmentCover& node_cover,
                       const std::vector<Transcript>& transcripts,
                       const TAllSegmentsMap& segments,
                       const std::pair<double,double>& fragLenBounds,
                       std::vector<std::pair<std::vector<UInt>, double> >& clusters,
                       int read_length)
{
  //makes sense for paired ends only
  if (node_cover.exons_2.empty())
  {
    return;
  }
  
  int gap = 0;
  Size idx = 0;
  
  typedef SegmentCover::TDistMap::const_iterator TGapIt;
  typedef std::map<std::pair<UInt,UInt>, std::pair<double,std::list<UInt> > > TTmpClusterMap;

  TTmpClusterMap tmpClusters;
  
  for (std::vector<Transcript>::const_iterator tIt = transcripts.begin(); tIt != transcripts.end(); ++tIt)
  {
    if(!isCoverSupported(node_cover, vector<Transcript>(1,*tIt)))
      continue;


    //get index of last subexon covered by left mate
  //  if (node_cover.exons_1.back() != node_cover.exons_2.front())
    if (std::find(node_cover.exons_2.begin(), node_cover.exons_2.end(), node_cover.exons_1.back()) == node_cover.exons_2.end())
    {
      //if pairs do not share a subexon, i.e. have a gap
      for (; idx < tIt->nodes.size() && tIt->nodes[idx] != node_cover.exons_1.back(); ++idx);
      
      for (++idx; idx < tIt->nodes.size() && tIt->nodes[idx] != node_cover.exons_2.front(); ++idx)
      {
        gap += segments[tIt->nodes[idx]].length;
      }
    }
    
    if (idx == tIt->nodes.size())
    {
      //last left or first right subexon not contained in transcript
      std::cerr << "last left or first right subexon not contained in transcript -- exit" << std::endl;
      std::cout << node_cover << std::endl;
      for (size_t kk = 0; kk < tIt->nodes.size(); ++kk)
        std::cout << tIt->nodes[kk] << '\t';
      std::cout<< std::endl;
      exit(1);
    }
    
    TGapIt it1, it2;
    

    int lb = static_cast<int>(fragLenBounds.first) - (2*read_length + gap);
    int ub = static_cast<int>(fragLenBounds.second) - (2*read_length + gap);
   
    //obtain iterators for the range of allowed dists
    it1 = node_cover.distCounts.lower_bound(lb);
    it2 = node_cover.distCounts.upper_bound(ub);
    
    std::pair<UInt,UInt>  key(std::distance(node_cover.distCounts.begin(), it1),
                              std::distance(node_cover.distCounts.begin(), it2));

    if (tmpClusters.find(key) == tmpClusters.end())
    {
      //add up associated counts
      double cnt = 0;
      while (it1 != it2)
      {
        cnt += it1->second;
        ++it1;
      }
      tmpClusters[key].first = cnt;
    }
      
    tmpClusters[key].second.push_back(std::distance(transcripts.begin(),tIt));
  }
  
  clusters.resize(tmpClusters.size());
  size_t pos = 0;
  for (TTmpClusterMap::const_iterator cIT = tmpClusters.begin(); cIT != tmpClusters.end(); ++cIT, ++pos)
  {
    clusters[pos].second = cIT->second.first;
    clusters[pos].first.assign(cIT->second.second.begin(),cIT->second.second.end());
  }
  
  return;
}


float getEmpiricalNormFactor(const SegmentCover& node_cover,
                             const Transcript& transcript,
                             const TAllSegmentsMap& segments,
                             int read_length,
                             int insert_size)
{
  static vector<double> probs(500, 0);
  static int last_insert_size = -1;

  if(insert_size != last_insert_size)
  {
    for(int i = 2*read_length; i < 500; ++i)
    {
      probs[i] = gsl_cdf_gaussian_P(i - insert_size, Parameters::s_dev) - gsl_cdf_gaussian_P(i - insert_size - 1, Parameters::s_dev);
    }
    last_insert_size = insert_size;
  }

  if(!isCoverSupported(node_cover, vector<Transcript>(1,transcript)))
  {
    return 0;
  }

  //case single end reads
  if(node_cover.exons_2.empty())
  {
    pair<int, int> pos;
    computeSingleRange(node_cover.exons_1, transcript, segments, read_length, pos);
    return(max(0, pos.second - pos.first + 1));
  }
  //case paired end reads
  else
  {
    double norm = 0.;
    double total_weight = 0.;
    int offset_pos_l = 0, offset_pos_r = 0;

    std::vector<UInt>::const_iterator left_exon = transcript.nodes.begin();
    UInt right_exon = node_cover.exons_1.front();
    while(*left_exon != right_exon)
    {
      offset_pos_l += segments[*left_exon].length;
      ++left_exon;
    }
    offset_pos_r += offset_pos_l;
    right_exon = node_cover.exons_2.front();
    while(*left_exon != right_exon)
    {
      offset_pos_r += segments[*left_exon].length;
      ++left_exon;
    }

    //std::cerr<<node_cover<<"  #reads: " << node_cover.mapped_reads.size() << std::endl;
    for (size_t read = 0; read < node_cover.mapped_reads.size(); ++read)
    {
      int read_pos_l = offset_pos_l + node_cover.mapped_reads[read].start_l;
      int read_pos_r = offset_pos_r + node_cover.mapped_reads[read].start_r;

      int emp_ins_size = read_length + read_pos_r - read_pos_l + 1;

      pair<int, int> pos_l, pos_r;
      computeSingleRange(node_cover.exons_1, transcript, segments, read_length, pos_l);
      computeSingleRange(node_cover.exons_2, transcript, segments, read_length, pos_r);
      int space_left = pos_l.second - pos_l.first + 1;
      int space_right = pos_r.second - pos_r.first + 1;
      int delta_s = pos_r.first - pos_l.first;

      int spacer = emp_ins_size - read_length;
      int norm_local = min(space_left, space_right);
      norm_local = min(norm_local, space_right - spacer + delta_s);
      norm_local = min(norm_local, space_left + spacer - delta_s);
      //cout<<"insert size: "<<insert_s<<" pos: " << norm_local << " prob:  " << probs[insert_s] <<std::endl;

      total_weight += (node_cover.mapped_reads[read].count * probs[emp_ins_size]);

      norm += float(max(0, norm_local)) * probs[emp_ins_size] * node_cover.mapped_reads[read].count;
    }
    if(norm != 0) norm /= total_weight;
    return(norm);
  }
}



void generate_all_transcripts(const vector<UInt>& locus, vector<Transcript> &transcripts)
{ 
  if(pow(2.,(int)locus.size())>ULONG_MAX || locus.empty())
  {
    cerr<<"too many possible transcripts -- Aborting"<<endl;
    return;
  }
  const UInt size=locus.size();
  UInt num = pow(2.,(int)size);

  Size start_id = transcripts.size();

  //transcripts.clear();
  //transcripts.reserve(num);
  Transcript tmp;
  tmp.nodes.reserve(locus.size());
  for(UInt i = 1; i < num; ++i)
  {

    for(UInt pos=0; pos < size; ++pos)
    {
      if(i& (1<<pos))
      {
        tmp.nodes.push_back(locus[pos]);
      }
    }
    tmp.id = start_id + i - 1;
    transcripts.push_back(tmp);
    tmp.nodes.clear();

  }

  //debug
  /* tmp.nodes.push_back(locus[0]);
  tmp.nodes.push_back(locus[1]);
  tmp.nodes.push_back(locus[2]);
  tmp.nodes.push_back(locus[3]);
  tmp.nodes.push_back(locus[4]);
  tmp.nodes.push_back(locus[6]);
  tmp.nodes.push_back(locus[7]);
  tmp.nodes.push_back(locus[8]);
  tmp.nodes.push_back(locus[9]);
  tmp.id = start_id;
  transcripts.push_back(tmp);
  tmp.nodes.clear();
 */

}


/* DEPRECATED
void appendValidTranscripts(const vector<UInt> &locus,
                            const std::vector<SegmentCover> &all_covers,
                            const std::vector<UInt> &selected_cover_ids,
                            vector<Transcript> &transcripts)
{
  //generate seqan graph with a node for each segment and edges are drawn according to cover overlaps
  typedef seqan::Graph<seqan::Directed<> > TGraph;
  typedef seqan::VertexDescriptor<TGraph>::Type VertexDescriptor;
  typedef seqan::StringSet<seqan::String<VertexDescriptor>, seqan::Owner<> > TStringSet;

  Size start_id = transcripts.size();

  std::map<UInt, VertexDescriptor> node_ids;
  std::map<UInt, bool> observed_node_ids;
  std::map<VertexDescriptor, UInt> node_ids_inv;

  TGraph graph;

  VertexDescriptor source = seqan::addVertex(graph);
  VertexDescriptor target = seqan::addVertex(graph);

  for(Size i = 0; i < locus.size(); ++i)
  {
    VertexDescriptor v_id = seqan::addVertex(graph);
    node_ids[locus[i]] = v_id;
    node_ids_inv[v_id] = locus[i];
  }

  //create edge set according to segment covers
  std::vector<UInt>::const_iterator cov_it;
  std::map<UInt, VertexDescriptor>::const_iterator left, right;

  for(cov_it = selected_cover_ids.begin(); cov_it != selected_cover_ids.end(); ++cov_it)
  {
    observed_node_ids[all_covers[*cov_it].exons_1[0]] = true;
    left = node_ids.find(all_covers[*cov_it].exons_1[0]);
    for(Size i = 0; i + 1 < all_covers[*cov_it].exons_1.size(); ++i)
    {
      observed_node_ids[all_covers[*cov_it].exons_1[i+1]] = true;
      right = node_ids.find(all_covers[*cov_it].exons_1[i+1]);
      if(left == node_ids.end() || right == node_ids.end())
      {
        left = right;
        continue;
      }
      if(!seqan::findEdge(graph, left->second, right->second))
      {
        seqan::addEdge(graph, left->second, right->second);
      }
      left = right;
    }

    if(!all_covers[*cov_it].exons_2.empty())
    {
      observed_node_ids[all_covers[*cov_it].exons_2[0]] = true;
      left = node_ids.find(all_covers[*cov_it].exons_2[0]);
      for(Size i = 0; i + 1 < all_covers[*cov_it].exons_2.size(); ++i)
      {
        observed_node_ids[all_covers[*cov_it].exons_2[i+1]] = true;
        right = node_ids.find(all_covers[*cov_it].exons_2[i+1]);
        if(left == node_ids.end() || right == node_ids.end())
        {
          left = right;
          continue;
        }
        if(!seqan::findEdge(graph, left->second, right->second))
        {
          seqan::addEdge(graph, left->second, right->second);
        }
        left = right;
      }
    }
  }


  //for all vertices add edge from the source and to the target vertex
  std::map<UInt, VertexDescriptor>::const_iterator vertices;

  for(vertices = node_ids.begin(); vertices != node_ids.end(); ++vertices)
  {
    if(observed_node_ids[vertices->first])
    {
      seqan::addEdge(graph, vertices->second, target);
      seqan::addEdge(graph, source, vertices->second);
    }
  }

  TStringSet all_paths;
  enumerateAcyclicPaths(graph, source, target, seqan::String<unsigned>(), -1u, all_paths); //seqan string is dummy. not used since next param is -1

  //paths to transcripts
  transcripts.resize(transcripts.size() + seqan::length(all_paths));

  typedef seqan::Iterator<TStringSet>::Type SetIterator;
  SetIterator set_it(all_paths);

  std::set<std::vector<UInt> >test_set;
  for(; !seqan::atEnd(set_it); seqan::goNext(set_it), ++start_id)
  {
    Transcript & t_ref = transcripts[start_id];
    t_ref.id = start_id;
    t_ref.nodes.reserve(seqan::length(*set_it));

    //do not copy the artificial source and target
    for(size_t i = 1; i + 1 < seqan::length(*set_it); ++ i)
    {
      t_ref.nodes.push_back(node_ids_inv[(*set_it)[i]]);
    }
    if(!test_set.insert(t_ref.nodes).second)
    {
      std::cerr<<"error repeated transcript!"<<endl;
      exit(-1);
    }

  }
}
*/


void generateFakeCovers(const vector<SegmentCover>& all_covers,
                        const vector<UInt>& cover_ids,
                        const vector<Transcript>& transcripts,
                        vector<SegmentCover>& fake_covers,
                        Parameters::FakeMode mode)
{
  if(mode == Parameters::FAKE_NONE)
  {
    return;
  }

  //as preprocessing we search for all junctions (edges in splice graph) that are not part of some cover. for all theses junctions we generate a node cover with
  //intensity 0, So selection of transcripts containing junctions not covered by any reads will be penalized
  std::set<pair<UInt, UInt> >abundant_junctions;
  std::set<UInt>abundant_singles;

  std::set<std::pair<std::vector<UInt>, std::vector<UInt> > >abundant_pairs;
  vector<UInt>::const_iterator cov_id_iter;

  for (cov_id_iter = cover_ids.begin(); cov_id_iter != cover_ids.end(); ++cov_id_iter)
  {
    if (mode != Parameters::FAKE_SINGLE_END)
    {
      abundant_pairs.insert(make_pair(all_covers[*cov_id_iter].exons_1, all_covers[*cov_id_iter].exons_2));
    }
    else
    {
      if (all_covers[*cov_id_iter].exons_1.size() == 1)
      {
        abundant_singles.insert(all_covers[*cov_id_iter].exons_1.front());
      }

      for (int j=0; j<(int)all_covers[*cov_id_iter].exons_1.size()-1; ++j)
      {
        abundant_junctions.insert(make_pair(all_covers[*cov_id_iter].exons_1[j], all_covers[*cov_id_iter].exons_1[j+1]));
      }

      if (all_covers[*cov_id_iter].exons_2.size() == 1)
      {
        abundant_singles.insert(all_covers[*cov_id_iter].exons_2.front());
      }

      for (int j=0; j<(int)all_covers[*cov_id_iter].exons_2.size()-1; ++j)
      {
        abundant_junctions.insert(make_pair(all_covers[*cov_id_iter].exons_2[j], all_covers[*cov_id_iter].exons_2[j+1]));
      }
    }
  }

  //find all transcripts that contain the junction
  //here we build the fake covers
  //if(Parameters::mode != Parameters::PAIRED_END)
  if(mode == Parameters::FAKE_SINGLE_END)
  {
    for(UInt i=0; i<transcripts.size(); ++i)
    {
      for(UInt j=0; j<transcripts[i].nodes.size(); ++j)
      {
        if(j<transcripts[i].nodes.size()-1 && abundant_junctions.find(make_pair(transcripts[i].nodes[j], transcripts[i].nodes[j+1])) == abundant_junctions.end())
        {
          fake_covers.push_back(SegmentCover());
          fake_covers.back().exons_1.push_back(transcripts[i].nodes[j]);
          fake_covers.back().exons_1.push_back(transcripts[i].nodes[j+1]);
          abundant_junctions.insert(make_pair(transcripts[i].nodes[j], transcripts[i].nodes[j+1]));
        }
        if(abundant_singles.find(transcripts[i].nodes[j]) == abundant_singles.end())
        {
          fake_covers.push_back(SegmentCover());
          fake_covers.back().exons_1.push_back(transcripts[i].nodes[j]);
          abundant_singles.insert(transcripts[i].nodes[j]);
        }
      }
    }
  }
  //if(Parameters::mode != Parameters::SINGLE_END)
  //TODO fake covers not based on transcripts but based on the whole locus
  if(mode != Parameters::FAKE_SINGLE_END)
  {
    int fake = 0;
    for(UInt i=0; i<transcripts.size(); ++i)
    {
      //std::cerr << i << " of trans " << transcripts.size() << std::endl;
      for(UInt j=0; j<transcripts[i].nodes.size(); ++j)
      {
        SegmentCover single_single, single_pair, pair_pair, pair_single;
        single_single.exons_1.push_back(transcripts[i].nodes[j]);
        single_pair.exons_1.push_back(transcripts[i].nodes[j]);

        if(j<transcripts[i].nodes.size()-1)
        {
          pair_pair.exons_1.push_back(transcripts[i].nodes[j]);
          pair_pair.exons_1.push_back(transcripts[i].nodes[j+1]);
          pair_single.exons_1.push_back(transcripts[i].nodes[j]);
          pair_single.exons_1.push_back(transcripts[i].nodes[j+1]);
        }

        for(UInt k = j; k<transcripts[i].nodes.size(); ++k)
        {
          pair_single.exons_2.clear();
          single_single.exons_2.clear();
          single_pair.exons_2.clear();
          pair_pair.exons_2.clear();

          single_single.exons_2.push_back(transcripts[i].nodes[k]);

          if(mode != Parameters::FAKE_PAIRED_END_TRUNC)
          {
            if(k<transcripts[i].nodes.size()-1)
            {
              pair_single.exons_2.push_back(transcripts[i].nodes[k + 1]);
              single_pair.exons_2.push_back(transcripts[i].nodes[k]);
              single_pair.exons_2.push_back(transcripts[i].nodes[k + 1]);
            }

            if(transcripts[i].nodes.size()>2 && k<transcripts[i].nodes.size()-2)
            {
              pair_pair.exons_2.push_back(transcripts[i].nodes[k+1]);
              pair_pair.exons_2.push_back(transcripts[i].nodes[k+2]);
              //if(j<transcripts[i].nodes.size()-1)
              //cout << pair_pair.exons_1[0] << " " << pair_pair.exons_1[1] << " " <<  pair_pair.exons_2[0] << " " << pair_pair.exons_2[1] << endl;
            }
          }

          if(!abundant_pairs.count(make_pair(single_single.exons_1, single_single.exons_2)))
          {
            fake_covers.push_back(single_single);
            abundant_pairs.insert(make_pair(single_single.exons_1, single_single.exons_2));
            ++fake;
          }

          if(mode != Parameters::FAKE_PAIRED_END_TRUNC)
          {
            if(!single_pair.exons_2.empty() && !abundant_pairs.count(make_pair(single_pair.exons_1, single_pair.exons_2)))
            {
              fake_covers.push_back(single_pair);
              abundant_pairs.insert(make_pair(single_pair.exons_1, single_pair.exons_2));
              ++fake;
            }

            if(!pair_single.exons_2.empty() && !abundant_pairs.count(make_pair(pair_single.exons_1, pair_single.exons_2)))
            {
              fake_covers.push_back(pair_single);
              abundant_pairs.insert(make_pair(pair_single.exons_1, pair_single.exons_2));
              ++fake;
            }

            if(!pair_pair.exons_1.empty() && !pair_pair.exons_2.empty() && !abundant_pairs.count(make_pair(pair_pair.exons_1, pair_pair.exons_2)))
            {
              fake_covers.push_back(pair_pair);
              abundant_pairs.insert(make_pair(pair_pair.exons_1, pair_pair.exons_2));
              ++fake;
            }
          }

        }
      }
    }
    std::cout<<"#Fakes: "<<fake <<std::endl;
  }
}



void generateFakeCovers(const std::vector<UInt>& locus,
                        const std::vector<SegmentCover>& all_covers,
                        const std::vector<UInt>& cover_ids,
                        const TAllSegmentsMap & all_segments,
                        std::vector<SegmentCover>& fake_covers,
                        const SpliceGraph& sGraph,
                        Parameters::FakeMode mode,
                        const std::pair<unsigned, unsigned>& fragLenBounds,
                        UInt end_space,                        
                        UInt maxUncovered,
                        UInt maxExonSkip,
                        const SpliceGraph::TLegalExonBoundMap& legalBlockBounds)
{
  const unsigned rl = Parameters::read_length_global;
  const unsigned isLb = fragLenBounds.first - 2 * rl;
  const unsigned isUb = fragLenBounds.second - 2 * rl;

  --end_space;

  std::set<SegmentCover> abundantCovers;
  const SpliceGraph *sg = &sGraph;

  bool exonAware = !legalBlockBounds.empty();

  bool countUncovered = false;

//  if (maxUncovered == -1u)
//    sg = new SpliceGraph(locus, all_segments, legalBlockBounds);
//  else
//  {
//    sg = new SpliceGraph(locus, all_covers, cover_ids, all_segments, false, true, legalBlockBounds);
//    countUncovered = true;
//  }
  
    if (maxUncovered != -1u)
      countUncovered = true;

  //std::map<UInt,Segment>::const_iterator iti, itj, itk, itl;


  //check which segment are on some source-sink path
  seqan::String<unsigned int> predMap;
  seqan::String<unsigned int> distMap;
  SpliceGraph::TGraph tmpG;// = sg->graph;
  sg->getCoveredGraph(tmpG); // if !only_covered this simply returns graph, otherwise graphCov

  //std::cerr << tmpG << std::endl;
  seqan::breadthFirstSearch(predMap, distMap, tmpG, sg->source);
  std::vector<bool> reachable(locus.size(), true);
  
#ifdef USEREACH
  for (size_t i = 0; i < locus.size(); ++i)
    reachable[i] = seqan::getProperty(distMap, sg->getNodeId(locus[i])) != seqan::_getInfinityDistance(distMap);
  
  seqan::transpose(tmpG);
  seqan::breadthFirstSearch(predMap, distMap, tmpG, sg->target);

  for (size_t i = 0; i < locus.size(); ++i)
    reachable[i] = reachable[i] && (seqan::getProperty(distMap, sg->getNodeId(locus[i])) != seqan::_getInfinityDistance(distMap));
#endif
  
  for (size_t i = 0; i < cover_ids.size(); ++i)
    abundantCovers.insert(all_covers[cover_ids[i]]);

  for (size_t i = 0; i < locus.size(); ++i)
  {     
    //cerr << i << ": " << fake_covers.size() << endl;
    //std::cerr << locus[i] << " " << sg->node_ids[locus[i]] << " " << reachable[i] << std::endl;   
    if (!reachable[i])
      continue;

    const Segment& segi = all_segments[locus[i]];

    for (size_t j = i; j < locus.size(); ++j)
    {
      if (!reachable[j])
        continue;
      if (i != j && !sg->findEdge(locus[i],locus[j], true))
        continue;

      UInt uncoveredL = 0;

      int minGap = isLb;
      int maxGap = isUb;

      const Segment& segj = all_segments[locus[j]];

      if (countUncovered && i != j && !sg->findEdge(locus[i],locus[j]))
      {
        if (++uncoveredL > maxUncovered)
          continue;
      }

      if (i == j && segi.length < rl + end_space)
        continue; //not enough space to place read on single exon

      if (i != j && (min(segi.length, rl - 1) +
                     min(segj.length, rl - 1) < rl + end_space))
        continue; //not enough space to place read on exon-exon junction

      if (mode == Parameters::FAKE_SINGLE_END)
      {
          SegmentCover tmp;
          tmp.exons_1.push_back(locus[i]);
          if (i != j)
              tmp.exons_1.push_back(locus[j]);

          if (!abundantCovers.count(tmp))
            fake_covers.push_back(tmp);
          continue;
      }

      minGap -= (static_cast<int>(segj.length) - static_cast<int>(rl));

      if (i != j)
      {
        minGap -= min(segi.length, rl - 1);
        //std::cerr << "B " << minGap << std::endl;
        maxGap -= max(0, (static_cast<int>(segj.length) - static_cast<int>(rl) + 1));
      }

      for (size_t k = j; k < locus.size(); ++k)
      {
        if (!reachable[k])
          continue;

        const Segment& segk = all_segments[locus[k]];

        for (size_t l = k; l < locus.size(); ++l)
        {
          if (!reachable[l])
            continue;

          if (k != l && !sg->findEdge(locus[k],locus[l], true))
            continue;

          UInt uncoveredB = uncoveredL;

          SegmentCover tmp;
          tmp.exons_1.push_back(locus[i]);
          tmp.exons_2.push_back(locus[k]);

          if (i != j)
            tmp.exons_1.push_back(locus[j]);
          if (k != l)
            tmp.exons_2.push_back(locus[l]);

          if (abundantCovers.count(tmp))
            continue;


          const Segment& segl = all_segments[locus[l]];

          if (countUncovered && k != l && !sg->findEdge(locus[k],locus[l]))
          {
            if (++uncoveredB > maxUncovered)
              continue;
          }

          if (k == l && segk.length < rl + end_space)
            continue; //not enough space to place read on single exon

          if (k != l && (min(segk.length, rl - 1) +
                         min(segl.length, rl - 1) < rl + end_space))
            continue; //not enough space to place read on exon-exon junction

          int minGapA = minGap - (static_cast<int>(segk.length) - static_cast<int>(rl));
          int maxGapA = maxGap;

          if (k != l)
          {
            minGapA -= min(segl.length, rl - 1);
            maxGapA -= max(0,(static_cast<int>(segk.length) - static_cast<int>(rl) + 1));
          }

          bool addNoGap = false;
          if (j == k) // no gap
          {
            if (i == j && k == l) //single exon
              addNoGap = (segj.length >= isLb + (2 * rl));

            else if (i != j && k == l) // i-j^j
              addNoGap = ((segj.length >= isLb + rl + 1) &&
                     (segi.length + segj.length >= isLb + (2 * rl)));

            else if (i == j && k != l) // j^j-l
              addNoGap = ((segj.length >= isLb + rl + 1) &&
                     (segl.length + segj.length >= isLb + (2 * rl)));

            else //i-j^j-l
            {
              addNoGap = ((segj.length <= isUb + (2* rl) - 2) &&
                          (segj.length >= isLb + 2) &&
                          (min(segi.length, rl - 1) + segj.length + min(segl.length, rl - 1) >= isLb + (2 * rl)));
            }
          }

          //std::cerr << i << " " << j << " " << k << " " << l << " add: " << addNoGap << " MinGap: " << minGapA << " MaxGap: " << maxGapA << std::endl;
          //std::cerr << locus[i] << " " << locus[j] << " " << locus[k] << " " << locus[l] << " add: " << addNoGap << " MinGap: " << minGapA << " MaxGap: " << maxGapA << std::endl;
          if ( (maxGapA < 0 || j==k ) && !addNoGap)
            continue;

          if (addNoGap || (minGapA + static_cast<int>(end_space) <= 0 && (sg->findEdge(locus[j],locus[k], maxUncovered - uncoveredB))))
          {            
            if (exonAware)
            {
              std::vector<UInt> tmpPath(&locus[i], &locus[l+1]);
              if (!sg->isExonAware(tmpPath))
                continue;
            }
            //can appear consecutively -- NO GAP
            //std::cerr << "consecutively" << std::endl;
            fake_covers.push_back(tmp);
            continue;
          }
          else
          {
            // otherwise we have to use the splice graph and check whether some valid combination
            // exists to fill up the gap
            if (k-j > 1)
            {
              bool isSup = sg->isCoverGapSupported(all_segments, tmp, std::make_pair(max(minGapA,0), maxGapA), maxUncovered - uncoveredB, maxExonSkip, exonAware);
              if (isSup)
                fake_covers.push_back(tmp);
            }
          }
        }
      }
    }
  }
  //delete sg;
}



ostream& operator<< (ostream& out, const SegmentCover& n)
{
  if(n.exons_1.empty())
  {
    return out;
  }
  SegmentCover::TContainerType::const_iterator it = n.exons_1.begin();

  out<<*(it++);
  for(; it!=n.exons_1.end(); ++it)
  {
    out <<"-"<<*it;
  }
  if(!n.exons_2.empty())
  {
    it=n.exons_2.begin();
    out<<"^"<<*(it++);

    for(; it!=n.exons_2.end(); ++it)
    {
      out <<"-"<<*it;
    }
  }
  return out;
}



void postproc_inf_trans(std::vector<Transcript>& input_trans,
                        map<UInt, double>& transcript_intens,
                        std::vector<Transcript>& output_trans,
                        std::vector<bool>& is_mandatory,
                        float thresh)
{
  //compute total intensity of all predicted transcripts
  float tot_intensity = 0.;
  double max_intensity = 0.;
  for (size_t i = 0; i < input_trans.size(); i++)
  {
    std::map<UInt, double>::const_iterator intens_it = transcript_intens.find(input_trans[i].id);
    if(intens_it == transcript_intens.end())
      continue;
    if(transcript_intens[intens_it->first] > 0)
    {
      max_intensity = std::max(transcript_intens[intens_it->first], max_intensity);
      tot_intensity += transcript_intens[intens_it->first];
    }
  }
  for (size_t i = 0; i < input_trans.size(); ++i)
  {
    std::map<UInt, double>::const_iterator intens_it = transcript_intens.find(input_trans[i].id);
    if(intens_it == transcript_intens.end())
      continue;
//    if (transcript_intens[intens_it->first]/tot_intensity > thresh)

    float thresh_ = thresh;
    if (Parameters::useKnownTrans)
        thresh_ = input_trans[i].isKnown ? thresh : 2*thresh;
    if (transcript_intens[intens_it->first]/max_intensity > thresh_ || is_mandatory[i])
    //if (transcript_intens[intens_it->first]/tot_intensity > 0)
    {
      output_trans.push_back(input_trans[i]);
    }
  }
}



bool enumGapExonSets(const SegmentCover& cover,
                     const TAllSegmentsMap& all_segments,
                     const std::vector<UInt>& locus,
                     vector<vector<bool> >& valid_candidates,
                     const std::map<UInt, std::set<UInt> >& legalBlockBounds,
                     int read_length,
                     int fragment_length,
                     double s_dev)
{  
  assert(!cover.exons_2.empty());
  if(cover.exons_1.back()>=cover.exons_2.front())
    return true;

  //store and sort segments in gap according to their length
  vector<UInt>::const_iterator low = lower_bound(locus.begin(), locus.end(), cover.exons_1.back());

  vector<pair<UInt,UInt> > segs;
  vector<SegmentType> seg_types;
  UInt i = 0;
  for(vector<UInt>::const_iterator it = low+1; *it<cover.exons_2.front(); ++it, ++i)
  {
    int length = all_segments[*it].length;
    SegmentType st = all_segments[*it].type;
    segs.push_back(make_pair(length,i));
    seg_types.push_back(st);
  }

  if(segs.size() == 0)
    return true;

//  sort(segs.begin(), segs.end());

  //mate lengths
  int l_mate_length = 0;
  vector<UInt>::const_iterator mate_it;
  for(mate_it = cover.exons_1.begin(); mate_it != cover.exons_1.end(); ++mate_it)
    l_mate_length += all_segments[*mate_it].length;
  int r_mate_length = 0;
  for(mate_it = cover.exons_2.begin(); mate_it != cover.exons_2.end(); ++mate_it)
    r_mate_length += all_segments[*mate_it].length;

  //prepare call to addRecursiveGapExons
  vector<bool> current_sol (segs.size(), false);
  set<vector<bool> > candidates;
  int lower = max((2 * read_length), (int)floor(gsl_cdf_gaussian_Pinv(Parameters::lower_percentile, Parameters::s_dev)) + fragment_length);
  int upper = ceil(gsl_cdf_gaussian_Pinv(Parameters::upper_percentile, s_dev)) + fragment_length;
  int length_lmate_last = all_segments[cover.exons_1.back()].length;
  int length_rmate_first = all_segments[cover.exons_2.front()].length;
  int lb_imprv =  min(l_mate_length, length_lmate_last - 1 + read_length) + min(r_mate_length, length_rmate_first - 1 + read_length);
  int length_lmate_first = all_segments[cover.exons_1.front()].length;
  int length_rmate_last = all_segments[cover.exons_2.back()].length;
  int ub_imprv = max(0, l_mate_length - length_lmate_first + 1 - read_length) + max(0, r_mate_length-length_rmate_last + 1 - read_length);

  SegmentType l_mate_end = all_segments[cover.exons_1.back()].type;
  SegmentType r_mate_start = all_segments[cover.exons_2.front()].type;

  std::set<UInt> legal_ends;
  addRecursiveGapExons(0, -1, segs, seg_types, l_mate_end, r_mate_start, max(0, lower - 1 - lb_imprv -1), upper - (2*read_length) - ub_imprv ,current_sol,
		  	  	  	   candidates, low, legalBlockBounds, legal_ends);
  //addRecursiveGapExons(0, segs, max(0,lower-1-l_mate_length-r_mate_length), upper+1-(2*read_length), current_sol, candidates);
  //addRecursiveGapExons(0, segs, max(0,lower-1-l_mate_length-r_mate_length), 275+1-(2*36), current_sol, candidates);
  //addRecursiveGapExons(0, segs, 0, 500, current_sol, candidates);


  for(set<vector<bool> >::const_iterator it = candidates.begin(); it != candidates.end(); it++) {
	/*Transcript tmp;
	for(mate_it = cover.exons_1.begin(); mate_it != cover.exons_1.end(); ++mate_it)
		tmp.nodes.push_back(*mate_it);
	for(int i=0; i<segs.size(); i++)
		if((*it)[i])
			tmp.nodes.push_back(*(low+i+1));
	for(mate_it = cover.exons_2.begin(); mate_it != cover.exons_2.end(); ++mate_it)
		tmp.nodes.push_back(*mate_it);
	double ll = getNormFactor(cover, tmp, all_segments);
	//cerr << ll << endl;
	if(ll<1 && cover.intensity==0)
		continue;
	*/
	valid_candidates.push_back(*it);
  }

  return !candidates.empty();
}



void addRecursiveGapExons(int pos,
                          int last_subexon,
                          vector<pair<UInt,UInt> >& segs,
                          vector<SegmentType>& seg_types,
                          SegmentType lmate_end,
                          SegmentType rmate_start,
                          UInt min_length,
                          UInt max_length,
                          vector<bool>& current_sol,
                          set<vector<bool> >& candidates,
                          std::vector<UInt>::const_iterator low,
                          const std::map<UInt, std::set<UInt> >& legalBlockBounds,
                          const std::set<UInt>& legal_ends)
{  
/*
  cout << "pos: " << pos << endl;
  for(int z=0; z<pos; z++)
	  cout << current_sol[z] << " ";
  cout << endl;
  cout << "legal ends:" << endl;
  for(std::set<UInt>::iterator it = legal_ends.begin(); it!=legal_ends.end(); it++)
	  cout << *it << " ";
  cout << endl;
  cout << "last subexon: " << last_subexon << endl;
*/

  //base case
  if(pos == segs.size()-1)
  {
    if(min_length == 0) {
    	if( rmate_start != RIGHTEND && rmate_start != INNER &&   //gap to first exon of right mate -> can't start with R or I and:
    	    ( (last_subexon == -1 && lmate_end != LEFTEND && lmate_end != INNER  && (Parameters::max_nof_exskip == 0 || segs.size() <= Parameters::max_nof_exskip) ) || //either nothing set provided left mate didn't end with L or I
    		  (last_subexon > -1 && seg_types[last_subexon] != LEFTEND && seg_types[last_subexon] != INNER && (Parameters::max_nof_exskip == 0 || pos-last_subexon <= Parameters::max_nof_exskip) ) ) )  //or something not L and not I set last
    	{
    		if(!legalBlockBounds.empty())
    		{
    			if(last_subexon == -1 || last_subexon < pos-1 || legal_ends.find(*(low+pos)) != legal_ends.end()) //if exon ends here it has to be legal
    			{											   //assumes that left mate ended an original exon!
    					candidates.insert(current_sol);
    			} else
    				cout << "base case: not allowed to end exon at last position: " << *(low+pos+1) << endl;
    		} else {
    			//cout << "last exon skipped in candidate" << endl;
    			candidates.insert(current_sol);
    		}
    	}
    }

    if(segs[pos].first <= max_length && (min_length==0 || segs[pos].first >= min_length)) {
    	//WRONG, we do not end here, cover continues!
    	//-------------------------------------------------------
//    	if(seg_types[pos] != LEFTEND && seg_types[pos] != INNER) {
    	//-------------------------------------------------------
    		if( last_subexon == pos-1 ||
    		    (seg_types[pos] != RIGHTEND  && seg_types[pos] != INNER) )  //we start here
    	    {
    			//cout << "last exon added in candidate" << endl;
    			current_sol[pos] = true;
    			candidates.insert(current_sol);
    			current_sol[pos] = false;
    		}
//    	}
    }
    //cout << "returning from base case" << endl;
    return;
  }

  if( (last_subexon == -1 && lmate_end != LEFTEND && lmate_end != INNER) ||
      (last_subexon > -1 && seg_types[last_subexon] != LEFTEND && seg_types[last_subexon] != INNER) )
  {
	  if(Parameters::max_nof_exskip == 0 || (last_subexon == -1 || pos-last_subexon <= Parameters::max_nof_exskip)) {
		//cout << "skipping" << endl;
		if( legalBlockBounds.empty() )
		{
			addRecursiveGapExons(pos+1, last_subexon, segs, seg_types, lmate_end, rmate_start, min_length, max_length, current_sol, candidates,
						         low, legalBlockBounds, legal_ends);
		} else { //exon aware
			if( last_subexon == -1 || last_subexon < pos-1  ) //no exon ends here
			{
				assert(legal_ends.empty());
				addRecursiveGapExons(pos+1, last_subexon, segs, seg_types, lmate_end, rmate_start, min_length, max_length, current_sol, candidates,
									 low, legalBlockBounds, legal_ends);
			} else if( legal_ends.find(*(low+pos)) != legal_ends.end() ) //legal exon ends here
			{
				addRecursiveGapExons(pos+1, last_subexon, segs, seg_types, lmate_end, rmate_start, min_length, max_length, current_sol, candidates,
									 low, legalBlockBounds);
			} else {
				cout << "cannot recurse without setting: " << *(low+pos+1) << endl;
			}
		}
	  } else if(min_length == 0) {
    	  if(rmate_start != RIGHTEND && rmate_start != INNER &&
    		 seg_types[last_subexon] != LEFTEND && seg_types[last_subexon] != INNER )
    	  {
    		  if( legalBlockBounds.empty() )
    		  {
    			  //cout << "found candidate without adding exon" << endl;
    			  candidates.insert(current_sol);
    		  } else //exon aware
    		  {
    			  if( last_subexon == -1 || last_subexon < pos-1  || legal_ends.find(*(low+pos)) != legal_ends.end() ) //if exon ends here it has to be legal
    			  {
    				  candidates.insert(current_sol);
    			  } else {
    				  cout << "cannot add solution with exon not set: " << *(low+pos+1) << endl;
    			  }
    		  }
    	  }
      }
  }

  if(segs[pos].first <= max_length)
  {
    if( last_subexon == pos-1 || (seg_types[pos] != RIGHTEND && seg_types[pos] != INNER))
    {
      std::set<UInt> bt_copy_ends;
      //cout << "adding" << endl;
      bool legal = true;
      if(!legalBlockBounds.empty()) //in case of exon awareness we might have to update set of legal exon ends
      {
    	  bt_copy_ends = legal_ends;
    	  if(last_subexon == -1 || last_subexon < pos-1) //if we start new legal exon initialize legal_ends
    	  {
    		  if(legalBlockBounds.find(*(low+pos+1)) != legalBlockBounds.end())
    		  {
    			  bt_copy_ends.clear();
    			  bt_copy_ends = legalBlockBounds.find(*(low+pos+1))->second;
    		  } else {
    			  legal = false; //not allowed to start new exon here
    			  cout << "not legal to start here" << pos << endl;
    		  }

    	  } else //if exon can be ended at previous position and startet at current, add new exits to legal_ends
    	  {
    		  assert(!bt_copy_ends.empty());
    		  std::set<UInt>::iterator e_it = bt_copy_ends.find(*(low+pos));
    		  if( e_it != bt_copy_ends.end() )
    		  {
    			  bt_copy_ends.erase(e_it);
    			  if(legalBlockBounds.find(*(low+pos+1)) != legalBlockBounds.end())
    				  bt_copy_ends.insert(legalBlockBounds.find(*(low+pos+1))->second.begin(), legalBlockBounds.find(*(low+pos+1))->second.end());
    		  }
    	  }
      }
      if(legalBlockBounds.empty() || legal)
      {
    	  current_sol[pos] = true;
    	  UInt new_min_length = (segs[pos].first >= min_length) ? 0 : min_length - segs[pos].first;
    	  if(!legalBlockBounds.empty())
    		  addRecursiveGapExons(pos+1, pos, segs, seg_types, lmate_end, rmate_start, new_min_length, max_length-segs[pos].first, current_sol, candidates,
    				   	   	   	   low, legalBlockBounds, bt_copy_ends);
    	  else
    		  addRecursiveGapExons(pos+1, pos, segs, seg_types, lmate_end, rmate_start, new_min_length, max_length-segs[pos].first, current_sol, candidates,
    				  	  	       low, legalBlockBounds, legal_ends);
    	  current_sol[pos] = false;
      }
    }
  }
  //cout << "returning" << endl;
  return;
}



void hasUnsupportedJunction(const std::vector<UInt>& locus,
                            const std::vector<SegmentCover>& allCovers,
                            const std::vector<UInt>& selectedCoverIds,
                            const TAllSegmentsMap& allSegments,
                            const std::vector<Transcript>& transcripts,
                            std::vector<UInt>& numUnsuppJunc,
                            UInt minGapLen,
                            bool inferNeighborEdges,
                            bool joinOnlyCoveredSubs)
{
  //create splice graph for locus
  numUnsuppJunc.clear();
  numUnsuppJunc.assign(transcripts.size(), 0);
  SpliceGraph sg(locus, allCovers, selectedCoverIds, allSegments, inferNeighborEdges, joinOnlyCoveredSubs);

  //for each known transcript check whether all of its junctions are supported
  std::vector<Transcript>::const_iterator tIt;
  for (tIt = transcripts.begin(); tIt != transcripts.end(); ++tIt)
  {
    for (size_t j = 1; j < tIt->nodes.size(); ++j)
    {
      Segment l_seg = allSegments[tIt->nodes[j-1]];
      Segment r_seg = allSegments[tIt->nodes[j]];

      if (abs((int)r_seg.beginPos - (int)l_seg.endPos) >= minGapLen && !sg.findEdge(tIt->nodes[j-1],tIt->nodes[j]))
      {
        //std::cout << tIt->nodes[j-1] << " ->  " << tIt->nodes[j] << std::endl;
        ++numUnsuppJunc[tIt - transcripts.begin()];
      }
    }
  }
}


double getNormalizedTranscriptLength (const Transcript& transcript,
                                    const TAllSegmentsMap& segments,
                                    int read_length,
                                    int fragment_length,
                                    double s_dev,
                                    double cutoff)
{
  UInt len = 0;
  for (size_t j = 0 ; j < transcript.nodes.size(); ++j)
  {
    len += segments[transcript.nodes[j]].length;
  }
  
  SegmentCover covTmp;
  covTmp.exons_1.push_back(0);
  covTmp.exons_2.push_back(0);
  TAllSegmentsMap mapTmp(1);
  mapTmp[0].beginPos = 0;
  mapTmp[0].endPos = len;
  mapTmp[0].length = len;
  Transcript transTmp;
  transTmp.nodes.push_back(0);

  return getNormFactor(covTmp, transTmp, mapTmp, read_length, fragment_length,s_dev);
  
}


//void getTranscriptSupports (std::vector<std::pair<double, UInt> >& supports,
//                              const std::vector<Transcript>& transcripts,
//                              const TAllSegmentsMap& all_segments,
//                              const std::vector<SegmentCover>& all_covers,
//                              const std::vector<UInt>& selected_cover_ids,
//                              int read_length,
//                              int fragment_length,
//                              double s_dev)
//{
//  //maps to avoid repeated computation of normfactors for the same subsequence between cover anchors
//  std::vector< std::map<unsigned, float> > known_normfs;
//  known_normfs.resize(selected_cover_ids.size());
//  
//  supports.resize(transcripts.size());
//  UInt total_calls = 0;
//  for (size_t j = 0; j < transcripts.size(); ++j)
//  {
//    const Transcript& tmp_trans = transcripts[j];
//    std::unordered_map<UInt, UInt> t_pos;
//    t_pos.reserve(100);
//    for (size_t s = 0; s < tmp_trans.nodes.size(); ++s)
//    {
//        t_pos[tmp_trans.nodes[s]] = s;
//    }
//
//    if (j%10000 == 0)
//        std::cout << j << " " << total_calls << std::endl;
//    supports[j] = std::pair<double, UInt>(0., j);
//    
//    for (size_t i = 0; i < selected_cover_ids.size(); ++i)
//    {
//      const SegmentCover& tmp_cov =  all_covers[selected_cover_ids[i]];
////      vector<UInt>::const_iterator it_l = search(tmp_trans.nodes.begin(),tmp_trans.nodes.end(), tmp_cov.exons_1.begin(), tmp_cov.exons_1.end());
//      std::unordered_map<UInt, UInt>::const_iterator m_it1 = t_pos.find(tmp_cov.exons_1.front());
//      if (m_it1  == t_pos.end())
//          continue;
//     
//      std::unordered_map<UInt, UInt>::const_iterator m_it2 = t_pos.find(tmp_cov.exons_2.front());
//      if (m_it2  == t_pos.end())
//          continue;
//
//      vector<UInt>::const_iterator it_l = tmp_trans.nodes.begin() + m_it1->second;
//      vector<UInt>::const_iterator it_r = tmp_trans.nodes.begin() + m_it2->second;
//      
//      if (tmp_trans.nodes.end() - it_l < tmp_cov.exons_1.size() || tmp_trans.nodes.end() - it_r < tmp_cov.exons_2.size() || 
//              !std::equal(it_l, it_l + tmp_cov.exons_1.size(), tmp_cov.exons_1.begin()) ||
//              !std::equal(it_r, it_r + tmp_cov.exons_2.size(), tmp_cov.exons_2.begin()))
//      {
//          continue;
//      }
///* 
//      if (it_l == tmp_trans.nodes.end() || tmp_trans.nodes.end() - it_l < tmp_cov.exons_1.size() || 
//              !std::equal(it_l, it_l + tmp_cov.exons_1.size(), tmp_cov.exons_1.begin()))
//      {
//          continue;
//      }
//
//      vector<UInt>::const_iterator it_r = find(it_l,tmp_trans.nodes.end(), tmp_cov.exons_2.front());
// 
//      if (it_r == tmp_trans.nodes.end() || tmp_trans.nodes.end() - it_r < tmp_cov.exons_2.size() || 
//              !std::equal(it_r, it_r + tmp_cov.exons_2.size(), tmp_cov.exons_2.begin()))
//      {
//          continue;
//      }
//*/
//
//      //vector<UInt>::const_iterator it_r = search(it_l,tmp_trans.nodes.end(), tmp_cov.exons_2.begin(), tmp_cov.exons_2.end());
//      //if (it_r == tmp_trans.nodes.end())
//      //    continue;
//
////      if(isCoverSupported(all_covers[selected_cover_ids[i]], tmp_trans))
////      {
//        //get the gap segment length
//        unsigned gap = 0;
////        if (all_covers[selected_cover_ids[i]].exons_1.back() < all_covers[selected_cover_ids[i]].exons_2.front())
////        {
//          std::vector<unsigned>::const_iterator gap_it = it_l + tmp_cov.exons_1.size();
//          
//          while (gap_it < it_r)
//          {
//            gap += all_segments[*gap_it].length;
//            ++gap_it;
//          }
////        }
//        
//        double normf;
//        std::map<unsigned, float>::const_iterator norm_it = known_normfs[i].find(gap);
//        if (norm_it == known_normfs[i].end())
//        {
//          ++total_calls;
//          normf = getNormFactor(all_covers[selected_cover_ids[i]], transcripts[j], all_segments, read_length, fragment_length, s_dev);
//          known_normfs[i][gap] = normf;
//        }
//        else
//        {
//          normf = norm_it->second;
//        }
//        supports[j].first += normf * tmp_cov.intensity;
////      }
//    }
//  }
//  std::sort(supports.begin(), supports.end(), std::greater<std::pair<double, UInt> >());
//  for (size_t i = 0; i < supports.size(); ++i)
//      std::cerr << transcripts[supports[i].second].name << " " << supports[i].first << std::endl;
//}



void markMandatoryCandidates(const std::vector<Transcript>& infTranscripts,
                             const TAllSegmentsMap& allSegments,
                             std::vector<bool>& isMandatory)
{
  typedef std::multimap<double, size_t> TOrderMap;
  isMandatory.assign(infTranscripts.size(), false);
  
  TOrderMap orderedTrans;
  
  for (size_t i = 0; i < infTranscripts.size(); ++i)
  {
    orderedTrans.insert(std::make_pair(infTranscripts[i].predictedExpression, i));
  }
  //sort transcripts by their predicted expression (descending)
  //std::sort(inf_transcripts.begin(), inf_transcripts.end(),[](const Transcript& t1, const Transcript& t2){return t1.predictedExpression > t2.predictedExpression})
  
  std::set<size_t> covered_tss, covered_tes;
  
  for (TOrderMap::const_reverse_iterator it = orderedTrans.rbegin(); it != orderedTrans.rend(); ++it)
  {
    size_t t = it->second;    
    if (allSegments[infTranscripts[t].nodes.front()].tss)
    {
      if (covered_tss.insert(infTranscripts[t].nodes.front()).second)
      {
        isMandatory[t] = true;
      }
    }
    if (allSegments[infTranscripts[t].nodes.back()].tes)
    {
      if (covered_tes.insert(infTranscripts[t].nodes.back()).second)
      {
        isMandatory[t] = true;
      }
    }
  }
}
