/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FunctionsIO.h>
#include <Parameters.h>
#include <Functions.h>


using namespace std;


void readQuadApproxCoeff(const string& filename, vector<double>& breakpoints, vector<pair<double,double> >& left_coeff, vector<pair<double,double> >& right_coeff, vector<pwQuadCoeff>& coeff_bp)
{
	string line;
	ifstream infile(filename.c_str());
	if(infile.is_open())
	{
		double bp;
		double ll, lq;
		double rl, rq;
		while(getline(infile, line))
		{
			stringstream line_s(line);
			line_s >> bp;
			line_s >> ll >> lq;
			line_s >> rl >> rq;

			breakpoints.push_back(bp);
			left_coeff.push_back(pair<double,double>(ll,lq));
			right_coeff.push_back(pair<double,double>(rl,rq));

			coeff_bp.push_back(pwQuadCoeff(bp,ll,lq,rl,rq));
		}
		infile.close();

		std::sort(coeff_bp.begin(), coeff_bp.end());
	}
	else
	{
		cerr<<"unable to open file: "<<filename<<endl;
	    exit(1);
	}
}

/*read a .mat file
- store all nodes for that locus ordered in node_ids
- store all transcripts in vector transcripts (a Transcript has an id and a vector of nodes)
*/
void readMatFile2(string filename, map<UInt, vector<Transcript> >& transcripts, map<UInt, vector<UInt> >& loci)
{
  UInt node_id, locus_id;
  string line;
  ifstream infile(filename.c_str());
  if(infile.is_open())
  {
    //read the nodes ids for that locus
    getline(infile, line);
    while(!infile.eof())
    {
      stringstream line_s(line);
      if(line[0]=='>')
      {
        line_s.seekg(1,ios::beg);
        line_s>>locus_id;
        //cout<<"locus_id: "<<locus_id<<endl;
        getline(infile, line);
        line_s.clear();
        line_s.str(line);
        line_s.seekg(1,ios::beg);
        while(!line_s.eof())
        {
          line_s>>node_id;
          //std::cout<<"Node ID: "<<node_id<<std::endl;
          loci[locus_id].push_back(node_id);
        }
      }

      //read the transcripts
      while(getline(infile, line))
      {
        if(line[0]=='>')
        {
          break;
        }
        UInt entry = 0;
        Transcript tmp_transcript;
        //cout<<"this should be empty: "<<line_s.str()<<endl;
        line_s.clear();
        line_s.str(line);

        //set the id of the transcript
        line_s>>tmp_transcript.id;
        //std::cout<<"Transcript id: "<<tmp_transcript.id<<std::endl;

        while(!line_s.eof())
        {
          line_s>>node_id;
          if(node_id)
          {
            tmp_transcript.nodes.push_back(loci[locus_id][entry]);
          }
          entry++;
        }
        transcripts[locus_id].push_back(tmp_transcript);
      }
    }
    infile.close();
  }
  else
  {
    cerr<<"unable to open file: "<<filename<<endl;
    exit(1);
  }
}



/*
Get the length information for each node from the .stat file
For each node, the length is stored in vector node_lengths (lenght of node with id x is node_length[x])
*/
void readSegmentLengths(string filename, map<UInt, Segment>& all_segments)
{
  all_segments.clear();

  ifstream infile(filename.c_str());
  if(infile.is_open())
  {
    string line;

    UInt node_id, length, line_count;
    float coverage;
    char  tmp_str[200];

    //skip first line
    infile.getline(tmp_str, 200);

    line_count=0;
    while(infile.getline(tmp_str, 200))
    {
      ++line_count;
    }

    //node_lengths.assign(line_count,0);
    infile.clear();
    infile.seekg(ios::beg);
    //skip first line
    infile.getline(tmp_str, 200);
    //cout<<tmp_str<<endl;

    while(getline(infile, line))
    {
      stringstream line_s(line);
      line_s >> node_id >> coverage >> length;
      Segment tmp;
      tmp.id = node_id;
      tmp.length = length;
      all_segments[node_id]=tmp;
    }
    infile.close();
  }
  else
  {
    cerr<<"unable to open file: "<<filename<<endl;
    exit(1);
  }
}


/*
read count file
*/
void readSegmentCovers(string filename, vector<SegmentCover>& node_counts, map<UInt, set<UInt> >& covers_to_segment)
{
  node_counts.clear();
  std::map<std::string, int>abund_covers;
  SegmentCover node_cov;
  //index.assign(number_nodes, set<UInt>());

  ifstream infile(filename.c_str());
  if(infile.is_open())
  {
    string line;
    //count number of entries in file
    UInt line_count=0;
    while(getline(infile, line))
    {
      ++line_count;
    }
    infile.clear();
    infile.seekg(ios::beg);
    node_counts.reserve(line_count);

    UInt index_nr=0;
    float coverage=0;

    //skip first line
    getline(infile, line);
    //parse file
    while(getline(infile, line))
    {
      stringstream strstr(line);
      std::string cover;
      strstr >> cover;
      strstr >> coverage;
      
      //try to read min dist entries if available
      int dist;
      double count;
      std::vector<std::pair<int, double> > minDistsTmp;
      
      while (strstr >> dist >> count)
      {
        minDistsTmp.push_back(std::pair<int, double>(dist,count));
      }
      
      //if the mirror of this cover is already abundant only add count
      size_t delim_pos = cover.find_first_of("^");

      if( ((Parameters::mode == Parameters::PAIRED_END) && (delim_pos == std::string::npos)) ||
          ((Parameters::mode == Parameters::SINGLE_END) && (delim_pos != std::string::npos)) )
      {
        continue;
      }

      //check whether reversed cover is already present. If so add counts
      if(delim_pos != std::string::npos)
      {
        string mirror = cover.substr(delim_pos + 1) + cover[delim_pos] + cover.substr(0, delim_pos);
        std::map<std::string, int>::iterator it = abund_covers.find(mirror);
        if(it != abund_covers.end())
        {
          node_counts[it->second].intensity +=  coverage;
          
          //add distCounts
          for (Size i = 0; i < minDistsTmp.size(); ++i)
          {
            node_counts[it->second].distCounts[minDistsTmp[i].first] += minDistsTmp[i].second;
          }
          continue;
        }
      }

      stringToCover(node_cov, cover, coverage);

      std::vector<UInt>::const_iterator ex_it;
      for(ex_it = node_cov.exons_1.begin(); ex_it != node_cov.exons_1.end(); ++ex_it)
      {
        covers_to_segment[*ex_it].insert(index_nr);
      }
      for(ex_it = node_cov.exons_2.begin(); ex_it != node_cov.exons_2.end(); ++ex_it)
      {
        covers_to_segment[*ex_it].insert(index_nr);
      }

      //correct ordering of pairs (right mate always after left mate)
      if(!node_cov.exons_2.empty() && node_cov.exons_1.back() > node_cov.exons_2.front())
      {
        std::swap(node_cov.exons_1, node_cov.exons_2);
      }

      node_counts.push_back(node_cov);
      
      //add distCounts
      for (Size i = 0; i < minDistsTmp.size(); ++i)
      {
        node_counts.back().distCounts[minDistsTmp[i].first] = minDistsTmp[i].second;
      }
      
      abund_covers[line] = index_nr;
      ++index_nr;
    }
    infile.close();
  }
  else
  {
    cerr<<"unable to open file: "<<filename<<endl;
    exit(1);
  }
}



void getSegmentCoverReads(string filename, std::vector<SegmentCover>& segment_covers, TCoverSegmentMap& covers_to_segment)
{
  segment_covers.clear();
  segment_covers.reserve(2000000);

  TAbundCoverMap abund_covers;

  ifstream infile;

  infile.open(filename.c_str());
  if(infile.is_open())
  {
    string cover_string;
    UInt pos_l, pos_r;
    double count;

    UInt index_nr = 0;
    UInt read_id = 0;
    char line[1001];
    const UInt BUFFSIZE = 100000;
    char buffer[BUFFSIZE];
    char separator[] = " \t";

    //parse file
    while(infile.good())
    {
      stringstream buff_stream;
      infile.read(buffer, BUFFSIZE);

      char * pch = strrchr(buffer,'\n');
      int length = pch - buffer;
      int offset = BUFFSIZE - length;
      buff_stream.rdbuf()->pubsetbuf(buffer, length);
      infile.seekg(-offset + 1, ios_base::cur);

      while(buff_stream.getline(line, 1000))
      {
        char *token;
        token = strtok(line, separator);
        read_id = atoi(token);
        token = strtok(NULL, separator);
        cover_string = std::string(token);
        token = strtok(NULL, separator);
        pos_l = atoi(token);
        token = strtok(NULL, separator);
        pos_r = atoi(token);
        token = strtok(NULL, separator);
        if(token == NULL)
        {
          continue;
        }
        count = atof(token);

        //std::cout<<"processing line: "<<read_nr++ << std::endl;
        //get number from read_id string
        //if((read_id % 100000) == 0)
        //std::cout<<"processing readd: "<<read_id << "  "<<count << std::endl;

        //if the cover or mirror of the cover is already abundant only add count
        TAbundCoverMap::iterator it = abund_covers.find(cover_string);

        if(it != abund_covers.end())
        {
          SegmentCover::MappedRead tmp = {read_id, pos_l, pos_r, count};
          segment_covers[it->second].mapped_reads.push_back(tmp);
          segment_covers[it->second].intensity +=  count;
          continue;
        }

        size_t delim_pos = cover_string.find_first_of("^");
        if(Parameters::mode == Parameters::PAIRED_END && delim_pos == std::string::npos)
        {
          continue;
        }

        if(Parameters::mode != Parameters::SINGLE_END && delim_pos != std::string::npos)
        {
          string mirror = cover_string.substr(delim_pos + 1) + cover_string[delim_pos] + cover_string.substr(0, delim_pos);
          it = abund_covers.find(mirror);

          if(it != abund_covers.end())
          {
            SegmentCover::MappedRead tmp = {read_id, pos_l, pos_r, count};
            segment_covers[it->second].mapped_reads.push_back(tmp);
            segment_covers[it->second].intensity +=  count;
            continue;
          }
        }

        SegmentCover node_cov;
        stringToCover(node_cov, cover_string, count);
        SegmentCover::MappedRead tmp = {read_id, pos_l, pos_r, count};
        node_cov.mapped_reads.push_back(tmp);

        std::vector<UInt>::const_iterator ex_it;

        for(ex_it = node_cov.exons_1.begin(); ex_it != node_cov.exons_1.end(); ++ex_it)
        {
          covers_to_segment[*ex_it].insert(index_nr);
        }
        if(!node_cov.exons_2.empty())
        {
          for(ex_it = node_cov.exons_2.begin(); ex_it != node_cov.exons_2.end(); ++ex_it)
          {
            covers_to_segment[*ex_it].insert(index_nr);
          }
        }

        //correct ordering of pairs (right mate always after left mate)
        if(!node_cov.exons_2.empty() && node_cov.exons_1.back() > node_cov.exons_2.front())
        {
          std::swap(node_cov.exons_1, node_cov.exons_2);
        }

        segment_covers.push_back(node_cov);
        abund_covers[cover_string] = index_nr;
        ++index_nr;
      }
    }
    infile.close();
  }
  else
  {
    cerr<<"unable to open file: "<<filename<<endl;
    exit(1);
  }
  /*
  std::vector<SegmentCover> &tmp = segment_covers;
  std::cout<<"start sorting"<<std::endl;
  std::sort(tmp.begin(), tmp.end());
  for(int i = 0; i < tmp.size() ; ++i)
  {
    std::cout << tmp[i] << '\t' << tmp[i].intensity << std::endl;
  }
  exit(1);
  */
}

void getSegmentCoverReads2(string filename,
                           std::vector<SegmentCover>& segment_covers,
                           std::vector<SuperSegmentCover>& segment_super_covers,
                           TCoverSegmentMap& covers_to_segment,
                           TCoverSegmentMap& covers_to_supercovers
                           )
{
  segment_covers.clear();
  segment_covers.reserve(2000000);
  Size uniq_counter = 0;


  TAbundCoverMap abund_super_covers, abund_covers;

  ifstream infile;
  infile.open(filename.c_str());

  if(!infile.is_open())
  {
    cerr<<"unable to open file: "<<filename<<endl;
    exit(1);
  }


  string cover_string;
  UInt pos_l, pos_r;
  double count;

  Size index_nr = 0;
  Size index_sup_nr = 0;

  UInt read_id = 0;
  UInt last_id = -1;

  std::map<std::string, SegmentCover::MappedRead> cache;

  char line[1001];
  const int BUFFSIZE = 100000;
  char buffer[BUFFSIZE];
  char separator[] = " \t";

  //parse file
  while(infile.good())
  {
    stringstream buff_stream;
    infile.read(buffer, BUFFSIZE);

    char * pch = strrchr(buffer,'\n');
    int length = pch - buffer;
    int offset = BUFFSIZE - length;
    buff_stream.rdbuf()->pubsetbuf(buffer, length);
    infile.seekg(-offset + 1, ios_base::cur);

    while(buff_stream.getline(line, 1000))
    {
      char *token;
      token = strtok(line, separator);
      read_id = atoi(token);
      token = strtok(NULL, separator);
      cover_string = std::string(token);

      size_t delim_pos = cover_string.find_first_of("^");
      if(Parameters::mode == Parameters::PAIRED_END && delim_pos == std::string::npos)
      {
        continue;
      }

      token = strtok(NULL, separator);
      pos_l = atoi(token);
      token = strtok(NULL, separator);
      pos_r = atoi(token);
      token = strtok(NULL, separator);
      if(token == NULL)
      {
        continue;
      }
      count = atof(token);

      if(read_id == last_id)
      {
        SegmentCover::MappedRead tmp = {read_id, pos_l, pos_r, count};
        cache[cover_string] = tmp;
        continue;
      }

      //some filtering
      std::map<std::string, SegmentCover::MappedRead>::const_iterator cache_it;
      double max_entry = 0;
      for(cache_it = cache.begin(); cache_it != cache.end(); ++cache_it)
      {
        max_entry = std::max(max_entry, cache_it->second.count);
      }

      std::vector<std::string> selected_cov;
      string super_cover_string;
      double selected_intens = 0.;
      for(cache_it = cache.begin(); cache_it != cache.end(); ++cache_it)
      {
        if(cache_it->second.count > 0.1 && cache_it->second.count / max_entry > 0.3)
        {
          selected_cov.push_back(cache_it->first);
          selected_intens += cache_it->second.count;
          super_cover_string +="\t" + cache_it->first;
        }
      }

      std::vector<std::string>::const_iterator sel_it = selected_cov.begin();
      for(; sel_it != selected_cov.end(); ++ sel_it)
      {
        cache[*sel_it].count = cache[*sel_it].count / selected_intens;
      }

      //for uniquely mapped reads
      if(selected_cov.size() == 1)
      {
        ++uniq_counter;
        std::vector<std::string>::const_iterator sel_it = selected_cov.begin();
        TAbundCoverMap::iterator it = abund_covers.find(*sel_it);
        if(it != abund_covers.end())
        {
          segment_covers[it->second].mapped_reads.push_back(cache[*sel_it]);
          segment_covers[it->second].intensity += 1;
        }

        else
        {
          SegmentCover node_cov;
          stringToCover(node_cov, *sel_it, 1);
          node_cov.mapped_reads.push_back(cache[*sel_it]);

          std::vector<UInt>::const_iterator ex_it;
          for(ex_it = node_cov.exons_1.begin(); ex_it != node_cov.exons_1.end(); ++ex_it)
          {
            covers_to_segment[*ex_it].insert(index_nr);
          }
          if(!node_cov.exons_2.empty())
          {
            for(ex_it = node_cov.exons_2.begin(); ex_it != node_cov.exons_2.end(); ++ex_it)
            {
              covers_to_segment[*ex_it].insert(index_nr);
            }
          }
          segment_covers.push_back(node_cov);
          abund_covers[*sel_it] = index_nr;
          ++index_nr;
        }
      }//endif uniquely mapped read

      //for multiply mapped reads
      else if(selected_cov.size() > 1)
      {
        TAbundCoverMap::iterator it = abund_super_covers.find(super_cover_string);

        //if super set already exists (and therefore also the single covers) just update count
        if(it != abund_super_covers.end())
        {
          segment_super_covers[it->second].intensity += 1;

          for(Size ii = 0; ii < selected_cov.size(); ++ii)
          {
            segment_covers[abund_covers[selected_cov[ii]]].intensity += cache[selected_cov[ii]].count;
          }
        }
        //else first add single covers if not present yet and then add the super cover
        else
        {
          SuperSegmentCover sup_cov;
          sup_cov.intensity = 1;

          std::vector<std::string>::const_iterator sel_it = selected_cov.begin();

          for( ; sel_it != selected_cov.end(); ++sel_it)
          {
            //std::cerr << cache[*sel_it].id<<std::endl;
            TAbundCoverMap::iterator it = abund_covers.find(*sel_it);
            if(it == abund_covers.end())
            {
              SegmentCover node_cov;
              stringToCover(node_cov, *sel_it, cache[*sel_it].count);

              std::vector<UInt>::const_iterator ex_it;
              for(ex_it = node_cov.exons_1.begin(); ex_it != node_cov.exons_1.end(); ++ex_it)
              {
                covers_to_segment[*ex_it].insert(index_nr);
              }
              if(!node_cov.exons_2.empty())
              {
                for(ex_it = node_cov.exons_2.begin(); ex_it != node_cov.exons_2.end(); ++ex_it)
                {
                  covers_to_segment[*ex_it].insert(index_nr);
                }
              }
              segment_covers.push_back(node_cov);
              abund_covers[*sel_it] = index_nr;
              sup_cov.covers.push_back(index_nr);
              covers_to_supercovers[index_nr].insert(index_sup_nr);
              ++index_nr;
            }
            else
            {
              segment_covers[it->second].intensity += cache[*sel_it].count;
              sup_cov.covers.push_back(it->second);
              covers_to_supercovers[it->second].insert(index_sup_nr);
            }
          }
          segment_super_covers.push_back(sup_cov);
          abund_super_covers[super_cover_string] = index_sup_nr;
          ++index_sup_nr;
        }
      }
      cache.clear();
      SegmentCover::MappedRead tmp = {read_id, pos_l, pos_r, count};
      cache[cover_string] = tmp;
      last_id = read_id;
    }
  }
  infile.close();

  std::cerr << "Number of single covers: "<< segment_covers.size()<<std::endl;
  std::cerr << "Number of super covers: "<< segment_super_covers.size()<<std::endl;
  std::cerr << "Number of single covers that are contained in at least one supercover: "<< covers_to_supercovers.size() << std::endl;
  std::cerr << "Number of uniquely mapped reads: " << uniq_counter << std::endl;
}


void readOrderingFile(const std::string& filename,
                      std::map<std::string, UInt>& t_lengths,
                      std::map<std::string, std::string>& transcript_to_gene,
                      std::map<std::string, std::set<std::string> >& transcripts_by_gene
                      /*std::map<std::string, std::vector<std::pair<int, UInt> > > &positions*/)
{
  //first get the transcript name - id pairs
  std::ifstream trans_file(filename.c_str());
  if(!trans_file.is_open())
  {
    std::cerr<<"Unable to open file: "<<filename<<" Abort"<<std::endl;
    exit(1);
  }

  size_t length;

  std::string line, gene, transcript;

  while(std::getline(trans_file,line))
  {
    if(line[0] == '>')
    {
      size_t left_delim_pos = line.find_first_of('_');
      size_t right_delim_pos = line.find_first_of('_', left_delim_pos + 1);

      gene = line.substr(left_delim_pos + 1, right_delim_pos - left_delim_pos - 1);

      left_delim_pos = line.find_first_of('_', right_delim_pos + 1);
      right_delim_pos = line.find_first_of('_', left_delim_pos + 1);

      transcript = line.substr(left_delim_pos + 1, right_delim_pos - left_delim_pos - 1);
      //std::cerr<<transcript<<std::endl;

      std::getline(trans_file,line);
      int pos = line.find_last_of(":");
      std::stringstream ss(line.substr(pos+1));
      ss >> length;

      t_lengths[transcript] = length;
      transcripts_by_gene[gene].insert(transcript);
      transcript_to_gene[transcript] = gene;
    }
  }
  trans_file.close();

}


//get the ground truth for our experiments
void getGroundTruth(std::string filename_trans,
                    std::string filename_exp,
                    std::map<UInt, double>& id_to_abundance,
                    std::map<UInt, std::string>& id_to_name)
{
  id_to_abundance.clear();

  //first get the transcript name - id pairs
  std::ifstream trans_file(filename_trans.c_str());

  if(!trans_file.is_open())
  {
    std::cerr<<"Unable to open file: "<<filename_trans<<" Abort"<<std::endl;
    exit(1);
  }

  //map transcript ids to names.
  std::map<std::string, int>name_to_id;

  std::string line, name;
  int id=0;

  std::getline(trans_file, line); //skip header line

  while(std::getline(trans_file,line))
  {
    std::stringstream ss(line);
    ss>>id;
    ss>>name;

    size_t l_pos = name.find("_Transcript_");
    if(l_pos != std::string::npos)
    {
      l_pos = name.find_first_of("_", l_pos + 1);
      size_t r_pos = name.find_first_of("_", l_pos + 1);
      name = name.substr(l_pos + 1, r_pos - l_pos - 1);
    }

    name_to_id[name] = id;
    id_to_name[id] = name;
  }
  trans_file.close();

  //now get the intensity values for each transcript
  std::ifstream exp_file(filename_exp.c_str());
  if(!exp_file.is_open())
  {
    std::cerr<<"Unable to open file: "<<filename_exp<<" Abort"<<std::endl;
    exit(1);
  }

  double abund;
  while(std::getline(exp_file, line))
  {
    std::stringstream ss(line);
    ss>>name;
    ss>>abund;

    size_t l_pos = name.find("_Transcript_");

    if(l_pos != std::string::npos)
    {
      l_pos = name.find_first_of("_", l_pos + 1);
      size_t r_pos = name.find_first_of("_", l_pos + 1);
      name = name.substr(l_pos + 1, r_pos - l_pos - 1);
    }

    if(name_to_id.count(name))
    {
      id_to_abundance[name_to_id[name]]=abund;
    }
    else
    {
      std::cerr<<"warning, the transcript "<<name<<" does not exist in our mapped data!";
    }
  }
  exp_file.close();
}


//get the ground truth for our experiments keeping original gene identifiers (strings)
void getGroundTruth(const std::string& filename_exp, map<std::string, double>& transcript_expression, std::map<std::string, UInt>& lengths, std::map<std::string, std::set<std::string> >& t_by_g)
{
  transcript_expression.clear();

  //now get the intensity values for each transcript
  std::ifstream exp_file(filename_exp.c_str());
  if(!exp_file.is_open())
  {
    std::cerr<<"Unable to open file: "<<filename_exp<<" Abort"<<std::endl;
    exit(1);
  }

  string line, name, gene;
  double abund;
  UInt length;


  while(std::getline(exp_file, line))
  {
    //std::cerr << line << std::endl;
    std::stringstream ss(line);
    ss >> name;
    ss >> gene;
    ss >> abund;
    ss >> length;
    transcript_expression[name] = abund;
    lengths[name] = length;
    t_by_g[gene].insert(name);

    /*
    size_t l_pos = name.find("_Transcript_");
    if(l_pos != std::string::npos)
    {
      l_pos = name.find_first_of("_", l_pos + 1);
      size_t r_pos = name.find_first_of("_", l_pos + 1);
      name = name.substr(l_pos + 1, r_pos - l_pos - 1);
      transcript_expression[name] = abund;
      std::cerr <<"ground Truth: " << name << " = " << abund << std::endl;
    }
    */
  }
  exp_file.close();
}





