/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FunctionsPreproc.h>

using namespace std;

void preprocessData(std::vector<SegmentCover>& covers,
                    std::vector<Transcript>& transcripts,
                    const std::vector<UInt>& locus,
                    const TAllSegmentsMap& segments)
{
  //we start by generating all possible transcripts for the given locus
  std::vector<Transcript>all_transcripts;
  generate_all_transcripts(locus, all_transcripts);

  std::vector<SegmentCover>::const_iterator covers_it;
  std::vector<UInt>::const_iterator it_l, it_r;

  //   for(size_t i=0; i<all_transcripts.size();++i)
  //   {
  //     cout<<" Transcript candidate "<<i;
  //     for(size_t j=0; j<all_transcripts[i].nodes.size();++j)
  //     {
  //       cout<<"  - "<<all_transcripts[i].nodes[j];
  //     }
  //     cout<<endl;
  //   }

  //from this set we now remove all transcripts that contain Segments,Junctions,SegmentPairs not supported by the data
  for(size_t i=0; i<all_transcripts.size();++i)
  {
    bool delete_transcript = false;
    std::vector<SegmentCover> candidate_covers;

    //check for single segments
    for(size_t j=0; j<all_transcripts[i].nodes.size(); ++j)
    {
      bool is_supported=false;

      for(covers_it=covers.begin(); covers_it!=covers.end(); ++covers_it)
      {
        it_l = find(covers_it->exons_1.begin(),covers_it->exons_1.end(), all_transcripts[i].nodes[j]);
        it_r = find(covers_it->exons_2.begin(),covers_it->exons_2.end(), all_transcripts[i].nodes[j]);
        if (it_l != covers_it->exons_1.end() || it_r != covers_it->exons_2.end())
        {
          is_supported = true;
        }
      }
      //check for junction
      if(is_supported && j+1<all_transcripts[i].nodes.size())
      {
        is_supported=false;
        for(covers_it=covers.begin(); covers_it!=covers.end(); ++covers_it)
        {
          it_l = search(covers_it->exons_1.begin(),covers_it->exons_1.end(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+2]);
          it_r = search(covers_it->exons_2.begin(),covers_it->exons_2.end(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+2]);
          if (it_l != covers_it->exons_1.end() || it_r != covers_it->exons_2.end())
          {
            is_supported=true;
          }
        }
        //cout<<"adjacent "<<all_transcripts[i].nodes[j]<<"  "<<all_transcripts[i].nodes[j+1]<<"  "<<isAdjacent(locus,all_transcripts[i].nodes[j],all_transcripts[i].nodes[j+1])<<endl;
        if(!is_supported && isAdjacent(locus,all_transcripts[i].nodes[j],all_transcripts[i].nodes[j+1]))
        {
          is_supported=true;
          SegmentCover candidate;
          candidate.exons_1.insert(candidate.exons_1.begin(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+2]);
          candidate_covers.push_back(candidate);
        }
      }
      //check for triplet
      if(is_supported && j+2<all_transcripts[i].nodes.size())
      {
        SegmentCover tmp;
        tmp.exons_1.insert(tmp.exons_1.begin(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+3]);
        //cout<<"checking tripplet: "<<tmp<<endl;
        UInt pos = getNormFactor(tmp, all_transcripts[i], segments, Parameters::read_length_global, Parameters::insert_size_global);
        cout << "Positions: " << pos << "  miss prob: " << getNoHitProb(Parameters::M_global, Parameters::alpha_global, pos)<<endl;
        if(pos>0)
        {
          is_supported = false;
          for(covers_it = covers.begin(); covers_it != covers.end(); ++covers_it)
          {
            it_l = search(covers_it->exons_1.begin(),covers_it->exons_1.end(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+3]);
            it_r = search(covers_it->exons_2.begin(),covers_it->exons_2.end(), &all_transcripts[i].nodes[j], &all_transcripts[i].nodes[j+3]);
            if (it_l != covers_it->exons_1.end() || it_r != covers_it->exons_2.end())
            {
              is_supported = true;
            }
          }
          if(!is_supported && getNoHitProb(Parameters::M_global, Parameters::alpha_global, pos) > 0.90)
          {
            is_supported = true;
            candidate_covers.push_back(tmp);
          }
        }
      }
      if(!is_supported)
      {
        delete_transcript = true;
        break;
      }
    }
    if(!delete_transcript)
    {
      transcripts.push_back(all_transcripts[i]);
      covers.insert(covers.end(), candidate_covers.begin(), candidate_covers.end());
      //tmp
      //       for(UInt x=0; x<candidate_covers.size();++x)
      //       {
      //         cout<<" adding cover:  "<<candidate_covers[x]<<endl;
      //       }

      //end tmp
    }
  }
}


void computeConnectedComponents(const std::vector<SegmentCover> &segment_covers,
                                const std::map<UInt, std::vector<UInt> > & all_loci,
                                std::map<UInt, std::vector<UInt> >&loci_cluster,
                                const std::vector<SuperSegmentCover> &all_segment_super_covers = std::vector<SuperSegmentCover>()
    )
{
  //we create a seqan graph with one vertex for each segment.
  //segments are connected if they:
  //                - are part of the same locus
  //                - are part of segment covers appearing in the same supercover
  //                - share a mapped read

  typedef seqan::Graph<seqan::Undirected<> > TGraph;
  typedef seqan::VertexDescriptor<TGraph>::Type TVertexDescriptor;
  typedef seqan::EdgeDescriptor<TGraph>::Type TEdgeDescriptor;

  TGraph graph;
  seqan::String<TVertexDescriptor> edges;
  std::set<std::pair<TVertexDescriptor, TVertexDescriptor> >edges_t;

  std::map<Size, std::set<Size> > all_segments;
  std::map<Size, std::vector<Size> > loci_inv;

  //connect all segments that are part of the same locus
  std::map<UInt, std::vector<UInt> >::const_iterator loc_it;
  for(loc_it = all_loci.begin(); loc_it != all_loci.end(); ++loc_it)
  {
    for(Size i = 0; i + 1 < loc_it->second.size(); ++i)
    {
      loci_inv[loc_it->second[i]].push_back(loc_it->first);

      edges_t.insert(std::make_pair(loc_it->second[i], loc_it->second[i+1]));
      all_segments[loc_it->second[i]].insert(loc_it->first);
    }

    loci_inv[loc_it->second.back()].push_back(loc_it->first);
    all_segments[loc_it->second.back()].insert(loc_it->first);
  }

  //Use the supercovers for joining segments due to multiple mapped reads
  std::vector<SuperSegmentCover>::const_iterator sup_it;

  for(sup_it = all_segment_super_covers.begin(); sup_it != all_segment_super_covers.end(); ++sup_it)
  {
    if(sup_it->covers.size() < 2)
    {
      cerr << "Supercover of size: " << sup_it->covers.size() <<" -- STRANGE! " << std::endl;
      continue;
    }

    SuperSegmentCover::TContainerType::const_iterator cov_it_l, cov_it_r;
    cov_it_l = cov_it_r = sup_it->covers.begin();
    ++cov_it_r;

    for( ; cov_it_r != sup_it->covers.end(); ++cov_it_r, ++cov_it_l)
    {
      Size left_id = segment_covers[*cov_it_l].exons_1[0];
      Size right_id = segment_covers[*cov_it_r].exons_1[0];

      if(left_id != right_id)
      {
        edges_t.insert(std::make_pair(left_id, right_id));
      }
    }
  }


  /*
  for(; cov_it != segment_covers.end(); ++cov_it)
  {
    //std::cerr << *cov_it << std::endl;
    for(Size i = 0; i + 1 < cov_it->exons_1.size(); ++i)
    {
      edges_t.insert(std::make_pair(cov_it->exons_1[i], cov_it->exons_1[i+1]));
    }
    for(Size i = 0; i + 1 < cov_it->exons_2.size(); ++i)
    {
      edges_t.insert(std::make_pair(cov_it->exons_2[i], cov_it->exons_2[i+1]));
    }
    if(!cov_it->exons_2.empty() && cov_it->exons_1[0] != cov_it->exons_2[0])
    {
      edges_t.insert(std::make_pair(cov_it->exons_1[0], cov_it->exons_2[0]));
    }
  }
*/


  std::cerr << "edge set constructed"<< std::endl;
  std::cerr << "number edges: " << edges_t.size() << std::endl;

  std::set<std::pair<TVertexDescriptor, TVertexDescriptor> >::const_iterator edge_it;
  for(edge_it = edges_t.begin(); edge_it != edges_t.end(); ++edge_it)
  {
    seqan::appendValue(edges, edge_it->first);
    seqan::appendValue(edges, edge_it->second);
  }

  seqan::addEdges(graph, edges, seqan::length(edges) / 2);

  seqan::String<TVertexDescriptor> comps;
  seqan::connectedComponents(comps, graph);
  std::map<Size, std::set<Size> >::iterator seg_it;
  std::set<Size> ids;
  std::map<Size, std::set<Size> > clusters, clustersS;
  for(seg_it = all_segments.begin(); seg_it != all_segments.end(); ++seg_it)
  {
    clusters[comps[seg_it->first]].insert(seg_it->second.begin(), seg_it->second.end());
    clustersS[comps[seg_it->first]].insert(seg_it->first);
    ids.insert(comps[seg_it->first]);
  }

  //Size number_comps = *std::max_element(seqan::begin(comps), seqan::end(comps));
  std::cerr << "Number of components: " << ids.size()<< std::endl;

  //  NOW JOIN SEGMENTS THAT SHARE MULTIPLE MAPPED READS

  //build index for multiple read mappings
  std::vector<std::vector<std::pair<unsigned, double> > >reads(1000000 * Parameters::M_global / 2);
  for(Size i = 0; i < segment_covers.size(); ++i)
  {
    double max_count = 0;
    for(Size read = 0; read < segment_covers[i].mapped_reads.size(); ++read)
    {
      max_count = std::max(max_count, segment_covers[i].mapped_reads[read].count);
    }
    for(Size read = 0; read < segment_covers[i].mapped_reads.size(); ++read)
    {
      if(segment_covers[i].mapped_reads[read].count > 0.05 && segment_covers[i].mapped_reads[read].count / max_count > 0.1)
      {
        reads[segment_covers[i].mapped_reads[read].id].push_back(std::make_pair(i, segment_covers[i].mapped_reads[read].count));
      }
    }
  }

  std::map<std::set<unsigned>, Size> combs;
  std::map<std::set<unsigned>, Size>::const_iterator combs_it;
  for(Size i= 0; i < reads.size(); ++i)
  {
    std::set<UInt>check;
    for(Size j = 0; j < reads[i].size(); ++j)
    {
      if(!check.insert(reads[i][j].first).second)
      {
        //invalid read
        reads[i].clear();
        break;
      }
    }
    if(reads[i].size() > 1)
    {
      combs[check] += 1;
    }
  }

  std::cerr<<"Number of combs: " << combs.size() << std::endl;
  for(combs_it = combs.begin(); combs_it != combs.end(); ++ combs_it)
  {
    std::cerr<<combs_it->second<<"\t";
    std::set<unsigned>::const_iterator seg_it;
    for(seg_it = combs_it->first.begin(); seg_it != combs_it->first.end(); ++seg_it)
    {
      std::cerr<< *seg_it << "\t";
    }
    std::cerr<<std::endl;
  }

  std::cerr << "index constructed" << std::endl;
  std::map<std::pair<Size, Size>, std::pair<double, double> > shared_counts;

  for(Size i = 0; i < reads.size(); ++i)
  {
    //if((i % 100000) == 0)
    //std::cerr<<"processing readd: "<< i << std::endl;
    for(Size l = 0; l + 1 < reads[i].size(); ++l)
    {
      for(Size r = l+1; r < reads[i].size(); ++r)
      {
        Size max_index, min_index, max_seg, min_seg;
        reads[i][l].first > reads[i][r].first ? max_index = l, min_index = r : max_index = r, min_index = l;

        max_seg = reads[i][max_index].first;
        min_seg = reads[i][min_index].first;

        shared_counts[std::make_pair(min_seg, max_seg)].first += reads[i][min_index].second;
        shared_counts[std::make_pair(min_seg, max_seg)].second += reads[i][max_index].second;

        if(segment_covers[reads[i][l].first].exons_1[0] != segment_covers[reads[i][r].first].exons_1[0])
        {
          edges_t.insert(std::make_pair(segment_covers[reads[i][l].first].exons_1[0], segment_covers[reads[i][r].first].exons_1[0]));
        }
      }
    }
  }

  std::map<std::pair<Size, Size>, std::pair<double, double> >::const_iterator shared_it;
  for(shared_it = shared_counts.begin(); shared_it != shared_counts.end(); ++shared_it)
  {
    double frac_f = shared_it->second.first / segment_covers[shared_it->first.first].intensity;
    double frac_l = shared_it->second.second / segment_covers[shared_it->first.second].intensity;

    if(std::max(frac_f, frac_l)>0.1)
    {
      std::cerr << frac_f << '\t' <<segment_covers[shared_it->first.first] << "\tA"<<std::endl;
      std::cerr << frac_l << '\t' <<segment_covers[shared_it->first.second] << "\tB" <<std::endl;
    }
    //    if(shared_it->second.first > segment_covers[shared_it->first.first].intensity)
    //    {
    //      std::cerr<<segment_covers[shared_it->first.first].intensity<<"\t"<<shared_it->second.first<<"\t"<<shared_counts2[shared_it->first]<<std::endl;
    //    }
  }

  //  for(Size read = 0; read < reads.size(); ++read)
  //  {
  //    double tmp_count = 0. ;
  //    for(Size mapping = 0; mapping <reads[read].size(); ++mapping)
  //    {
  //      tmp_count += reads[read][mapping].second;
  //    }
  //    if(std::abs(tmp_count - 1.) >= 0.05)
  //    {
  //      std::cerr<<read<<"\t"<<tmp_count<<"\t"<<reads[read].size()<<std::endl;
  //    }
  //  }


  std::cerr << "edge set constructed"<< std::endl;
  std::cerr << "number edges: " << edges_t.size() << std::endl;

  seqan::clearEdges(graph);
  seqan::clear(edges);
  seqan::clear(comps);
  for(edge_it = edges_t.begin(); edge_it != edges_t.end(); ++edge_it)
  {
    seqan::appendValue(edges, edge_it->first);
    seqan::appendValue(edges, edge_it->second);
  }

  seqan::addEdges(graph, edges, seqan::length(edges) / 2);
  std::cerr << "edges constructed" << std::endl;

  seqan::connectedComponents(comps, graph);

  ids.clear();
  for(seg_it = all_segments.begin(); seg_it != all_segments.end(); ++seg_it)
  {
    ids.insert(comps[seg_it->first]);
  }

  std::map<UInt, set<UInt> > tmp_components;
  std::map<UInt, set<UInt> >::const_iterator tmp_iter;
  for(seg_it = all_segments.begin(); seg_it != all_segments.end(); ++seg_it)
  {
    for(Size i = 0; i < loci_inv[seg_it->first].size(); ++i)
    {
      tmp_components[comps[seg_it->first]].insert(loci_inv[seg_it->first][i]);
    }
  }

  loci_cluster.clear();
  for(tmp_iter = tmp_components.begin(); tmp_iter != tmp_components.end(); ++tmp_iter)
  {
    loci_cluster[tmp_iter->first].assign(tmp_iter->second.begin(), tmp_iter->second.end());
  }

  std::cerr << "Number of components after read joining: " << ids.size() << std::endl;
}


void clusterLoci(const std::map<UInt, std::vector<UInt> > & all_loci, std::map<UInt, std::vector<UInt> >& loci_clusters)
{
  std::cout << "clustering loci..." << std::endl;
  //we start with a cluster for each exon. This cluster then contains all genes that have this exon
  std::map<UInt, std::set<UInt> >clusters;

  std::map<UInt, std::vector<UInt> >::const_reverse_iterator m_it;
  for(m_it=all_loci.rbegin(); m_it!=all_loci.rend(); ++m_it)
  {
    for(UInt i=0; i<m_it->second.size(); ++i)
    {
      clusters[m_it->second[i]].insert(m_it->first);
    }
  }
  UInt max_id = seqan::maxValue<UInt>();

  std::map<UInt, std::set<UInt> >::iterator c_it;
  std::map<UInt, UInt>gene_to_cluster;
  std::set<UInt>keys;
  for(c_it=clusters.begin(); c_it!=clusters.end(); ++c_it)
  {
    std::set<UInt>::iterator s_it;
    for(s_it=c_it->second.begin(); s_it!= c_it->second.end(); ++s_it)
    {
      //if(*s_it == 7530 || *s_it == 7531 || *s_it == 7711 || *s_it == 7713 || *s_it == 7755 || *s_it == 7756 || *s_it == 7875 || *s_it == 7876 || *s_it == 7973 || *s_it == 7974 || *s_it == 8098 || *s_it == 8099 || *s_it == 8121)
      //std::cerr<<*s_it<<"  "<<c_it->first<<std::endl;
      gene_to_cluster[*s_it] = c_it->first;
    }
    keys.insert(c_it->first);
  }

  //std::cerr<<"7530 !!"<<gene_to_cluster[7530]<<std::endl;
  //std::cerr<<"7711 !!"<<gene_to_cluster[7711]<<std::endl;

  //UInt output_count = 0;
  std::set<UInt>::const_iterator k_it;

  for (k_it = keys.begin(); k_it != keys.end(); ++k_it)
  {
    //std::cerr<<++output_count<<"  / "<<clusters.size()<<std::endl;
    std::set<UInt>::iterator s_it;
    bool remap=false;
    UInt map_id = max_id;
    if(clusters[*k_it].size() != 1)
    {
      for(s_it=clusters[*k_it].begin(); s_it!= clusters[*k_it].end(); ++s_it)
      {
        //if(*s_it == 8121)
        //std::cerr<<"to: " << map_id;
        if(gene_to_cluster[*s_it] != *k_it)
        {
          remap=true;
          map_id = std::min(map_id,gene_to_cluster[*s_it]);
        }
      }
    }
    if(remap)
    {
      for(s_it=clusters[*k_it].begin(); s_it!= clusters[*k_it].end(); ++s_it)
      {
        std::set<UInt>::iterator join_it;
        for(join_it=clusters[gene_to_cluster[*s_it]].begin(); join_it!=clusters[gene_to_cluster[*s_it]].end(); ++join_it)
        {
          clusters[map_id].insert(*join_it);
        }
      }

      std::set<UInt>::const_iterator update_it;
      for(update_it = clusters[map_id].begin(); update_it!=clusters[map_id].end(); ++update_it)
      {
        gene_to_cluster[*update_it]=map_id;
      }
    }
  }

  UInt cluster_id=0;
  std::set<UInt>inserted;
  std::map<UInt, UInt>::const_iterator g_it;

  for(g_it=gene_to_cluster.begin(); g_it!=gene_to_cluster.end(); ++g_it)
  {
    if(inserted.insert(g_it->second).second)
    {
      loci_clusters[cluster_id] = std::vector<UInt>(clusters[g_it->second].begin(), clusters[g_it->second].end());
      ++cluster_id;
    }
  }
}


void identifyTssTes (const std::vector<SegmentCover> &allCovers,
                     TAllSegmentsMap &allSegments,
                     double threshold,
                     bool onlyCoveredTerminals)
{
  typedef std::vector<SegmentCover>::const_iterator TCovIt;
  std::map<UInt, double> leftCounts, rightCounts;
  const double eps = std::numeric_limits<double>::epsilon();
  
  //fill coverage values for segments
  for (TCovIt covIt = allCovers.begin(); covIt != allCovers.end(); ++covIt)
  {
    for (Size i = 0; i < covIt->exons_1.size(); ++i)
    {
      allSegments[covIt->exons_1[i]].coverage += covIt->intensity;
    }
    
    for (Size i = 0; i < covIt->exons_2.size(); ++i)
    {
      //do not double count segment in the same segment cover e.g. i^i-j
      if (covIt->exons_2[i] <= covIt->exons_1.back())
        continue;

      allSegments[covIt->exons_2[i]].coverage += covIt->intensity;
    }
  }
  
  //reset previous entries
  for (Size i = 0; i < allSegments.size(); ++i)
  {
    if (allSegments[i].coverage > 0 || !onlyCoveredTerminals)
    {
      if (allSegments[i].type == LEFTEND || allSegments[i].type == FULL)
        allSegments[i].tes = true;
      if (allSegments[i].type == RIGHTEND || allSegments[i].type == FULL)
        allSegments[i].tss = true;
    }
    else
    {
      allSegments[i].tes = false;
      allSegments[i].tss = false;
    }
  }

  for (TCovIt covIt = allCovers.begin(); covIt != allCovers.end(); ++covIt)
  {
    for (Size i = 0; i + 1 < covIt->exons_1.size(); ++i)
    {
      if (abs(static_cast<int>(allSegments[covIt->exons_1[i]].endPos) - static_cast<int>(allSegments[covIt->exons_1[i+1]].beginPos)) > 1 )
      {
        rightCounts[covIt->exons_1[i]] += covIt->intensity;
        leftCounts[covIt->exons_1[i+1]] += covIt->intensity;
      }
    }

    for (Size i = 0; i + 1 < covIt->exons_2.size(); ++i)
    {
      if (abs(static_cast<int>(allSegments[covIt->exons_2[i]].endPos) - static_cast<int>(allSegments[covIt->exons_2[i+1]].beginPos)) > 1)
      {
        rightCounts[covIt->exons_2[i]] += covIt->intensity;
        leftCounts[covIt->exons_2[i+1]] += covIt->intensity;
      }
    }
    
    //counts implied by left/right mate without explicit junction coverage
    if (!covIt->exons_2.empty() && allSegments[covIt->exons_1.back()] < allSegments[covIt->exons_2.front()])
    {
      //check whether left subexon is really a right boundary of a real exon or if not, whether its adjacent
      //neighbor is covered
      if (allSegments[covIt->exons_1.back()].endPos != allSegments[1 + covIt->exons_1.back()].beginPos/* ||
          allSegments[1 + covIt->exons_1.back()].coverage - threshold > eps*/)
      {
        rightCounts[covIt->exons_1.back()] += covIt->intensity;
      }
      
      //check whether right subexon is really a left boundary of a real exon or if not, whether its adjacent
      //neighbor is covered
      if (allSegments[covIt->exons_2.front() - 1].endPos != allSegments[covIt->exons_2.front()].beginPos /*||
          allSegments[covIt->exons_2.front() - 1].coverage - threshold > eps */)
      {
        leftCounts[covIt->exons_2.front()] += covIt->intensity;
      }
    }
  }

  for (std::map<UInt, double>::const_iterator it = leftCounts.begin(); it != leftCounts.end(); ++it)
  {
    if (it->second - threshold > eps)
      allSegments[it->first].tss = false;
  }

  for (std::map<UInt, double>::const_iterator it = rightCounts.begin(); it != rightCounts.end(); ++it)
  {
    if (it->second - threshold > eps)
      allSegments[it->first].tes = false;
  }
}



void getTssTesFromAnnotation (TAllSegmentsMap& allSegments,
                              const TKnownTranscriptsMap& knownTranscripts)
{
  //reset all entries to false
  for (TAllSegmentsMap::iterator sIt = allSegments.begin(); sIt != allSegments.end(); ++sIt)
  {
    sIt->tes = false;
    sIt->tss = false;
  }
  
  //mark first and last subexon of every annotated transcript as possible TSS and TES
  for (TKnownTranscriptsMap::const_iterator tmapIt = knownTranscripts.begin(); tmapIt != knownTranscripts.end(); ++tmapIt)
  {
    typedef TKnownTranscriptsMap::value_type::const_iterator TTransIt;
    for (TTransIt transIt = tmapIt->begin(); transIt != tmapIt->end(); ++transIt)
    {
      allSegments[transIt->nodes.front()].tss = true;
      allSegments[transIt->nodes.back()].tes = true;
    }
  }
}



void regroupLoci(const std::vector<SegmentCover>& segment_covers,
                 TAllLociMap& all_loci,
                 const TAllSegmentsMap& all_segments)
{
  typedef seqan::Graph<seqan::Undirected<> > TGraph;
  typedef seqan::VertexDescriptor<TGraph>::Type TVertexDescriptor;
  typedef seqan::EdgeDescriptor<TGraph>::Type TEdgeDescriptor;

  TGraph graph;
  seqan::String<TVertexDescriptor> edges;
  std::set<std::pair<TVertexDescriptor, TVertexDescriptor> > edges_t;


  typedef std::vector<SegmentCover>::const_iterator TSegIt;
  for (TSegIt segIt = segment_covers.begin(); segIt != segment_covers.end(); ++segIt)
  {
    Size left_id = segIt->exons_1[0];
    for (size_t i = 1; i < segIt->exons_1.size(); ++i)
    {
      edges_t.insert(std::make_pair(left_id, segIt->exons_1[i]));
    }
    for (size_t i = 1; i < segIt->exons_2.size(); ++i)
    {
      edges_t.insert(std::make_pair(left_id, segIt->exons_2[i]));
    }
  }
  
  for (size_t i = 0; i+1 < all_segments.size(); ++i)
  {
    if (all_segments[i].contigId == all_segments[i+1].contigId &&
        all_segments[i].endPos == all_segments[i+1].beginPos)
    {
        edges_t.insert(std::make_pair(all_segments[i].id, all_segments[i+1].id));
    }
  }

  typedef std::set<std::pair<TVertexDescriptor, TVertexDescriptor> >::const_iterator TEdgeIt;
  for(TEdgeIt edge_it = edges_t.begin(); edge_it != edges_t.end(); ++edge_it)
  {
    seqan::appendValue(edges, edge_it->first);
    seqan::appendValue(edges, edge_it->second);
  }

  seqan::addEdges(graph, edges, seqan::length(edges) / 2);

  seqan::String<TVertexDescriptor> comps;
  seqan::connectedComponents(comps, graph);

  all_loci.clear();
  std::vector<size_t> mappIds;
  size_t nextId(0);

  for (size_t i = 0; i < seqan::length(comps); ++i)
  {
    size_t oldId = comps[i];
    if (oldId >= mappIds.size())
      mappIds.resize(oldId + 1, -1);

    if(mappIds[oldId] == -1)
      mappIds[oldId] = nextId++;

    if (all_loci.size() <= mappIds[oldId])
    {
      all_loci.reserve(2*mappIds[oldId]);
      all_loci.resize(mappIds[oldId]+1);
    }
    all_loci[mappIds[oldId]].push_back(i);
  }
  
  for (TAllLociMap::iterator it = all_loci.begin(); it != all_loci.end(); ++it)
  {
    std::sort(it->begin(), it->end());
  }
}
