/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Structures.h>
#include <seqan/graph_algorithms.h>
#include <Functions.h>
#include <FunctionsGraph.h>

SpliceGraph::SpliceGraph(const std::vector<UInt>& locus,
                         const TAllSegmentsMap& allSegments,
                         const TLegalExonBoundMap& legalBlockBounds)
{  

  _coverBased = false;
  _buildGraph(locus, allSegments);

  //seqan::resizeEdgeMap(graph, _isCovered, true);
  seqan::resizeEdgeMap(edgeCargo, graph, EdgeCargo(1));

  //set edge weights
  TNodeIdMap::const_iterator nIt;
  for (nIt = nodeIds.begin(); nIt != nodeIds.end(); ++nIt)
  {
    unsigned weight = allSegments[nIt->first].length;
    seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type eIt(graph, nIt->second);
    while (!seqan::atEnd(eIt))
    {
      seqan::property(edgeCargo, *eIt)._edgeWeight = weight;
      ++eIt;
    }
  }

  _setEdgeLengths(allSegments);

  _exonAware = !legalBlockBounds.empty();

  if (_exonAware)
  {
    _setLegalBlockBounds(locus, allSegments, legalBlockBounds);
  }
}



void SpliceGraph::_buildGraph(const std::vector<UInt>& locus,
                              const TAllSegmentsMap& allSegments)
{
  nodeIdsInv.clear();
  nodeIds.clear();
  seqan::clear(graph);
  
  source = seqan::addVertex(graph);
  for(Size i = 0; i < locus.size(); ++i)
  {
    TVertexDescriptor v_id = seqan::addVertex(graph);
    nodeIds[locus[i]] = v_id;
    nodeIdsInv[v_id] = locus[i];
  }
  target = seqan::addVertex(graph);


  for(Size i = 0; i < locus.size(); ++i)
  {
    //connect vertex to source and sink
    const Segment& seg = allSegments[locus[i]];

    SegmentType typeL = seg.type;

    if (typeL != RIGHTEND && typeL != INNER && seg.tss)
      seqan::addEdge(graph, source, nodeIds[locus[i]]);

    if (typeL != LEFTEND && typeL != INNER && seg.tes)
      seqan::addEdge(graph, nodeIds[locus[i]], target);

    //neighboring subexons can always be connected
    if (i+1 < locus.size())
      seqan::addEdge(graph, nodeIds[locus[i]], nodeIds[locus[i+1]]);

    //continue if only edge to next subexon on right side
    if (typeL != RIGHTEND && typeL != FULL)
      continue;

    for(Size j = i+2; j < locus.size(); ++j)
    {
      const Segment& seg2 = allSegments[locus[j]];

      SegmentType typeR = seg2.type;

      if (typeR != LEFTEND && typeR != FULL)
        continue;

      seqan::addEdge(graph, nodeIds[locus[i]], nodeIds[locus[j]]);
    }
  }
}

/*
SpliceGraph::SpliceGraph(const std::vector<UInt> &locus,
                         const std::vector<SegmentCover> &allCovers,
                         const std::vector<UInt> &selectedCoverIds,
                         const std::map<UInt, Segment> &allSegments,
                         bool inferNeighborEdges,
                         const TLegalExonBoundMap &legalBlockBounds)
{

  _exonAware = !legalBlockBounds.empty();

  //std::map<UInt, TVertexDescriptor> node_ids;
  std::map<UInt, bool> observedNodeIds;

  nodeIdsInv.clear();
  nodeIds.clear();
  seqan::clear(graph);

  source = seqan::addVertex(graph);
  target = seqan::addVertex(graph);

  for(Size i = 0; i < locus.size(); ++i)
  {
    TVertexDescriptor v_id = seqan::addVertex(graph);
    nodeIds[locus[i]] = v_id;
    nodeIdsInv[v_id] = locus[i];
  }

  //create edge set according to segment covers
  std::vector<UInt>::const_iterator covIt;
  std::map<UInt, TVertexDescriptor>::const_iterator left, right;

  for(covIt = selectedCoverIds.begin(); covIt != selectedCoverIds.end(); ++covIt)
  {
    left = right = nodeIds.find(allCovers[*covIt].exons_1[0]);
    observedNodeIds[left->first] = true;

    for(Size i = 0; i + 1 < allCovers[*covIt].exons_1.size(); ++i)
    {
      right = nodeIds.find(allCovers[*covIt].exons_1[i+1]);
      if(left == nodeIds.end() || right == nodeIds.end())
      {
        left = right;
        continue;
      }
      observedNodeIds[right->first] = true;
      if(!seqan::findEdge(graph, left->second, right->second))
      {
        seqan::addEdge(graph, left->second, right->second);
      }
      left = right;
    }

    if(!allCovers[*covIt].exons_2.empty())
    {
      observedNodeIds[allCovers[*covIt].exons_2[0]] = true;
      left = nodeIds.find(allCovers[*covIt].exons_2[0]);

      //connect neighboring exons j,j+1 if they are covered by a pair x-j^j+1-y
      if (inferNeighborEdges && left != nodeIds.end() && right != nodeIds.end() &&
          left->second - right->second == 1 && !seqan::findEdge(graph, right->second, left->second)
          )
      {
        //right is still the last node of exons 1
        seqan::addEdge(graph, right->second, left->second);
      }


      for(Size i = 0; i + 1 < allCovers[*covIt].exons_2.size(); ++i)
      {
        observedNodeIds[allCovers[*covIt].exons_2[i+1]] = true;
        right = nodeIds.find(allCovers[*covIt].exons_2[i+1]);
        if(left == nodeIds.end() || right == nodeIds.end())
        {
          left = right;
          continue;
        }
        if(!seqan::findEdge(graph, left->second, right->second))
        {
          seqan::addEdge(graph, left->second, right->second);
        }
        left = right;
      }
    }
  }


  //for all vertices add edge from the source and to the target vertex
  std::map<UInt, TVertexDescriptor>::const_iterator vertices;

  for(vertices = nodeIds.begin(); vertices != nodeIds.end(); ++vertices)
  {
    if(observedNodeIds[vertices->first])
    {
      //connect vertex to source and sink
      const std::map<UInt,Segment>::const_iterator it = allSegments.find(vertices->first);
      if (it == allSegments.end())
      {
        std::cerr << "Critical Error. Unknown Segment id in SpliceGraph Ctor" << std::endl;
        exit(-1); 
      }

      SegmentType typeL = it->second.type;

      if (typeL != RIGHTEND && typeL != INNER && it->second.tss)
        seqan::addEdge(graph, source, vertices->second);

      if (typeL != LEFTEND && typeL != INNER && it->second.tes)
        seqan::addEdge(graph, vertices->second, target);
    }
  }

  if (_exonAware)
  {
    _setLegalBlockBounds(locus, allSegments, legalBlockBounds);
  }
}*/

SpliceGraph::SpliceGraph(const std::vector<UInt>& locus,
                         const std::vector<SegmentCover>& allCovers,
                         const std::vector<UInt>& selectedCoverIds,
                         const TAllSegmentsMap& allSegments,
                         bool inferNeighborEdges,
                         bool joinOnlyCoveredSubs,
                         const TLegalExonBoundMap& legalBlockBounds)
{
  _coverBased = true;
  _exonAware = !legalBlockBounds.empty();
  _buildGraph(locus, allSegments);

  //seqan::resizeEdgeMap(graph, _isCovered, -1);
  seqan::resizeEdgeMap(edgeCargo, graph, EdgeCargo(-1));

  //std::map<UInt, TVertexDescriptor> node_ids;  
  std::set<TEdgeDescriptor> coveredEdges;

  //create edge set according to segment covers
  std::vector<UInt>::const_iterator covIt;
  std::map<UInt, TVertexDescriptor>::const_iterator left, right;

  for(covIt = selectedCoverIds.begin(); covIt != selectedCoverIds.end(); ++covIt)
  {
    left = right = nodeIds.find(allCovers[*covIt].exons_1[0]);
    _observedNodeIds.insert(left->first);

    for(Size i = 0; i + 1 < allCovers[*covIt].exons_1.size(); ++i)
    {
      right = nodeIds.find(allCovers[*covIt].exons_1[i+1]);
      if(left == nodeIds.end() || right == nodeIds.end())
      {
        left = right;
        continue;
      }
      _observedNodeIds.insert(right->first);
      if (TEdgeDescriptor tmp = seqan::findEdge(graph, left->second, right->second))
      {
        //seqan::assignProperty(_isCovered, tmp, 1);
        seqan::property(edgeCargo, tmp)._isCovered = 1;
        coveredEdges.insert(tmp);
      }
      left = right;
    }

    if(!allCovers[*covIt].exons_2.empty())
    {
      _observedNodeIds.insert(allCovers[*covIt].exons_2[0]);
      left = nodeIds.find(allCovers[*covIt].exons_2[0]);

      //connect neighboring exons j,j+1 if they are covered by a pair x-j^j+1-y
      if (inferNeighborEdges && left != nodeIds.end() && right != nodeIds.end() &&
          left->second - right->second == 1)
      {
        //right is still the last node of exons 1
        if (TEdgeDescriptor tmp = seqan::findEdge(graph, right->second, left->second))
        {
          //seqan::assignProperty(_isCovered, tmp, 1);
          seqan::property(edgeCargo, tmp)._isCovered = 1;
          coveredEdges.insert(tmp);
        }
      }

      for(Size i = 0; i + 1 < allCovers[*covIt].exons_2.size(); ++i)
      {
        _observedNodeIds.insert(allCovers[*covIt].exons_2[i+1]);
        right = nodeIds.find(allCovers[*covIt].exons_2[i+1]);
        if(left == nodeIds.end() || right == nodeIds.end())
        {
          left = right;
          continue;
        }
        if (TEdgeDescriptor tmp = seqan::findEdge(graph, left->second, right->second))
        {
          //seqan::assignProperty(_isCovered, tmp, 1);
          seqan::property(edgeCargo, tmp)._isCovered = 1;
          coveredEdges.insert(tmp);
        }
        left = right;
      }
    }
  }


  //connect neighboring segments of the same exon even if they are not covered!
  size_t runLength = 0;
  bool anySupported = false;
  for(Size i = 0; i < locus.size(); ++i)
  {
    anySupported = !joinOnlyCoveredSubs && (anySupported || _observedNodeIds.count(locus[i]));

    if (i+1 < locus.size() && allSegments[locus[i]].endPos == allSegments[locus[i+1]].beginPos)
    {
      if (!joinOnlyCoveredSubs || (_observedNodeIds.count(locus[i]) && _observedNodeIds.count(locus[i+1])))
      {
        TEdgeDescriptor tmp = seqan::findEdge(graph, getNodeId(locus[i]), getNodeId(locus[i+1]));
        //seqan::assignProperty(_isCovered, tmp, 1);
        seqan::property(edgeCargo, tmp)._isCovered = 1;
        coveredEdges.insert(tmp);
        ++runLength;
      }
    }
    else
    {
      if (anySupported)
      {
        for (size_t b = 0; b <= runLength; ++b)
          _observedNodeIds.insert(locus[i-b]);
      }
      runLength = 0;
      anySupported = false;
    }
  }


  //for all vertices add edge from the source and to the target vertex and flag observed vertices in isCovered
  std::map<UInt, TVertexDescriptor>::const_iterator vertices;

  for(vertices = nodeIds.begin(); vertices != nodeIds.end(); ++vertices)
  {
    if(_observedNodeIds.count(vertices->first))
    {
      //if node is observed all uncovered outgoing edges are flagged as 0 in isCovered
      seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type outEdgeIt(graph, vertices->second);
      for(;!atEnd(outEdgeIt);goNext(outEdgeIt))
      {
        if (seqan::property(edgeCargo, *outEdgeIt)._isCovered == -1)
        {
          //seqan::assignProperty(_isCovered, *outEdgeIt, 0);
          seqan::property(edgeCargo, *outEdgeIt)._isCovered = 0;
        }
      }

      //connect vertex to source and sink
      const Segment& seg = allSegments[vertices->first];

      SegmentType typeL = seg.type;

      if (typeL != RIGHTEND && typeL != INNER && seg.tss)
      {
        TEdgeDescriptor tmp = seqan::findEdge(graph, source, vertices->second);
        //seqan::assignProperty(_isCovered, tmp, 1);
        seqan::property(edgeCargo, tmp)._isCovered = 1;
        coveredEdges.insert(tmp);
      }

      if (typeL != LEFTEND && typeL != INNER && seg.tes)
      {
        TEdgeDescriptor tmp = seqan::findEdge(graph, vertices->second, target);
        //seqan::assignProperty(_isCovered, tmp, 1);
        seqan::property(edgeCargo, tmp)._isCovered = 1;
        coveredEdges.insert(tmp);
      }
    }
  }

  //Create the filtered graph graphCov
  seqan::_createVertices(graphCov, target);
  std::vector<TVertexDescriptor> edges;
  edges.reserve(2* coveredEdges.size());
  std::set<TEdgeDescriptor>::const_iterator eIt;
  for (eIt = coveredEdges.begin(); eIt != coveredEdges.end(); ++eIt)
  {    
    edges.push_back(seqan::sourceVertex(graph, *eIt));
    edges.push_back(seqan::targetVertex(graph, *eIt));
  }

  seqan::addEdges(graphCov, edges, coveredEdges.size());
  seqan::resizeEdgeMap(edgeCargoCov, graphCov, EdgeCargo(1));


  //set edge weights of both graphs
  TNodeIdMap::const_iterator nIt;
  for (nIt = nodeIds.begin(); nIt != nodeIds.end(); ++nIt)
  {
    unsigned weight = allSegments[nIt->first].length;
    seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type eIt(graph, nIt->second);
    while (!seqan::atEnd(eIt))
    {
      seqan::property(edgeCargo, *eIt)._edgeWeight = weight;
      ++eIt;
    }
    seqan::Iterator<TGraph, seqan::OutEdgeIterator>::Type eIt2(graphCov, nIt->second);
    while (!seqan::atEnd(eIt2))
    {
      seqan::property(edgeCargoCov, *eIt2)._edgeWeight = weight;
      ++eIt2;
    }
  }


  //set exon awareness information
  if (_exonAware)
  {
    _setLegalBlockBounds(locus, allSegments, legalBlockBounds);
  }

  //compute how for each edge how many exon skipping events it implies
  _setEdgeLengths(allSegments);


  //check supported pairs of subexons (a pair is supported if it appears in the same cover)
  //later used for transcript filtering
  suppPairs.assign(nodeIds.size()+2, std::vector<bool>(nodeIds.size()+2, false));
  for(covIt = selectedCoverIds.begin(); covIt != selectedCoverIds.end(); ++covIt)
  {
    SegmentCover::TContainerType::const_iterator ex1It = allCovers[*covIt].exons_1.begin();

    for (; ex1It != allCovers[*covIt].exons_1.end(); ++ex1It)
    {
      SegmentCover::TContainerType::const_iterator ex2It = allCovers[*covIt].exons_2.begin();
      for (; ex2It != allCovers[*covIt].exons_2.end(); ++ex2It)
      {
        suppPairs[getNodeId(*ex1It)][getNodeId(*ex2It)] = true;
      }
    }
  }
}



void SpliceGraph::getCoveredGraph(TGraph &g) const
{
  if (_coverBased)
    g = graphCov;
  else
    g = graph;
}



void SpliceGraph::_setEdgeLengths(const TAllSegmentsMap& allSegments)
{
  //std::cerr << "Start set Edge Lengths!" << std::endl;
  if (nodeIds.size() < 2)
    return;

  std::vector<bool> hasGap(nodeIds.size()-1, false);
  TNodeIdMap::const_iterator lIt, rIt1;

  rIt1 = nodeIds.begin();
  lIt = rIt1++;

  for (size_t i = 0; rIt1 != nodeIds.end(); ++lIt, ++rIt1, ++i)
  {
    if (allSegments[lIt->first].endPos != allSegments[rIt1->first].beginPos)
      hasGap[i] = true;
  }

  //std::cerr << "AA" << std::endl;
  std::vector<std::vector<unsigned> > lengths(nodeIds.size(), std::vector<unsigned>(nodeIds.size(),0));

  for (size_t i = 0; i+1 < nodeIds.size(); ++i)
  {
    int len = 0;
    for (size_t j = i+1; j < nodeIds.size(); ++j)
    {
      len += hasGap[j-1];
      lengths[i][j] = len;
    }
  }

  //std::cerr << "BB" << std::endl;
  TEdgeIterator eIt(graph);
  while(!seqan::atEnd(eIt))
  {
    TVertexDescriptor src = seqan::sourceVertex(eIt);
    TVertexDescriptor tgt = seqan::targetVertex(eIt);
    if (src != source && tgt != target)
    {
      unsigned len = lengths[src-1][tgt-1];
      if (len > 1)
      {
        seqan::property(edgeCargo, *eIt)._edgeLen = len-1;
        //std::cerr << src-1 << " " << tgt-1 << "  " << len-1 << std::endl;
      }      
    }
    seqan::goNext(eIt);
  }

  //std::cerr << "CC" << std::endl;
  if (_coverBased)
  {
    TEdgeIterator eIt2(graphCov);
    while(!seqan::atEnd(eIt2))
    {
      TVertexDescriptor src = seqan::sourceVertex(graphCov, *eIt2);
      TVertexDescriptor tgt = seqan::targetVertex(graphCov, *eIt2);
      if (src != source && tgt != target)
      {
        unsigned len = lengths[src-1][tgt-1];
        if (len > 1)
        {
          seqan::property(edgeCargoCov, *eIt2)._edgeLen = len-1;
          //std::cerr << src-1 << " " << tgt-1 << "  " << len-1 << std::endl;
        }
      }
      ++eIt2;
    }
  }

  //std::cerr << "Done set Edge Lengths!" << std::endl;

/*
  TNodeIdMap::const_iterator lIt, rIt1, rIt2;
  for (lIt = nodeIds.begin(); lIt != nodeIds.end(); ++lIt)
  {
    int len = 0;
    rIt1 = rIt2 = lIt;
    ++rIt2;
    for (; rIt2 != nodeIds.end(); ++rIt1, ++rIt2)
    {
      if (allSegments.find(rIt1->first)->second.endPos != allSegments.find(rIt2->first)->second.beginPos)
      {
        ++len;
      }

      if (TEdgeDescriptor tmp = seqan::findEdge(graph, lIt->second, rIt2->second))
      {
        if (len > 1)
          seqan::property(edgeCargo, tmp)._edgeLen = len-1;
      }
      if (_coverBased)
      {
        if (TEdgeDescriptor tmp = seqan::findEdge(graphCov, lIt->second, rIt2->second))
        {
          if (len > 1)
            seqan::property(edgeCargoCov, tmp)._edgeLen = len-1;
        }
      }
    }
  }
  */
}


void SpliceGraph::_setLegalBlockBounds(const std::vector<UInt>& locus,
                                       const TAllSegmentsMap& allSegments,
                                       const TLegalExonBoundMap& legalBlockBounds)
{
  //for exon awareness mark edges representing real introns and set legal start-end combinations of
  //subexon blocks
  //seqan::resizeEdgeMap(graph, _intronEdge, true);

  //if (_coverBased)
  //  seqan::resizeEdgeMap(graph, _intronEdgeCov, true);

  for (size_t i = 0; i + 1 < locus.size(); ++i)
  {
    if (allSegments[locus[i]].endPos == allSegments[locus[i+1]].beginPos)
    {
      //seqan::assignProperty(_intronEdge, seqan::findEdge(graph, nodeIds[locus[i]], nodeIds[locus[i+1]]), false);
      seqan::property(edgeCargo, seqan::findEdge(graph, getNodeId(locus[i]), getNodeId(locus[i+1])))._intronEdge = false;

      if (_coverBased)
      {
        if (TEdgeDescriptor tmp = seqan::findEdge(graphCov, getNodeId(locus[i]), getNodeId(locus[i+1])))
        {
          //seqan::assignProperty(_intronEdgeCov, tmp, false);
          seqan::property(edgeCargoCov, tmp)._intronEdge = false;
        }
      }
    }
  }

  seqan::resizeVertexMap(_validBlockExit, graph);
  for (size_t i = 0; i < locus.size(); ++i)
  {
    TVertexDescriptor entryV = getNodeId(locus[i]);
    std::map<UInt, std::set<UInt> >::const_iterator entryIt = legalBlockBounds.find(locus[i]);
    if (entryIt != legalBlockBounds.end())
    {
      std::set<UInt>::const_iterator exitIt = entryIt->second.begin();
      for (; exitIt != entryIt->second.end(); ++exitIt)
      {
        TNodeIdMap::const_iterator nIt = nodeIds.find(*exitIt);
        if (nIt != nodeIds.end())
        {
          seqan::property(_validBlockExit, entryV).push_back(nIt->second);
          //std::cerr << entryV << " : " << nIt->second << "  (" << locus[i] << " , " << *exitIt << ")" << std::endl;
        }
      }
    }
  }
  for (size_t i = 0; i < seqan::length(_validBlockExit); ++i)
  {
    std::sort(_validBlockExit[i].begin(), _validBlockExit[i].end());
  }
}



void SpliceGraph::appendValidTranscripts(std::vector<Transcript>& transcripts,
                                         unsigned maxUncovered,
                                         unsigned maxExonSkip,
                                         bool applyFilter,
                                         bool exonAware) const
{
  if (exonAware && !_exonAware)
  {
    std::cerr << "calling appendValidTranscripts as exon aware without aware initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  if (maxUncovered != -1u && !_coverBased)
  {
    std::cerr << "calling appendValidTranscripts with maxUncovered != -1u without cover based initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  TStringSet allPaths;

  if (!exonAware)
  {
    if (maxUncovered && _coverBased)
    {
      //enumerateAcyclicPaths(graph, source, target, _isCovered, maxUncovered, _edgeLen, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graph, source, target, edgeCargo, maxUncovered, maxExonSkip, allPaths);
      detectFeasiblePath(graph, source, target, edgeCargo, maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
    else if (!maxUncovered && _coverBased)
    {
      //enumerateAcyclicPaths(graphCov, source, target, _edgeLenCov, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graphCov, source, target, edgeCargoCov, maxExonSkip, allPaths);
      detectFeasiblePath(graphCov, source, target, edgeCargoCov, maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
    else
    {
      //enumerateAcyclicPaths(graph, source, target, _edgeLen, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graph, source, target, edgeCargo, maxExonSkip, allPaths);
      detectFeasiblePath(graph, source, target, edgeCargo, maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
  }
  else
  {
    if (maxUncovered && _coverBased)
    {
      //enumerateAcyclicPaths(graph, source, target, _isCovered, maxUncovered, _intronEdge, _validBlockExit, _edgeLen, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graph, source, target, edgeCargo, maxUncovered, _validBlockExit, maxExonSkip, allPaths);
      detectFeasiblePath(graph, source, target, source, edgeCargo, _validBlockExit, seqan::String<bool>(), maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
    else if (!maxUncovered && _coverBased)
    {
      //enumerateAcyclicPaths(graphCov, source, target, _intronEdgeCov, _validBlockExit, _edgeLenCov, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graphCov, source, target, edgeCargoCov, _validBlockExit, maxExonSkip, allPaths);
      detectFeasiblePath(graphCov, source, target, source, edgeCargoCov, _validBlockExit, seqan::String<bool>(), maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
    else
    {
      //enumerateAcyclicPaths(graph, source, target, _intronEdge, _validBlockExit, _edgeLen, maxExonSkip, allPaths);
      //enumerateAcyclicPaths(graph, source, target, edgeCargo, _validBlockExit, maxExonSkip, allPaths);
      detectFeasiblePath(graph, source, target, source, edgeCargo, _validBlockExit, seqan::String<bool>(), maxUncovered, std::make_pair(0u, -1u), std::make_pair(0u, -1u), maxExonSkip, allPaths);
    }
  }

  //paths to transcripts  
  transcripts.reserve(transcripts.size() + seqan::length(allPaths));

  typedef seqan::Iterator<TStringSet>::Type TSetIterator;
  TSetIterator setIt(allPaths);

  std::set<std::vector<UInt> >testSet;
  for(; !seqan::atEnd(setIt); seqan::goNext(setIt))
  {
    if (applyFilter && _coverBased && maxUncovered && !_hasTransSupport(*setIt))
      continue;

    transcripts.resize(transcripts.size()+1);

    Transcript & tRef = transcripts.back();
    tRef.id = transcripts.size()-1;
    tRef.nodes.reserve(seqan::length(*setIt));

    //do not copy the artificial source and target
    for(size_t i = 1; i + 1 < seqan::length(*setIt); ++i)
    {
      tRef.nodes.push_back(getSegmentId((*setIt)[i]));
      if (_coverBased && i+2 < seqan::length(*setIt) &&
          /*seqan::property(edgeCargo, seqan::findEdge(graph, (*setIt)[i], (*setIt)[i+1]))._intronEdge &&*/
          !seqan::findEdge(graphCov, (*setIt)[i], (*setIt)[i+1]))
      {
        //std::cerr << "error should not be uncovered! " << std::endl;  
        //exit(1);  
        ++tRef.uncovered;
      }
    }
    if(!testSet.insert(tRef.nodes).second)
    {
      std::cerr << "error repeated transcript!" << std::endl;
      exit(-1);
    }
  }
}


/*
bool SpliceGraph::isCoverGapSupported(const SegmentCover &cov,
                                      const std::pair<unsigned, unsigned> &minmaxInsert,                                      
                                      const std::map<UInt,Segment> &allSegments,
                                      UInt maxUncovered,
                                      UInt maxExonSkip,
                                      bool exonAware)
{

  if (maxUncovered != -1u && !_coverBased)
  {
    std::cerr << "calling isCoverGapSupported with maxUncoverd on spliceGraph not based on counts -- Terminating" << std::endl;
    exit(1);
  }

  if (exonAware && !_exonAware)
  {
    std::cerr << "calling appendValidTranscripts as exon aware without aware initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  if (cov.exons_2.empty())
    return false;

  TVertexDescriptor start = nodeIds[cov.exons_1.back()];
  TVertexDescriptor end = nodeIds[cov.exons_2.front()];
  TVertexDescriptor blockStart = start;


  unsigned offset(-1u); //we have to subtract the length of start exon as this length is summed to complete length during path search.

  if (exonAware)
  {
    //find blockEntry
    size_t lastEndPos = allSegments.find(cov.exons_1[0])->second.endPos;
    std::map<UInt,Segment>::const_iterator segIt;
    for (size_t i = 1; i < cov.exons_1.size(); ++i)
    {
      segIt = allSegments.find(cov.exons_1[i]);
      if (segIt->second.beginPos != lastEndPos)
        blockStart = nodeIds[cov.exons_1[i]];

      lastEndPos = segIt->second.endPos;
    }
    offset = segIt->second.length;
  }
  else
  {
    offset = allSegments.find(cov.exons_1.back())->second.length;
  }

  seqan::String<TVertexDescriptor> path;

  TGraph *gPtr = &graph;
  //seqan::String<unsigned> *edgeLenPtr = &_edgeLen;
  seqan::String<EdgeCargo> *edgeCargoPtr = &edgeCargo;

  if (_coverBased && !maxUncovered)
  {
    gPtr = &graphCov;
    //edgeLenPtr = &_edgeLenCov;
    edgeCargoPtr = &edgeCargoCov;
  }

  if (!exonAware)
    detectFeasiblePath(*gPtr,
                       start,
                       end,
                       *edgeCargoPtr,
                       maxUncovered,
                       minmaxInsert.first + offset,
                       minmaxInsert.second + offset,
                       maxExonSkip,
                       path);
  else
    detectFeasiblePath(*gPtr,
                       start,
                       end,
                       blockStart,
                       *edgeCargoPtr,
                       _validBlockExit,
                       seqan::String<bool>(),
                       maxUncovered,
                       minmaxInsert.first + offset,
                       minmaxInsert.second + offset,
                       maxExonSkip,
                       path);

  //if (!seqan::empty(path))
  //  std::cerr << path << " :: " << cov << std::endl;
  return !seqan::empty(path);
}
*/



bool SpliceGraph::_hasTransSupport (const seqan::String<TVertexDescriptor>& trans) const
{
  for (size_t i = 0; i+1 < seqan::length(trans); ++i)
  {
    if (!seqan::findEdge(graphCov, trans[i], trans[i+1]))
    {
      //is this missed junction covered by a pair?
      bool supp = false;
      for (size_t l = i+1; l > 0; --l)
      {
        for (size_t r = i+1; r < seqan::length(trans); ++r)
        {
          if (suppPairs[trans[l-1]][trans[r]])
          {
            supp = true;
            break;
          }
        }
        if (supp)
          break;
      }

      if (!supp)
        return false;
    }
  }
  return true;
}


bool SpliceGraph::isCoverGapSupported(const TAllSegmentsMap& allSegments,
                                      const SegmentCover& cover,
                                      const std::pair<unsigned, unsigned>& minMaxGap,
                                      UInt maxUncovered,
                                      UInt maxExonSkip,
                                      bool exonAware) const
{
  std::vector<std::vector<UInt> > exonPaths;
  _getExonPaths(allSegments, cover, minMaxGap, maxUncovered, maxExonSkip, exonAware, exonPaths, true);

  return !exonPaths.empty();
}



void SpliceGraph::_getExonPaths(const TAllSegmentsMap& allSegments,
                        const SegmentCover& cover,
                        const std::pair<unsigned, unsigned>& fragLenBounds,
                        UInt maxUncovered,
                        UInt maxSkip,
                        bool exonAware,
                        std::vector<std::vector<UInt> >& exonPaths,
                        bool stopAfterFirstPath) const
{
  if (exonAware && !_exonAware)
  {
    std::cerr << "calling getExonPaths as exon aware without aware initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  if (maxUncovered != -1u && !_coverBased)
  {
    std::cerr << "calling getExonPaths as with maxUncovere != -1u without cover based initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  const TGraph *graphPtr = 0;
  const seqan::String<EdgeCargo> * cargoPtr = 0;

  if (maxUncovered)
  {
    graphPtr = &graph;
    cargoPtr = &edgeCargo;
  }
  else
  {
    graphPtr = &graphCov;
    cargoPtr = &edgeCargoCov;
  }


  const int lb = fragLenBounds.first;
  const int ub = fragLenBounds.second;

  if (ub > 0)
  {
    assert(nodeIds.count(cover.exons_1.back()) == 1);
    assert(nodeIds.count(cover.exons_2.front()) == 1);

    seqan::StringSet<seqan::String<TVertexDescriptor> > paths;
    int offset = allSegments[cover.exons_1.back()].length;
    TVertexDescriptor start = getNodeId(cover.exons_1.back());
    TVertexDescriptor end = getNodeId(cover.exons_2.front());

    if (start >= end)
    {
      std::cerr << "invalid call of getExonPaths with overlapping left and right anchor -- EXIT" << std::endl;
      exit(1);
    }

    if (exonAware)
    {
      //Check: can we build a path without intron between the anchors?
      UInt gap = allSegments[cover.exons_2.front()].beginPos - allSegments[cover.exons_1.back()].endPos;
      bool singleExPath = (gap <= ub && gap >= lb);

      for (TVertexDescriptor v = start; v+1 != end && singleExPath; ++v)
      {
        singleExPath = !seqan::property(*cargoPtr, seqan::findEdge(*graphPtr, v, v+1))._intronEdge;
      }

      //If yes, is it exon aware?
      if (singleExPath)
      {
        std::vector<UInt> tmpPath;
        TVertexDescriptor startOuter = getNodeId(cover.exons_1.front());
        TVertexDescriptor endOuter = getNodeId(cover.exons_2.back());

        while(startOuter <= endOuter)
        {
          tmpPath.push_back(getSegmentId(startOuter));
          ++startOuter;
        }

        if(isExonAware(tmpPath))
        {
          exonPaths.push_back(std::vector<UInt>(tmpPath.begin() + cover.exons_1.size(), tmpPath.end() - cover.exons_2.size()));

          if (stopAfterFirstPath || end - start == 1)
          {            
            return;
          }
        }
      }
      //Else: Use path search to fill the gap with paths containing at least one real intron edge


      //check possible entries for left anchor
      size_t blockEntryId = 0;
      for (size_t i = 0; i + 1 < cover.exons_1.size(); ++i)
      {
        if (allSegments[cover.exons_1[i]].endPos != allSegments[cover.exons_1[i+1]].beginPos)
          blockEntryId = i+1;
      }
      std::vector<TVertexDescriptor> blockEntries, InvalidEntriesR;

      if (blockEntryId == 0)
      {
        //check for possible block entries left of the left anchor
        TVertexDescriptor node = getNodeId(cover.exons_1.front());

        if (node == 1)
        {
          blockEntries.push_back(node);
          //std::cout << "entry " << node << std::endl;
        }
        else
        {
          size_t seg = getSegmentId(node), newSeg = -1;
          //std::map<UInt,Segment>::const_iterator seg, newSeg;
          //seg = allSegments.find(getSegmentId(node));
          while(node > 0)
          {
            if (node == 1u || (allSegments[seg].type != INNER && allSegments[seg].type != RIGHTEND))
              blockEntries.push_back(node);
            if (node == 1u)
              break;

            newSeg = getSegmentId(node-1);

            if (allSegments[newSeg].endPos != allSegments[seg].beginPos)
              break;

            seg = newSeg;
            --node;
          }          
        }
      }
      else
      {
        blockEntries.push_back(getNodeId(cover.exons_1[blockEntryId]));
      }

      //check invalid entries for right anchor

      TVertexDescriptor v = getNodeId(cover.exons_2.front());
      if (v != 1 && allSegments[getSegmentId(v-1)].endPos == allSegments[cover.exons_2.front()].beginPos)
      {
        TVertexDescriptor blockExitId = -1u;
        TVertexDescriptor last = getNodeId(cover.exons_2.back());
        for (size_t i = cover.exons_2.size() - 1; i > 0; --i)
        {
          if (allSegments[cover.exons_2[i-1]].endPos != allSegments[cover.exons_2[i]].beginPos)
            blockExitId = getNodeId(cover.exons_2[i-1]);
        }

        size_t seg = getSegmentId(v), newSeg = -1;
        while(v > 0)
        {
          if (blockExitId != -1u && (_validBlockExit[v].empty() || std::find(_validBlockExit[v].begin(), _validBlockExit[v].end(), blockExitId) == _validBlockExit[v].end()))
          {
            InvalidEntriesR.push_back(v);
            //std::cout << "Invalid entry1: " << v << std::endl;
          }
          else if (blockExitId == -1u  && (_validBlockExit[v].empty() || _validBlockExit[v].back() < last)) //last is the maximal exit id
          {
            InvalidEntriesR.push_back(v);
            //std::cout << "Invalid entry2: " << v << std::endl;
          }
          if (v == 1)
            break;

          newSeg = getSegmentId(v-1);

          if (allSegments[newSeg].endPos != allSegments[seg].beginPos)
            break;

          seg = newSeg;
          --v;
        }
      }

      seqan::String<bool> validEntries;
      seqan::resizeVertexMap(validEntries, graph, true); //can be used for both graphs as they have same vertex ids
      for (size_t j = 0; j < InvalidEntriesR.size(); ++j)
      {
        seqan::property(validEntries, InvalidEntriesR[j]) = false;
      }
      unsigned maxPathLen = singleExPath ? end-start : -1u;
      for (size_t i = 0; i < blockEntries.size(); ++i)
      {
        TVertexDescriptor blockEntry = blockEntries[i];

        //std::cerr << "start at " << getSegmentId(blockEntry) << std::endl;
        if (stopAfterFirstPath)
        {
          seqan::String<TVertexDescriptor> path;
          detectFeasiblePath(*graphPtr, start, end, blockEntry, *cargoPtr, _validBlockExit, validEntries, maxUncovered, std::make_pair(std::max(lb,0)+offset, ub+offset), std::make_pair(0u, maxPathLen), maxSkip, path);
          if (!empty(path))
          {            
            seqan::appendValue(paths, path);
            break;
          }
        }
        else
          detectFeasiblePath(*graphPtr, start, end, blockEntry, *cargoPtr, _validBlockExit, validEntries, maxUncovered, std::make_pair(std::max(lb,0)+offset, ub+offset), std::make_pair(0u, maxPathLen), maxSkip, paths);
      }
    }

    else //not exon aware
    {
      if (stopAfterFirstPath)
      {
        seqan::String<TVertexDescriptor> path;
        detectFeasiblePath(*graphPtr, start, end, *cargoPtr, maxUncovered, std::make_pair(std::max(lb,0)+offset, ub+offset), std::make_pair(0u, -1u), maxSkip, path);
        if (!empty(path))
        {
          seqan::appendValue(paths, path);
        }
      }
      else
        detectFeasiblePath(*graphPtr, start, end, *cargoPtr, maxUncovered, std::make_pair(std::max(lb,0)+offset, ub+offset), std::make_pair(0u, -1u), maxSkip, paths);
    }

    //translate into exonIds
    size_t pOffset = exonPaths.size();
    exonPaths.resize(pOffset + seqan::length(paths));
    for (size_t i = 0; i < seqan::length(paths); ++i, ++pOffset)
    {
      for (size_t j = 1; j + 1 < seqan::length(paths[i]); ++j)
      {
        exonPaths[pOffset].push_back(getSegmentId(paths[i][j]));
      }
    }
    //remove duplicates
    std::sort(exonPaths.begin(), exonPaths.end());
    exonPaths.resize(std::distance(exonPaths.begin(), std::unique(exonPaths.begin(), exonPaths.end())));
  }
}



void SpliceGraph::getExonPaths(const TAllSegmentsMap& allSegments,
                        const SegmentCover& cover,
                        const std::pair<unsigned, unsigned>& fragLenBounds,
                        UInt maxUncovered,
                        UInt maxSkip,
                        bool exonAware,
                        std::vector<std::vector<UInt> >& exonPaths) const
{

  if (exonAware && !_exonAware)
  {
    std::cerr << "calling getExonPaths as exon aware without aware initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  if (maxUncovered != -1u && !_coverBased)
  {
    std::cerr << "calling getExonPaths as with maxUncovere != -1u without cover based initialization of SpliceGraph -- Terminating" << std::endl;
    exit(1);
  }

  const int rl = Parameters::read_length_global;
  const int isLb = fragLenBounds.first - 2 * rl;
  const int isUb = fragLenBounds.second - 2 * rl;

  std::pair<int, int> rangeL, rangeR;
  Transcript tmp;
  tmp.nodes = cover.exons_1;
  computeSingleRange(cover.exons_1, tmp, allSegments, rl, rangeL);
  tmp.nodes = cover.exons_2;
  computeSingleRange(cover.exons_2, tmp, allSegments, rl, rangeR);

  //get length of left cover
  size_t lenL(0);
  for (size_t i = 0; i < cover.exons_1.size(); ++i)
    lenL += allSegments[cover.exons_1[i]].length;

  int lb = isLb - (lenL - rangeL.first - rl) - rangeR.second;
  int ub = isUb - (lenL - rangeL.second - rl) - rangeR.first;

  if (ub > 0)
    _getExonPaths(allSegments, cover, std::make_pair(lb,ub), maxUncovered, maxSkip, exonAware, exonPaths);
}


bool SpliceGraph::isExonAware (const std::vector<UInt>& subexonChain) const
{
  //find real introns in chain
  if (subexonChain.empty())
    return true;

  size_t pos = 0;
  std::vector<size_t> exExits;
  exExits.reserve(subexonChain.size());

  size_t lastPos = subexonChain.size() - 1;
  while (pos < lastPos)
  {
    TEdgeDescriptor e = findEdge(subexonChain[pos], subexonChain[pos+1], true);
    if (!e)
    {
      std::cerr << "invalid asking for edge " << subexonChain[pos] << " => " << subexonChain[pos+1] << std::endl;
      return false;
    }

    if (seqan::property(edgeCargo, e)._intronEdge)
      exExits.push_back(pos);
    ++pos;
  }
  exExits.push_back(lastPos);

  TVertexDescriptor v = getNodeId(subexonChain[0]);
  //if (exExits[0] != 0)
  {
    //check to the left for some start with legal exit = exExits[0]
    TVertexDescriptor v2 = getNodeId(subexonChain[exExits[0]]);
    while(v > 0)
    {

      if (exExits[0] != lastPos && std::find(_validBlockExit[v].begin(), _validBlockExit[v].end(), v2) != _validBlockExit[v].end())
        break;
      else if (exExits[0] == lastPos && !_validBlockExit[v].empty() && _validBlockExit[v].back() >= v2)
        break;
      else if (v > 1 && !seqan::property(edgeCargo, seqan::findEdge(graph, v-1, v))._intronEdge)
        --v;
      else
      {
        //std::cerr << "A " << subexonChain[exExits[0]] << std::endl;
        return false;
      }
    }
  }

  TVertexDescriptor v1, v2;
  //v1 = nodeIds[subexonChain[0]];
  v1 = v;
  for(size_t i = 0; i < exExits.size(); ++i)
  {    
    v2 = getNodeId(subexonChain[exExits[i]]);
    //std::cerr << v1  << " :: " << v2 << std::endl;

    if (exExits[i] != lastPos && std::find(_validBlockExit[v1].begin(), _validBlockExit[v1].end(), v2) == _validBlockExit[v1].end())
    {
      //std::cerr << "B " << v1  << " - " << v2 << std::endl;
      return false;
    }
    else if (exExits[i] == lastPos)
    {
      if (_validBlockExit[v1].empty() || _validBlockExit[v1].back() < v2)
      {
        //std::cerr << "C" << std::endl;
        return false;
      }
      else
        return true;
    }

    v1 = getNodeId(subexonChain[exExits[i]+1]);
  }

  return true;
}


void pwLinApprox::set_d(double d)
{
	this->d = d;
	delta = sqrt(d*d*Parameters::epsilon); //not sqrt(2*d*Parameters::epsilon) ???
	//std::cout << "delta = " << delta << std::endl;
	nof_bp_in = (floor(c*d/delta) * 2) + 1;
	nof_bp_out = 5; //add 5 bp to each side, 3d, 4d, 5d
	//std::cout << "nof_bp_in = " << nof_bp_in << std::endl;
	//std::cout << "nof_bp_out = " << nof_bp_out << std::endl;
}

double pwLinApprox::f(size_t i)
{
	assert( i < nof_bp_in +  (2*nof_bp_out) );
	if(i<nof_bp_in) {
		//std::cout << "< nof_bp_in ";
		if(i*delta > c*d) {
			//std::cout << "and > d" << std::endl;
			int adj_i = ((int)i)-((nof_bp_in-1)/2);
			return adj_i * delta * adj_i * delta / d;
		} else
			//std::cout << "and <= d" << std::endl;
			return i*delta*i*delta/d;
	} else
	{
		//std::cout << ">= nof_bp_in ";
		if( ((int)i)-nof_bp_in < nof_bp_out ) {
			//std::cout << "and i-nof_bp_in < nof_bp_out" << std::endl;
			return (((int)i)+1-nof_bp_in+1+(c-1)) * d * (((int)i)+1-nof_bp_in+1+(c-1));
		}
		else {
			//std::cout << "and i-nof_bp_in >= nof_bp_out" << std::endl;
			return (((int)i)+1-nof_bp_in-nof_bp_out+1+(c-1)) * d * (((int)i)+1-nof_bp_in-nof_bp_out+1+(c-1));
		}
	}
}

double pwLinApprox::df(size_t i)
{
	assert( i < nof_bp_in + (2*nof_bp_out) );
	return 2*bp(i)/d;
}

double pwLinApprox::bp(size_t i)
{
	assert( i < nof_bp_in + (2*nof_bp_out) );
	if(i<nof_bp_in) {
		if(i*delta > c*d) {
			return -(((int)i)-((nof_bp_in-1)/2))*delta;
		}
		else
			return i*delta;
	} else
	{
		if( ((int)i)-nof_bp_in < nof_bp_out )
			return (((int)i)+1-nof_bp_in+1+(c-1)) * d;
		else {
			//std::cout << i << " " << nof_bp_in << " " << nof_bp_out << " " << d << std::endl;
			return -(((int)i)+1-nof_bp_in-nof_bp_out+1+(c-1)) * d;
		}
	}
}

int pwLinApprox::getNofBp()
{
	return nof_bp_in + (2 * nof_bp_out);
}


pwQuadCoeff::pwQuadCoeff(double bp, double left_lin, double left_quad, double right_lin, double right_quad) :
				bp(bp), left_lin(left_lin), left_quad(left_quad), right_lin(right_lin), right_quad(right_quad) {}

bool pwQuadCoeff::operator<(pwQuadCoeff const& rhs) const {
	return bp < rhs.bp;
}

bool pwQuadCoeff::operator<(double rhs) const {
	return bp < rhs;
}

//bool pwQuadCoeff::operator=(double rhs)	const {
//	return bps == rhs;
//}

void pwQuadApprox::set_d(double d)
{
	this->d = d;
	delta = sqrt(d*d*Parameters::epsilon); //TODO adjust to new loss function
	//std::cout << "delta = " << delta << std::endl;
	nof_bp_in = (floor(c*d/delta) * 2) + 1;
	nof_bp_out = 5;
	//std::cout << "nof_bp_in = " << nof_bp_in << std::endl;
	//std::cout << "nof_bp_out = " << nof_bp_out << std::endl;

	//linear interpolation between enclosing neighbors
	std::vector<pwQuadCoeff>::const_iterator low;
	low = std::lower_bound(coeff_bp.begin(), coeff_bp.end(), d);
	if(low == coeff_bp.end())
	{
		//std::cout << "low = end" << std::endl;
		l_lin_coeff = coeff_bp.back().left_lin;
		l_quad_coeff = coeff_bp.back().left_quad;
		r_lin_coeff = coeff_bp.back().right_lin;
		r_quad_coeff = coeff_bp.back().right_quad;
	} else if(low == coeff_bp.begin())
	{
		//std::cout << "low = begin" << std::endl;
		l_lin_coeff = coeff_bp.front().left_lin;
		l_quad_coeff = coeff_bp.front().left_quad;
		r_lin_coeff = coeff_bp.front().right_lin;
		r_quad_coeff = coeff_bp.front().right_quad;
	} else
	{
		double r_neighbor = low->bp;
		double l_neighbor = (low-1)->bp;
		assert(d<=r_neighbor);
		assert(d>=l_neighbor);
		double frac_loc = (d-l_neighbor)/(r_neighbor-l_neighbor);
		//std::cout << "neighbors: " << l_neighbor << " " << r_neighbor << " " << frac_loc << std::endl;

		double smaller;
		double adj_frac_loc;
		if((low-1)->left_lin <= low->left_lin)
		{
			smaller = (low-1)->left_lin;
			adj_frac_loc = frac_loc;
		} else
		{
			smaller = low->left_lin;
			adj_frac_loc = 1.-frac_loc;
		}
		l_lin_coeff = smaller + ( adj_frac_loc * fabs(low->left_lin - (low-1)->left_lin) );

		if((low-1)->left_quad <= low->left_quad)
		{
			smaller = (low-1)->left_quad;
			adj_frac_loc = frac_loc;
		} else
		{
			smaller = low->left_quad;
			adj_frac_loc = 1.-frac_loc;
		}
		l_quad_coeff = smaller + ( adj_frac_loc * fabs(low->left_quad - (low-1)->left_quad) );

		if((low-1)->right_lin <= low->right_lin)
		{
			smaller = (low-1)->right_lin;
			adj_frac_loc = frac_loc;
		} else
		{
			smaller = low->right_lin;
			adj_frac_loc = 1.-frac_loc;
		}
		r_lin_coeff = smaller + ( adj_frac_loc * fabs(low->right_lin - (low-1)->right_lin) );

		if((low-1)->right_quad <= low->right_quad)
		{
			smaller = (low-1)->right_quad;
			adj_frac_loc = frac_loc;
		} else
		{
			smaller = low->right_quad;
			adj_frac_loc = 1.-frac_loc;
		}
		r_quad_coeff = smaller + ( adj_frac_loc * fabs(low->right_quad - (low-1)->right_quad) );

		//std::cout << (low-1)->left_lin << " " << low->left_lin << " " << l_lin_coeff << std::endl;
		//std::cout << (low-1)->left_quad << " " << low->left_quad << " " << l_quad_coeff << std::endl;
		//std::cout << (low-1)->right_lin << " " << low->right_lin << " " << r_lin_coeff << std::endl;
		//std::cout << (low-1)->right_quad << " " << low->right_quad << " " <<  r_quad_coeff << std::endl;

		//std::cout << l_lin_coeff << " " << l_quad_coeff << " " << r_lin_coeff << " " << r_quad_coeff << std::endl;
	}
}

pwQuadApprox::pwQuadApprox(const std::vector<pwQuadCoeff>& coeff_bp)
	: coeff_bp(coeff_bp)
{}

double pwQuadApprox::f(size_t i)
{
	assert( i < nof_bp_in + (2*nof_bp_out) );

	if(i<nof_bp_in) //within d (pos or neg)
	{
		if(i*delta > c*d) //e negative, thus Vexp < V (left case)
		{
			int adj_i = ((int)i)-((nof_bp_in-1)/2);
			return (l_quad_coeff * (adj_i*delta) * (adj_i*delta)) + (l_lin_coeff * fabs(adj_i*delta));
		} else { //e positive, thus Vexp >= V (right case)
			return (r_quad_coeff * (i*delta) * (i*delta)) + (r_lin_coeff * fabs(i*delta));
		}
	} else
	{
		if( ((int)i)-nof_bp_in < nof_bp_out ) //e positive, right case
		{
			int adj_i = ((int)i)+1-nof_bp_in+1+(c-1);
			return (r_quad_coeff * (adj_i*d) * (adj_i*d)) + (r_lin_coeff * fabs(adj_i*d));
		} else { //e negative, left case
			int adj_i = ((int)i)+1-nof_bp_in-nof_bp_out+1+(c-1);
			return (l_quad_coeff * (adj_i*d) * (adj_i*d)) + (l_lin_coeff * fabs(adj_i*d));
		}
	}
}

double pwQuadApprox::bp(size_t i)
{
	assert(i < nof_bp_in + (2*nof_bp_out));
	if(i<nof_bp_in) {
		if(i*delta > c*d)
			return -(((int)i)-((nof_bp_in-1)/2))*delta;
		else
			return i*delta;
	} else
	{
		if( ((int)i)-nof_bp_in < nof_bp_out )
			return (((int)i)+1-nof_bp_in+1+(c-1)) * d;
		else
			return -(((int)i)+1-nof_bp_in-nof_bp_out+1+(c-1)) * d;
	}
}

double pwQuadApprox::df(size_t i)
{
	assert(i < nof_bp_in + (2*nof_bp_out));
	if( (i<nof_bp_in && i*delta > c*d) || (i>=nof_bp_in && ((int)i)-nof_bp_in >= nof_bp_out) ) { //e negative
		return (2*l_quad_coeff*bp(i)) - l_lin_coeff;
	}
	else { //e positive
		return (2*r_quad_coeff*bp(i)) + r_lin_coeff;
	}
}

int pwQuadApprox::getNofBp()
{
	return nof_bp_in + (2*nof_bp_out);
}
