/* This program is free software: you can redistribute it and/or modify
 *
 * Copyright (C) 2014 Stefan Canzar, Sandro Andreotti
 *
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <priceilp.h>

#include <Parameters.h>
#include <Functions.h>

#include <gsl/gsl_cdf.h>

using std::cout;
using std::endl;

PriceIlp::PriceIlp(const Data& ilpData, const Options & options)
	: _ilpData(ilpData)
	, _options(options)
	, _env()
 	, _model(_env)
	, _x(_env, ilpData.locus.size())
	, _y(_env, ilpData.covers.size())
	, _y2exons(ilpData.covers.size())
	, _cover_valid(ilpData.covers.size(), true)
	, _es(_env, ilpData.locus.size())
	, _ee(_env, ilpData.locus.size())
    , _s(_env, ilpData.locus.size())
	, _e(_env, ilpData.locus.size())
	, _pre(_env, ilpData.locus.size()-1)
	, _suff(_env, ilpData.locus.size()-1)  //attention: suff[i] refers to variable x[i+1]
	, _ir(_env, ilpData.intron_retention.size())
{
	for (size_t i = 0; i < ilpData.locus.size(); i++)
	{
		_exon_order[ilpData.locus[i]] = i;
	}

	//because ir vars used in exon awareness
	if(!_options._ignoreIntronRet && !_ilpData.intron_retention.empty()) {
		//one variable for each intron retentions
		for(size_t i = 0; i<ilpData.intron_retention.size(); ++i)
    {
      std::stringstream ss;
      ss << i;
			std::string varname = std::string("ir_") + ss.str();
			_ir[i] = IloBoolVar(_env, varname.c_str());
		}
	}
}

PriceIlp::~PriceIlp()
{
	_x.end();
	_y.end();
	_es.end();
	_ee.end();
	_s.end();
	_e.end();
	_pre.end();
	_suff.end();
	_ir.end();
}

bool PriceIlp::init()
{
	bool successful;
	initXY();
	if(_options._exonAware)
		initExonAwareNEW();
	if(!_options._ignoreIntronRet && !_ilpData.intron_retention.empty())
		initIntronRet();
	if(_options._tsstes)
		initTssTesNEW();
	if(_options._sgComp)
		initSgComp(); //allows to relax at most Parameters::max_nof_relaxed_edges edge(s) to explain new covers
	if(Parameters::explainNewCover)
		successful = explainNewCover();
	initObjective();
	initCplex();
	//_cplex.exportModel("test.lp");

	return successful;
}

void PriceIlp::initXY()
{
	for(long long int i=0; i<_ilpData.locus.size(); i++)
	{
		std::stringstream ss;
		ss << i;
		std::string varname = std::string("x_") + ss.str();
		_x[i] = IloBoolVar(_env, varname.c_str());
	}

	//_model.add(_x[_exon_order[239436]]==1);
	//_model.add(_x[_exon_order[239437]]==0);
	//_model.add(_x[_exon_order[239438]]==1);
	//_model.add(_x[_exon_order[239439]]==1);
	//_model.add(_x[_exon_order[15595]]==1);
	//_model.add(_x[_exon_order[15596]]==0);
	//_model.add(_x[_exon_order[15597]]==1);
	//_model.add(_x[_exon_order[15598]]==1);
	//_model.add(_x[_exon_order[15599]]==1);
	//_model.add(_x[_exon_order[15600]]==0);
	//_model.add(_x[_exon_order[15601]]==0);
	//_model.add(_x[_exon_order[15602]]==0);
	//_model.add(_x[_exon_order[15603]]==0);
	//_model.add(_x[_exon_order[15604]]==1);
/*	_model.add(_x[_exon_order[773]]==1);
	_model.add(_x[_exon_order[774]]==0);
	_model.add(_x[_exon_order[775]]==1);
	_model.add(_x[_exon_order[776]]==1);
	_model.add(_x[_exon_order[777]]==1);
	_model.add(_x[_exon_order[778]]==1);
	_model.add(_x[_exon_order[779]]==1);
	_model.add(_x[_exon_order[780]]==1);
	_model.add(_x[_exon_order[781]]==0);
	_model.add(_x[_exon_order[782]]==1);
	_model.add(_x[_exon_order[783]]==1);
	_model.add(_x[_exon_order[784]]==1);
	_model.add(_x[_exon_order[785]]==0);
	_model.add(_x[_exon_order[786]]==1);
	_model.add(_x[_exon_order[787]]==1);
	_model.add(_x[_exon_order[788]]==0);
	_model.add(_x[_exon_order[789]]==0);
	_model.add(_x[_exon_order[790]]==1);
	_model.add(_x[_exon_order[791]]==1);
	_model.add(_x[_exon_order[792]]==1);
	_model.add(_x[_exon_order[793]]==1);
	_model.add(_x[_exon_order[794]]==1);
	_model.add(_x[_exon_order[797]]==0);
*/


/*	_model.add(_x[_exon_order[27973]]==0);
	_model.add(_x[_exon_order[27974]]==0);
	_model.add(_x[_exon_order[27975]]==0);
	_model.add(_x[_exon_order[27976]]==0);
	_model.add(_x[_exon_order[27977]]==0);
	_model.add(_x[_exon_order[27978]]==0);
	_model.add(_x[_exon_order[27979]]==1);
	_model.add(_x[_exon_order[27980]]==1);
	_model.add(_x[_exon_order[27981]]==1);
	_model.add(_x[_exon_order[27982]]==1);
	_model.add(_x[_exon_order[27983]]==1);
	_model.add(_x[_exon_order[27984]]==1);
	_model.add(_x[_exon_order[27985]]==0);
	_model.add(_x[_exon_order[27986]]==1);
	_model.add(_x[_exon_order[27987]]==1);
	_model.add(_x[_exon_order[27988]]==1);
	_model.add(_x[_exon_order[27989]]==0);
	_model.add(_x[_exon_order[27990]]==1);
	_model.add(_x[_exon_order[27991]]==0);
	_model.add(_x[_exon_order[27992]]==1);
	_model.add(_x[_exon_order[27993]]==1);
	_model.add(_x[_exon_order[27994]]==0);
	_model.add(_x[_exon_order[27995]]==0);
	_model.add(_x[_exon_order[27996]]==0);
	_model.add(_x[_exon_order[27997]]==0);
	_model.add(_x[_exon_order[27998]]==1);
	_model.add(_x[_exon_order[27999]]==1);
*/

	/*
			//obey subexon structure
			//----------------------
			assert(segments.find(locus[0])->second.type != RIGHTEND && segments.find(locus[0])->second.type != INNER);
			for(UInt i=1; i<locus.size(); i++)
				if(segments.find(locus[i])->second.type == RIGHTEND || segments.find(locus[i])->second.type == INNER) {
					model_prcg.add(x[i]<=x[i-1]);
				}
			assert(segments.find(locus[locus.size()-1])->second.type != LEFTEND && segments.find(locus[locus.size()-1])->second.type != INNER);
			for(UInt i=0; i<locus.size()-1; i++)
				if(segments.find(locus[i])->second.type == LEFTEND || segments.find(locus[i])->second.type == INNER) {
					model_prcg.add(x[i]<=x[i+1]);
				}
	 */

	IloExpr y_lb(_env);

	for(int i=0; i<_ilpData.covers.size(); i++)
	{
/*
		std::cout << "cover " << i << std::endl;

			    	cout << "left mate:" << endl;
			    				for(int t=0; t<_ilpData.covers[i].exons_1.size(); ++t)
			    					cout << _ilpData.covers[i].exons_1[t] << " ";
			    				cout << endl;
			    				cout << "right mate:" << endl;
			    				for(int t=0; t<_ilpData.covers[i].exons_2.size(); ++t)
			    					cout << _ilpData.covers[i].exons_2[t] << " ";
			    				cout << endl;
			    				cout << "intensity: " << _ilpData.covers[i].intensity << endl;
*/

		//bool valid = enumGapExonSets(covers[i], segments, locus, y2exons[i], legalBlockBounds);
		//bool valid = enumGapExonSets(covers[i], segments, locus, y2exons[i]);
		//      for(int zz=0; zz<y2exons[i].size(); zz++) {
		//    	  cout << "current sol: " << endl;
		//    	  for(vector<bool>::const_iterator cs_it = y2exons[i][zz].begin(); cs_it!=y2exons[i][zz].end(); cs_it++)
		//    		  cout << *cs_it << " ";
		//    	  cout << endl;
		//      }

		//cout << "number of possibilities " << y2exons[i].size() << endl;

		std::pair<unsigned, unsigned> bounds;
		bounds.first = std::max((2 * Parameters::read_length_global), (int)floor(gsl_cdf_gaussian_Pinv(Parameters::lower_percentile, Parameters::s_dev)) + Parameters::insert_size_global);
		bounds.second = ceil(gsl_cdf_gaussian_Pinv(Parameters::upper_percentile, Parameters::s_dev)) + Parameters::insert_size_global;
		std::vector<std::vector<UInt> > exon_paths;
		if(_ilpData.covers[i].exons_1.back() < _ilpData.covers[i].exons_2.front()) {
			//bounds.first = 0;
			//bounds.second = 10000;
			if(Parameters::explainNewCover)
			{
				if(i<_ilpData.cov_expl.size() && !_ilpData.cov_expl[i])
					_ilpData.sg.getExonPaths(_ilpData.segments, _ilpData.covers[i], bounds, _options._maxUncovered, _options._maxSkip, _options._exonAware, exon_paths);
				else
					_ilpData.sg.getExonPaths(_ilpData.segments, _ilpData.covers[i], bounds, 0, _options._maxSkip, _options._exonAware, exon_paths);
			} else
				_ilpData.sg.getExonPaths(_ilpData.segments, _ilpData.covers[i], bounds, _options._maxUncovered, _options._maxSkip, _options._exonAware, exon_paths);
			//_ilpData.sg.getExonPaths(_ilpData.segments, _ilpData.covers[i], bounds, 1, _options._maxSkip, _options._exonAware, exon_paths);
		}

		//convert exon path to y2exons
		vector<vector<bool> >& valid_candidates = _y2exons[i];
		vector<UInt>::const_iterator low = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_1.back());
		std::vector<std::vector<UInt> >::const_iterator exon_path_it = exon_paths.begin();
		for(; exon_path_it != exon_paths.end(); exon_path_it++)
		{
			//std::cout << "next path" << std::endl;
			//assert((*exon_path_it).size() > 0);
			if((*exon_path_it).size() > 0) {
				UInt next_idx = 0;
				vector<bool> current_sol;
				for(vector<UInt>::const_iterator it = low+1; *it<_ilpData.covers[i].exons_2.front(); ++it)
				{
					if(next_idx < (*exon_path_it).size() && (*exon_path_it)[next_idx] == *it) {
						current_sol.push_back(true);
						//std::cout << " 1 ";
						next_idx++;
					}
					else {
						current_sol.push_back(false);
						//std::cout << " 0 ";
					}
				}
				//std::cout << std::endl;
				valid_candidates.push_back(current_sol);
				//    	  cout << "current sol: " << endl;
				//    	  for(vector<bool>::const_iterator cs_it = current_sol.begin(); cs_it!=current_sol.end(); cs_it++)
				//    		  cout << *cs_it << " ";
				//    	  cout << endl;
			} else
			{
				//cout << "oh gott" << endl;
				vector<bool> empty_path;
				for(vector<UInt>::const_iterator it = low+1; *it<_ilpData.covers[i].exons_2.front(); ++it)
				{
					empty_path.push_back(false);
				}
				//if(empty_path.size()>0)
				valid_candidates.push_back(empty_path);
			}
		}
		//note: valid_candidates = y2exons[i] should be empty only if middle part of fragment non-empty but infeasible
		if(_ilpData.covers[i].exons_1.back() >= _ilpData.covers[i].exons_2.front()) {
			vector<bool> empty_path;
			valid_candidates.push_back(empty_path);
		}

		//cout << "number of possibilities " << y2exons[i].size() << endl;

		bool valid = true; //??? Gegenbeispiele???

		if(!valid) {
			_cover_valid[i] = false;
			continue;
		}

		//y[i] = IloBoolVarArray(env_prcg, max(1,(int)y2exons[i].size()));
		_y[i] = IloBoolVarArray(_env, (int)_y2exons[i].size());  //remember: y2exons[i] = valid_candidates

		//at most one variable from each y family
		IloExpr sumofy(_env);
		//for(int j=0; j<max(1,(int)y2exons[i].size()); j++) {
		for(int j=0; j<(int)_y2exons[i].size(); j++)
		{
			std::stringstream ssi, ssj;
			ssi << i;
			ssj << j;
			std::string varname = std::string("y_") + ssi.str() + std::string("_") + ssj.str();
			_y[i][j] = IloBoolVar(_env, varname.c_str());
			sumofy += _y[i][j];
		}
		_model.add(sumofy<=1);
		sumofy.end();

		//connect y to x
		//vector<UInt>::const_iterator low = lower_bound(locus.begin(), locus.end(), _ilpData.covers[i].exons_1.back());
		//for(int j=0; j<max(1,(int)y2exons[i].size()); j++) {
		for(int j=0; j<(int)_y2exons[i].size(); j++) {
			int cnt_overlap = 0;
			int tot_span = 0;
			y_lb.clear();
			std::vector<UInt>::const_iterator ex_it;
			vector<UInt>::const_iterator locus_it = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_1.front());
			for(;locus_it!=_ilpData.locus.end() && *locus_it <= _ilpData.covers[i].exons_1.back(); locus_it++)
			{
				ex_it = find(_ilpData.covers[i].exons_1.begin(), _ilpData.covers[i].exons_1.end(), *locus_it);
				if(ex_it==_ilpData.covers[i].exons_1.end()) {
					//std::cout << "1-" << *locus_it << std::endl;
					y_lb += 1-_x[ _exon_order[*locus_it] ];
					_model.add(_y[i][j]<=1-_x[ _exon_order[*locus_it] ]);
					tot_span++;
				} else
				{
					y_lb += _x[ _exon_order[*ex_it] ];
					//std::cout << *ex_it << std::endl;
					_model.add(_y[i][j]<=_x[ _exon_order[*ex_it] ]);
					tot_span++;
				}
			}
			locus_it = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_2.front());
			for(;locus_it!=_ilpData.locus.end() && *locus_it <= _ilpData.covers[i].exons_2.back(); locus_it++)
			{
				if(*locus_it <= _ilpData.covers[i].exons_1.back()) {
					cnt_overlap++;
					continue;
				}
				ex_it = find(_ilpData.covers[i].exons_2.begin(), _ilpData.covers[i].exons_2.end(), *locus_it);
				if(ex_it==_ilpData.covers[i].exons_2.end()) {
					y_lb += 1-_x[ _exon_order[*locus_it] ];
					//std::cout << "1-" << *locus_it << std::endl;
					_model.add(_y[i][j]<=1-_x[ _exon_order[*locus_it] ]);
					tot_span++;
				} else
				{
					y_lb += _x[ _exon_order[*ex_it] ];
					//std::cout << *ex_it << std::endl;
					_model.add(_y[i][j]<=_x[ _exon_order[*ex_it] ]);
					tot_span++;
				}
			}
			//if(y2exons[i].size()>0) {
			for(int z=0; z<_y2exons[i][j].size(); z++) {
				if(_y2exons[i][j][z]) {
					y_lb += _x[ _exon_order[*(low+z+1)] ];
					//std::cout << *(low+z+1) << std::endl;
					_model.add(_y[i][j]<=_x[ _exon_order[*(low+z+1)] ]);
				}
				else {
					y_lb += 1 - _x[ _exon_order[*(low+z+1)] ];
					//std::cout << "1-" << *(low+z+1) << std::endl;
					_model.add(_y[i][j]<=1-_x[ _exon_order[*(low+z+1)] ]);
				}
			}
			//}
			//if(y2exons[i].size()>0) {
			y_lb += 1 - (int)_y2exons[i][j].size() - tot_span;
			//std::cout << " finally: " << 1 - (int)_y2exons[i][j].size() - tot_span << std::endl;
			//}
			//else {
			//y_lb += 1 - tot_span;
			//}
			_model.add(_y[i][j]>=y_lb);

		}
	}
	y_lb.end();
}

bool PriceIlp::updateNewCover()
{
	_model.remove(_force_expl_constr);
	return explainNewCover();
}

bool PriceIlp::explainNewCover()
{
	bool nonzero = false;
	IloExpr sumofy(_env);
	for(int i=0; i<_ilpData.cov_expl.size(); i++)
	{
		if(!_ilpData.cov_expl[i])
		{
			//cout << " cover " << i << endl;

			vector<UInt>::const_iterator low;
			if(_y2exons[i].size()>0)
				low = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_1.back());
			for(int j=0; j<(int)_y2exons[i].size(); j++)
			{
				//determine normfactor
				Transcript tmp;
				for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_1.begin(); ex_it != _ilpData.covers[i].exons_1.end(); ex_it++)
					tmp.nodes.push_back(*ex_it);
				if(_y2exons[i].size()>0)
				{
					for(int z=0; z<_y2exons[i][j].size(); z++)
						if(_y2exons[i][j][z])
							tmp.nodes.push_back(*(low+z+1));
				}
				for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_2.begin(); ex_it != _ilpData.covers[i].exons_2.end(); ex_it++)
				{
					if(*ex_it <= _ilpData.covers[i].exons_1.back())
						continue;
					tmp.nodes.push_back(*ex_it);
				}
				//std::cout << "tmp: " << std::endl;
				//for(int k=0; k<tmp.nodes.size(); k++)
					//std::cout << tmp.nodes[k] << " ";
				//std::cout << std::endl;
				double ll = getNormFactor(_ilpData.covers[i], tmp, _ilpData.segments);
				//std::cout << "normfactor: " << ll << std::endl;

				if(_ilpData.covers[i].intensity >= Parameters::unexpl_cov_bound)
				{
					nonzero = true;
					if(Parameters::unexpl_cov_sum_bound)
						sumofy += _ilpData.covers[i].intensity*_y[i][j];
					else
						sumofy += _y[i][j];
				}
			}
		}
	}
	if(nonzero) 
    {
        bool force_by_rel_edges = false;
        if (force_by_rel_edges)
        {
           for (size_t rel = 0; rel < _rx_sgc.size(); ++rel)
           {
              if(Parameters::unexpl_cov_sum_bound)
                  _force_expl_constr = _model.add(sumofy >= _rx_sgc[rel] * Parameters::unexpl_cov_sum_bound); //can cause infeasibility
              else
                  _force_expl_constr = _model.add(sumofy >= _rx_sgc[rel]);
           }
        }
        else
        {
            if(Parameters::unexpl_cov_sum_bound)
                _force_expl_constr = _model.add(sumofy>=Parameters::unexpl_cov_sum_bound); //can cause infeasibility
            else
                _force_expl_constr = _model.add(sumofy>=1);
        }
	}
	else
  {
    if (Parameters::verbose)
      std::cerr << "cannot explain new cover" << endl;
    
		return false;
	}
	sumofy.end();
	return true;
}

void PriceIlp::initExonAware()
{
	vector<bool> valid_exon_end(_ilpData.locus.size());
	for(int i=0; i<_ilpData.locus.size(); i++)
		valid_exon_end[i]=false;

	std::string varname;
	for(long long int i=0; i<_ilpData.locus.size(); i++)
	{
		std::stringstream ss;
		ss << i;
		varname = std::string("s_") + ss.str();
		_es[i] = IloBoolVar(_env, varname.c_str());
		varname = std::string("_e_") + ss.str();
		_ee[i] = IloBoolVar(_env, varname.c_str());
	}

	//compute for every cluster the set of intron retentions that start/end in that cluster
	//note touching neighboring exons must be ignored, since unaffected by exon start/end definition
	vector<vector<UInt> > starting_intrets(_ilpData.exon_cluster.size());
	vector<vector<UInt> > ending_intrets(_ilpData.exon_cluster.size());
	for(UInt i=0; i<_ilpData.intron_retention.size(); ++i) {
		UInt startloc = _ilpData.locus[_ilpData.intron_retention[i].first];
		//cout << "startloc = " << startloc << endl;
		if(_ilpData.locid2exidx.find(startloc) != _ilpData.locid2exidx.end()) {
			UInt start_cluster = _ilpData.locid2exidx.at(startloc);
			starting_intrets[start_cluster].push_back(i);
			//cout << "added start " << i << " to " << startloc << endl;
		}
		UInt endloc = _ilpData.intron_retention[i].second;
		//cout << "endloc = " << endloc << endl;
		if(_ilpData.locid2exidx.find(endloc) != _ilpData.locid2exidx.end()) {
			UInt end_cluster = _ilpData.locid2exidx.at(endloc);
			ending_intrets[end_cluster].push_back(i);
			//cout << "added end " << i << " to " << endloc << endl;
		}
	}


	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		//std::string varname = std::string("s_") + std::to_string(i);
		//_es[i] = IloBoolVar(_env, varname.c_str());
		//varname = std::string("_e_") + std::to_string(i);
		//_ee[i] = IloBoolVar(_env, varname.c_str());
		std::map<UInt, std::set<UInt> >::const_iterator entryIt = _ilpData.legalBlockBounds.find(_ilpData.locus[i]);
		if(entryIt == _ilpData.legalBlockBounds.end())
		{
			_model.add(_es[i]==0);
		} else
		{
			for(std::set<UInt>::const_iterator exitIt = entryIt->second.begin(); exitIt != entryIt->second.end(); ++exitIt)
				valid_exon_end[ _exon_order[*exitIt] ] = true;  //remember for later
		}

		//definition of es and ee
		//notice that all x between selected es and ee are one, otherwise additional es/ee necessary
		if(i>0)
			_model.add(_es[i]>=_x[i]+1-_x[i-1]-1);
		else
			_model.add(_es[i]>=_x[i]);

		if(i<_ilpData.locus.size()-1)
			_model.add(_ee[i]>=_x[i]+1-_x[i+1]-1);
		else
			_model.add(_ee[i]>=_x[i]);

		_model.add(_x[i]>=_es[i]);
		_model.add(_x[i]>=_ee[i]);

		if(i>0)
			_model.add(_es[i]<=_ee[i-1]+1-_x[i-1]);
		if(i<_ilpData.locus.size()-1)
			_model.add(_ee[i]<=_es[i+1]+1-_x[i+1]);
	}

	IloExpr sumof_ee(_env);
	IloExpr sumof_sir(_env);
	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		if(!valid_exon_end[i])
			_model.add(_ee[i]==0);

		if(_ilpData.locid2exidx.find(_ilpData.locus[i])==_ilpData.locid2exidx.end())
			continue;

		sumof_sir.clear();
		UInt exidx =_ilpData.locid2exidx.at(_ilpData.locus[i]);
		for(UInt j=0; j<starting_intrets[exidx].size(); j++) {
			if(_ilpData.intron_retention[starting_intrets[exidx][j]].first==i)
				sumof_sir+=_ir[starting_intrets[exidx][j]];
		}

		std::map<UInt, std::set<UInt> >::const_iterator entryIt = _ilpData.legalBlockBounds.find(_ilpData.locus[i]);
		if(entryIt != _ilpData.legalBlockBounds.end())
		{
			sumof_ee.clear();
			for(std::set<UInt>::const_iterator exitIt = entryIt->second.begin(); exitIt != entryIt->second.end(); ++exitIt)
			{
				int eo = _exon_order[*exitIt];
				if( _ilpData.locid2exidx.find(*exitIt) != _ilpData.locid2exidx.end() &&
				   _ilpData.locid2exidx.at(*exitIt) == _ilpData.locid2exidx.at(_ilpData.locus[i]) ) //have to be in same cluster (intron retention!)
					sumof_ee+=_ee[eo];
			}
			_model.add(sumof_ee + sumof_sir - _es[i] >= 0); //if decided to start exon here it has to match one of its legal ends
		}
	}
	sumof_sir.end();

	//in every exon cluster at most one es,ee pair
	IloExpr sumof_es(_env);
	//IloExpr sumof_es_intret(_env);
	//IloExpr sumof_ee_intret(_env);
	//IloExpr sumof_ee;
	for(UInt i=0; i<_ilpData.exon_cluster.size(); i++)
	{
		UInt leftmost = _ilpData.locus.size();
		UInt rightmost = _ilpData.locus.size();
		//compute left- and rightmost segment in cluster
		for(UInt j=0; j<_ilpData.exon_cluster[i].size(); j++)
		{
			if(leftmost == _ilpData.locus.size() || _ilpData.exon_cluster[i][j].first < leftmost)
				leftmost = _ilpData.exon_cluster[i][j].first;
			if(rightmost == _ilpData.locus.size() || _exon_order[_ilpData.exon_cluster[i][j].second] > rightmost)
				rightmost = _exon_order[_ilpData.exon_cluster[i][j].second];
		}
		//std::cout << "look: " << leftmost << " " << rightmost << std::endl;
		sumof_es.clear();
		sumof_ee.clear();
		for(UInt j=leftmost; j<rightmost+1; j++) {
			sumof_es+=_es[j];
			sumof_ee+=_ee[j];
		}
		_model.add(sumof_es<=1);
		_model.add(sumof_ee<=1);

		for(UInt j=0; j<starting_intrets[i].size(); j++) {
			sumof_es-=_ir[starting_intrets[i][j]];
			//cout << " subtracting start intret x from y: " << _ilpData.locus[leftmost] << " " << _ilpData.locus[rightmost] << ": " << starting_intrets[i][j] << endl;
		}
		for(UInt j=0; j<ending_intrets[i].size(); j++) {
			sumof_ee-=_ir[ending_intrets[i][j]];
			//cout << " subtracting end intret x from y: " << _ilpData.locus[leftmost] << " " << _ilpData.locus[rightmost] << ": " << ending_intrets[i][j] << endl;
		}

		_model.add(sumof_es==sumof_ee);
	}
	sumof_ee.end();
	sumof_es.end();
}

void PriceIlp::initExonAwareNEW()
{
  
  std::vector<std::pair<UInt, UInt> > _knownExonVarBounds;
  //get known exons from legalBlockBounds
  for(int i=0; i<_ilpData.locus.size(); i++)
  {
	std::map<UInt, std::set<UInt> >::const_iterator entryIt = _ilpData.legalBlockBounds.find(_ilpData.locus[i]);
    if (entryIt == _ilpData.legalBlockBounds.end())
        continue;

    std::set<UInt>::const_iterator legalEndIt;
    for (legalEndIt = entryIt->second.begin(); legalEndIt != entryIt->second.end(); ++legalEndIt)
    {
      if (Parameters::verbose)
        std::cout << "exon: " << _ilpData.locus[i] << " <-> " << *legalEndIt << std::endl;
      UInt rightId = _exon_order[*legalEndIt];
      std::stringstream ss;
      ss << i;
      std::string varname = std::string("exon_") + ss.str();
      _knownExonVars.push_back(IloBoolVar(_env, varname.c_str()));
      _knownExonVarBounds.push_back(std::make_pair(i,rightId));
    }
  }
  
  //bind segments to exon vars
  for(int i=0; i<_ilpData.locus.size(); i++)
  {
    IloExpr exAware(_env);
    for (size_t j = 0; j < _knownExonVarBounds.size(); ++j)
    {
      if (_knownExonVarBounds[j].first <= i && _knownExonVarBounds[j].second >= i)
        exAware += _knownExonVars[j];
    }
    _model.add(_x[i] == exAware);
    exAware.end();
  }
}


void PriceIlp::initIntronRet()
{
	//assuming that no duplicates exist
/*
	//one variable for each intron retentions
	for(size_t i = 0; i<_ilpData.intron_retention.size(); ++i) {
		std::string varname = std::string("ir_") + std::to_string(static_cast<long long unsigned>(i));
		_ir[i] = IloBoolVar(_env, varname.c_str());
	}
*/

	//forbid pairwise overlapping intron retentions?
	const map<UInt, UInt>& exidx = _ilpData.locid2exidx;
	const vector<UInt>& locus = _ilpData.locus;
	const TAllSegmentsMap& segments = _ilpData.segments;
	for(size_t i=0; i<_ilpData.intron_retention.size()-1; ++i)
		for(size_t j=i+1; j<_ilpData.intron_retention.size(); ++j) {
			UInt i_left_seg_idx = _ilpData.intron_retention[i].first;
			//while(i_left_seg_idx<locus.size() && exidx.find(locus[i_left_seg_idx++])==exidx.end());
			UInt i_right_seg_idx = _exon_order[_ilpData.intron_retention[i].second];
			//while(i_right_seg_idx>=0 && exidx.find(locus[i_right_seg_idx--])==exidx.end());
			UInt j_left_seg_idx = _ilpData.intron_retention[j].first;
			//while(j_left_seg_idx<locus.size() && exidx.find(locus[j_left_seg_idx++])==exidx.end());
			UInt j_right_seg_idx = _exon_order[_ilpData.intron_retention[j].second];
			//while(j_right_seg_idx>=0 && exidx.find(locus[j_right_seg_idx--])==exidx.end());

			/*
			assert(_ilpData.locid2exidx.find(_ilpData.locus[_ilpData.intron_retention[i].first])!=_ilpData.locid2exidx.end());
			UInt i_left_cluster_idx = _ilpData.locid2exidx.at(_ilpData.locus[_ilpData.intron_retention[i].first]);
			assert(_ilpData.locid2exidx.find(_ilpData.intron_retention[i].second)!=_ilpData.locid2exidx.end());
			UInt i_right_cluster_idx = _ilpData.locid2exidx.at(_ilpData.intron_retention[i].second);
			assert(_ilpData.locid2exidx.find(_ilpData.locus[_ilpData.intron_retention[j].first])!=_ilpData.locid2exidx.end());
			UInt j_left_cluster_idx = _ilpData.locid2exidx.at(_ilpData.locus[_ilpData.intron_retention[j].first]);
			assert(_ilpData.locid2exidx.find(_ilpData.intron_retention[j].second)!=_ilpData.locid2exidx.end());
			UInt j_right_cluster_idx = _ilpData.locid2exidx.at(_ilpData.intron_retention[j].second);
			*/
			if( !( i_left_seg_idx > j_right_seg_idx + 1 || j_left_seg_idx > i_right_seg_idx + 1 ||
				  (i_left_seg_idx == j_right_seg_idx + 1 && segments[locus[i_left_seg_idx]].beginPos != segments[locus[j_right_seg_idx]].endPos) ||
				  (j_left_seg_idx == i_right_seg_idx + 1 && segments[locus[j_left_seg_idx]].beginPos != segments[locus[i_right_seg_idx]].endPos) ) )
			//if( !( i_left_seg_idx > j_right_seg_idx || j_left_seg_idx > i_right_seg_idx ))
			{
				//std::cout << "conflicting intron retentions: " << std::endl;
				//std::cout << locus[i_left_seg_idx] << "-" << locus[i_right_seg_idx] << " " << locus[j_left_seg_idx] << "-" << locus[j_right_seg_idx] << std::endl;
				_model.add(_ir[i]+_ir[j]<=1);
			}
		}

	IloExpr sumof_es(_env);
	IloExpr sumof_ee(_env);
	//forbid exon clusters spanned by intron retentions
	for(size_t i=0; i<_ilpData.intron_retention.size(); ++i)
	{
		/*
		assert(_ilpData.locid2exidx.find(_ilpData.locus[_ilpData.intron_retention[i].first])!=_ilpData.locid2exidx.end());
		UInt left_cluster_idx = _ilpData.locid2exidx.at(_ilpData.locus[_ilpData.intron_retention[i].first]);
		assert(_ilpData.locid2exidx.find(_ilpData.intron_retention[i].second)!=_ilpData.locid2exidx.end());
		UInt right_cluster_idx = _ilpData.locid2exidx.at(_ilpData.intron_retention[i].second);
		*/

		//std::cerr << locus[_ilpData.intron_retention[i].first] << " " << _ilpData.intron_retention[i].second << std::endl;
		UInt left_seg_idx = _ilpData.intron_retention[i].first;
		int right_seg_idx = _exon_order[_ilpData.intron_retention[i].second];
		for(int j=left_seg_idx; j<=right_seg_idx; ++j)
			_model.add(_x[j]>=_ir[i]);
		if(left_seg_idx>0 && segments[locus[left_seg_idx-1]].endPos==segments[locus[left_seg_idx]].beginPos) //include touching neighbors
			left_seg_idx--;
		if(right_seg_idx<locus.size()-1 && segments[locus[right_seg_idx]].endPos==segments[locus[right_seg_idx+1]].beginPos) //include touching neighbors
			right_seg_idx++;
		while(left_seg_idx<locus.size() && exidx.find(locus[left_seg_idx])==exidx.end())
			left_seg_idx++;
		while(right_seg_idx>=0 && exidx.find(locus[right_seg_idx])==exidx.end())
			right_seg_idx--;
		assert(left_seg_idx<locus.size() && right_seg_idx>=0);
		assert(exidx.find(locus[left_seg_idx])!=exidx.end());
		assert(exidx.find(locus[right_seg_idx])!=exidx.end());

		//std::cout << "intron retention spans: " << std::endl;

		UInt left_cluster_idx = exidx.at(locus[left_seg_idx]);
		UInt right_cluster_idx = exidx.at(locus[right_seg_idx]);
		for(UInt cluster_idx = left_cluster_idx; cluster_idx <= right_cluster_idx; ++cluster_idx)
		{
			UInt leftmost = _ilpData.locus.size();
			UInt rightmost = _ilpData.locus.size();
			//compute left- and rightmost segment in cluster
			for(UInt j=0; j<_ilpData.exon_cluster[cluster_idx].size(); j++)
			{
				if(leftmost == _ilpData.locus.size() || _ilpData.exon_cluster[cluster_idx][j].first < leftmost)
					leftmost = _ilpData.exon_cluster[cluster_idx][j].first;
				if(rightmost == _ilpData.locus.size() || _exon_order[_ilpData.exon_cluster[cluster_idx][j].second]>rightmost)
					rightmost = _exon_order[_ilpData.exon_cluster[cluster_idx][j].second];
			}
			//std::cout << _ilpData.locus[leftmost] << " " << _ilpData.locus[rightmost] << std::endl;

			sumof_es.clear();
			sumof_ee.clear();
			for(UInt j=leftmost; j<rightmost+1; j++) {
				if(j!=_ilpData.intron_retention[i].first) {
					sumof_es+=_es[j];
				}
				if(j!=_exon_order[_ilpData.intron_retention[i].second]) {
					sumof_ee+=_ee[j];
				}
			}
			_model.add(sumof_es <= 1-_ir[i]);
			_model.add(sumof_ee <= 1-_ir[i]);
		}
	}
	sumof_es.end();
	sumof_ee.end();
}

//transcript must start at given TSS/TES
void PriceIlp::initTssTes()
{
	//variables
	//---------
	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		std::stringstream ss;
		ss << i;
		std::string varname = std::string("tss_") + ss.str();
		_s[i] = IloBoolVar(_env, varname.c_str());
		varname = std::string("tes_") + ss.str();
		_e[i] = IloBoolVar(_env, varname.c_str());
		if(i < _ilpData.locus.size()-1)
		{
			varname = std::string("pre_") + ss.str();
			_pre[i] = IloBoolVar(_env, varname.c_str());
			varname = std::string("suff_") + ss.str();
			_suff[i] = IloBoolVar(_env, varname.c_str());
		}
	}

	//constraints
	//-----------
	//precisely one TSS/TES
	IloExpr sumofv(_env);
	for(size_t i=0; i<_ilpData.locus.size(); i++)
		sumofv += _s[i];
	_model.add(sumofv==1);
	sumofv.clear();
	for(size_t i=0; i<_ilpData.locus.size(); i++)
		sumofv += _e[i];
	_model.add(sumofv==1);
	sumofv.end();

	//only given TSS/TES
	for(size_t i=0; i<_ilpData.locus.size(); i++)
	{
		if(!_ilpData.segments[_ilpData.locus[i]].tss || !_ilpData.sg.isSegmentCovered(_ilpData.locus[i]))
			_model.add(_s[i]==0);
		else if(_ilpData.segments[_ilpData.locus[i]].type != LEFTEND && _ilpData.segments[_ilpData.locus[i]].type != FULL) // ? why?
			_model.add(_s[i]==0);
		if(!_ilpData.segments[_ilpData.locus[i]].tes || !_ilpData.sg.isSegmentCovered(_ilpData.locus[i]))
			_model.add(_e[i]==0);
		else if(_ilpData.segments[_ilpData.locus[i]].type != RIGHTEND && _ilpData.segments[_ilpData.locus[i]].type != FULL) // ? why?
			_model.add(_e[i]==0);
	}

	//set TSS/TES
	for(size_t i=0; i<_ilpData.locus.size(); i++)
	{
		_model.add(_x[i]>=_s[i]);
		_model.add(_x[i]>=_e[i]);
	}

	//define prefix/suffix
	for(size_t i=0; i<_ilpData.locus.size()-1; i++)
	{
		_model.add(_pre[i]>=_s[i+1]);
		if(i<_ilpData.locus.size()-2)
			_model.add(_pre[i]>=_pre[i+1]);
		_model.add(_suff[i]>=_e[i]);
		if(i>0)
			_model.add(_suff[i]>=_suff[i-1]);
	}

	//forbid prefix/suffix
	for(size_t i=0; i<_ilpData.locus.size()-1; i++)
	{
		_model.add(_x[i]<=1-_pre[i]);
		_model.add(_x[i+1]<=1-_suff[i]);
	}

}

//transcript must start at given TSS/TES
void PriceIlp::initTssTesNEW()
{
	//variables
	//---------
	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		std::stringstream ss;
		ss << i;
		std::string varname = std::string("tss_") + ss.str();
		_s[i] = IloBoolVar(_env, varname.c_str());
		varname = std::string("tes_") + ss.str();
		_e[i] = IloBoolVar(_env, varname.c_str());
    //		if(i < _ilpData.locus.size()-1)
    //		{
    //			varname = std::string("pre_") + ss.str();
    //			_pre[i] = IloBoolVar(_env, varname.c_str());
    //			varname = std::string("suff_") + ss.str();
    //			_suff[i] = IloBoolVar(_env, varname.c_str());
    //		}
	}
  
	//constraints
	//-----------
	//precisely one TSS/TES
	IloExpr sumofv(_env);
	for(size_t i=0; i<_ilpData.locus.size(); i++)
		sumofv += _s[i];
	_model.add(sumofv==1);
	sumofv.clear();
	for(size_t i=0; i<_ilpData.locus.size(); i++)
		sumofv += _e[i];
	_model.add(sumofv==1);
	sumofv.end();
  
	//only given TSS/TES
	for(size_t i=0; i<_ilpData.locus.size(); i++)
	{
		if(!_ilpData.segments[_ilpData.locus[i]].tss || !_ilpData.sg.isSegmentCovered(_ilpData.locus[i]))
			_model.add(_s[i]==0);
		else if(_ilpData.segments[_ilpData.locus[i]].type != LEFTEND && _ilpData.segments[_ilpData.locus[i]].type != FULL) // ? why?
			_model.add(_s[i]==0);
		if(!_ilpData.segments[_ilpData.locus[i]].tes || !_ilpData.sg.isSegmentCovered(_ilpData.locus[i]))
			_model.add(_e[i]==0);
		else if(_ilpData.segments[_ilpData.locus[i]].type != RIGHTEND && _ilpData.segments[_ilpData.locus[i]].type != FULL) // ? why?
			_model.add(_e[i]==0);
	}
  
	//set TSS/TES
	for(size_t i=0; i<_ilpData.locus.size(); i++)
	{
		_model.add(_x[i]>=_s[i]);
		_model.add(_x[i]>=_e[i]);
	}
  
  for(size_t i = 0; i < _ilpData.locus.size(); i++)
  {
    IloExpr prefix(_env);
    IloExpr suffix(_env);
    
    //no segments selected to the right of TES
    for(size_t j = 0; j < i; j++)
    {
      prefix += _e[j];
    }
    
    //no segments selected to the left of TSS
    for(size_t j = i+1; j < _ilpData.locus.size(); j++)
    {
      suffix += _s[j];
    }
    
    _model.add(_x[i]<= 1 - prefix);
    _model.add(_x[i]<= 1 - suffix);
    
    prefix.end();
    suffix.end();
  }
}

void PriceIlp::initObjective()
{
	_obj = IloMinimize(_env);
	_model.add(_obj);
}

void PriceIlp::initCplex()
{
	_cplex = IloCplex(_model);
	_cplex.setParam(IloCplex::EpGap, 0.00001);

  if (Parameters::verbose)
  {
    std::cout << "new pricing ILP has " << _cplex.getNcols() << " columns and " << _cplex.getNrows() << " rows." << std::endl;
    std::cout << "number of nonzero elements: " << _cplex.getNNZs() << std::endl;
  }
  else
  {
    _cplex.setOut(_env.getNullStream());
  }
}

void PriceIlp::updateObjDual(const IloNumArray& duals)
{
	//adapt weight based on dual solution
	for(size_t i=0; i<_ilpData.covers.size(); i++)
	{
		if(!_cover_valid[i]) {
			continue;
		}

		vector<UInt>::const_iterator low;
		if(_y2exons[i].size()>0)
			low = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_1.back());
		//for(int j=0; j<max(1,(int)y2exons[i].size()); j++)
		for(int j=0; j<(int)_y2exons[i].size(); j++)
		{
			Transcript tmp;
			for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_1.begin(); ex_it != _ilpData.covers[i].exons_1.end(); ex_it++) {
				tmp.nodes.push_back(*ex_it);
			}
			if(_y2exons[i].size()>0)
			{
				for(int z=0; z<_y2exons[i][j].size(); z++)
				{
					if(_y2exons[i][j][z]) {
						tmp.nodes.push_back(*(low+z+1));
					}
				}
			}
			for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_2.begin(); ex_it != _ilpData.covers[i].exons_2.end(); ex_it++)
			{
				if(*ex_it <= _ilpData.covers[i].exons_1.back())
					continue;
				tmp.nodes.push_back(*ex_it);
			}
			double ll = getNormFactor(_ilpData.covers[i], tmp, _ilpData.segments);

			//cout << "come one " << i << "/" << j << ":  " << approxSolver.getDual(covered[i]) << " " << ll << endl;
			//for(int zz=0; zz<tmp.nodes.size(); zz++)
			//cout << tmp.nodes[zz] << " ";
			//cout << endl;

			_obj.setLinearCoef( _y[i][j], (-duals[i] * ll));
		}
	}
}

bool PriceIlp::solve()
{
	if( !_cplex.solve() )
  {
    if (Parameters::verbose)
      _env.error() << "Failed to optimize LP" << std::endl;
	
    return false;
	}
	assert(_cplex.getCplexStatus()==IloCplex::Optimal || _cplex.getCplexStatus()==IloCplex::OptimalTol);
	return true;
}

double PriceIlp::getObjValue() const
{
/*
				std::cout << "using following y vars:  " << std::endl;
	    		for(size_t i=0; i<_ilpData.covers.size(); i++)
	    		{
	    			//for(int j=0; j<max(1,(int)y2exons[i].size()); j++)
	    			for(int j=0; j<(int)_y2exons[i].size(); j++)
	    			{
	    				if((int)(_cplex.getValue(_y[i][j])+0.5) == 1)
	    					std::cout << i << "/" << j << ": " << std::endl;
	    			}
	    		}
*/

	/*
	std::cout << "es: " << std::endl;
	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		if((int)(_cplex.getValue(_es[i])+0.5)==1)
			std::cout << _ilpData.locus[i] << std::endl;
	}
	std::cout << "ee: " << std::endl;
	for(int i=0; i<_ilpData.locus.size(); i++)
	{
		if((int)(_cplex.getValue(_ee[i])+0.5)==1)
			std::cout << _ilpData.locus[i] << std::endl;
	}
	std::cout << "ir: " << std::endl;
	for(int i=0; i<_ilpData.intron_retention.size(); ++i)
	{
		if((int)(_cplex.getValue(_ir[i])+0.5)==1)
			std::cout << _ilpData.locus[i] << std::endl;
	}
	*/
	return _cplex.getObjValue();
}

void PriceIlp::printCplexStatus() const
{
	std::cout << _cplex.getCplexStatus() << std::endl;
}

bool PriceIlp::is_infeasible() const
{
	return _cplex.getCplexStatus()==IloCplex::Infeasible;
}

void PriceIlp::printX()
{
	for(size_t i=0; i<_ilpData.locus.size(); ++i)
	{
		if((int)(_cplex.getValue(_x[i])+0.5)==1)
			std::cout << _ilpData.locus[i] << " ";
	}
	std::cout << std::endl;
}

//derive from optimal ILP solution transcript and norm factors as new column to LP
void PriceIlp::getNewTrans(vector<bool>& new_trans, vector<double>& ll)
{
	for(size_t i=0; i<_ilpData.covers.size(); i++)
	{
		if(!_cover_valid[i]) {
			ll[i]=-1;
			continue;
		}
//		std::cout << "cover " << i << std::endl;
//		std::cout << " ===== " << std::endl;

/*
		    			std::cout << "left mate:" << std::endl;
		    			for(int t=0; t<_ilpData.covers[i].exons_1.size(); ++t)
		    				std::cout << _ilpData.covers[i].exons_1[t] << " ";
		    			std::cout << std::endl;
		    			std::cout << "right mate:" << std::endl;
		    			for(int t=0; t<_ilpData.covers[i].exons_2.size(); ++t)
		    				std::cout << _ilpData.covers[i].exons_2[t] << " ";
		    			std::cout << std::endl;
*/

		vector<UInt>::const_iterator low;
		if(_y2exons[i].size()>0)
			low = lower_bound(_ilpData.locus.begin(), _ilpData.locus.end(), _ilpData.covers[i].exons_1.back());
		for(int j=0; j<(int)_y2exons[i].size(); j++)
		{
			if((int)(_cplex.getValue(_y[i][j])+0.5)==1)
			{
				Transcript tmp;
				for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_1.begin(); ex_it != _ilpData.covers[i].exons_1.end(); ex_it++) {
					tmp.nodes.push_back(*ex_it);
					new_trans[_exon_order[*ex_it]] = true;
				}
				if(_y2exons[i].size()>0)
				{
					for(int z=0; z<_y2exons[i][j].size(); z++)
					{
						if(_y2exons[i][j][z]) {
							tmp.nodes.push_back(*(low+z+1));
							new_trans[_exon_order[*(low+z+1)]] = true;
						}
					}
				}
				for(vector<UInt>::const_iterator ex_it = _ilpData.covers[i].exons_2.begin(); ex_it != _ilpData.covers[i].exons_2.end(); ex_it++)
				{
					if(*ex_it <= _ilpData.covers[i].exons_1.back())
						continue;
					tmp.nodes.push_back(*ex_it);
					new_trans[_exon_order[*ex_it]] = true;
				}
				//std::cout << "tmp: " << std::endl;
				//for(int k=0; k<tmp.nodes.size(); k++)
					//std::cout << tmp.nodes[k] << " ";
				//std::cout << std::endl;
				ll[i] = getNormFactor(_ilpData.covers[i], tmp, _ilpData.segments);
				//std::cout << i << " " << ll[i] << std::endl;
				break;
			}
			//if(j == (int)_y2exons[i].size() - 1)
				//std::cout << i << " 0" << endl;
		}
	}
	if(_options._tsstes) {
		//fixing tss and tes, they might not be consistent with any cover
		for(UInt i=0; i<_ilpData.locus.size(); i++)
			if((int)(_cplex.getValue(_s[i])+0.5)==1 || (int)(_cplex.getValue(_e[i])+0.5)==1)
				new_trans[i] = true;
	}
	if(_options._exonAware) {
		//reads might not cover all exons
		//if exon aware all x=1 make sense (not random assignment)
		for(UInt i=0; i<_ilpData.locus.size(); i++)
			if((int)(_cplex.getValue(_x[i])+0.5)==1)
				new_trans[i] = true;
	}

  if (Parameters::verbose)
  {
    std::cout << "added following transcript after ILP pricing: " << std::endl;
    for(UInt i=0; i<_ilpData.locus.size(); i++)
      if(new_trans[i])
        std::cout << _ilpData.locus[i] << " ";
    std::cout << std::endl;
  }
}


//void PriceIlp::updateSgComp(const std::vector<int>& explained)
bool PriceIlp::updateSgComp()
{
	//cout << "number of relaxed edges before: " << _relaxed_edges.size() << endl;
	std::vector<std::pair<UInt,UInt> > to_undo_relax;

	//over all relaxed edges
	map<std::pair<UInt,UInt>, UInt>::const_iterator it;
	for(it=_relaxed_edges.begin(); it!=_relaxed_edges.end(); it++)
	{
		bool spanning_unexpl_remains = false;
		for(int k=0; k<_ilpData.cov_expl.size(); k++)
		{
			if(!_ilpData.cov_expl[k])
				if(_ilpData.covers[k].exons_1.back()<=it->first.first && _ilpData.covers[k].exons_2.front()>=it->first.second)
				{
					spanning_unexpl_remains = true;
					break;
				}
		}
		if(!spanning_unexpl_remains)
			to_undo_relax.push_back(it->first);
	}

	std::vector<std::pair<UInt,UInt> >::const_iterator undo_it;
	for(undo_it=to_undo_relax.begin(); undo_it!=to_undo_relax.end(); undo_it++)
	{
    if (Parameters::verbose)
      cout << "undo relaxation for: " << undo_it->first << "-" << undo_it->second << endl;
    
		_model.add(_rx_sgc[_relaxed_edges.at(*undo_it)]==0);
		_relaxed_edges.erase(*undo_it);
	}
  if (Parameters::verbose)
    cout << "number of relaxed edges after: " << _relaxed_edges.size() << endl;
	//_cplex.exportModel("test.lp");
	if(_relaxed_edges.empty())
		return false;
	else
		return true;
}

void PriceIlp::initSgComp()
{
	if(_options._exonAware)
	{
		IloExpr sumofx(_env);
		for(UInt i=0; i<_ilpData.exon_cluster.size(); i++)
		{
			std::set<UInt> right_boundaries;
			for(UInt j=0; j<_ilpData.exon_cluster[i].size(); j++)
			{
				right_boundaries.insert(_ilpData.exon_cluster[i][j].second);
			}
			for(UInt ii=i+1; ii<_ilpData.exon_cluster.size(); ii++)
			{
				for(UInt jj=0; jj<_ilpData.exon_cluster[ii].size(); jj++)
				{
					for(std::set<UInt>::iterator rb_it=right_boundaries.begin(); rb_it!=right_boundaries.end(); ++rb_it)
					{
						//std::cout << _ilpData.locus[_ilpData.exon_cluster[ii][jj]] << " "  << *rb_it << std::endl;
						UInt left_seg_idx = _ilpData.exon_cluster[ii][jj].first;
						if(_ilpData.sg.findEdge(*rb_it, _ilpData.locus[left_seg_idx])==0)
						{
							bool spanned_by_unexpl = false;
							if(Parameters::max_nof_relaxed_edges)
							{
								for(int k=0; k<_ilpData.cov_expl.size(); k++)
								{
									if(!_ilpData.cov_expl[k])
									{
										if(_ilpData.covers[k].exons_1.back()<=*rb_it && _ilpData.covers[k].exons_2.front()>=_ilpData.locus[left_seg_idx])
										{
											//cout << " relax: " << *rb_it << " " << _ilpData.locus[left_seg_idx] << endl;
											spanned_by_unexpl = true;
											break;
										}
										//cout << _ilpData.covers[k].exons_1.back() << " " << *rb_it << " " << _ilpData.covers[k].exons_2.front() <<  " " << _ilpData.locus[left_seg_idx] << endl;
									}
								}
								//if(spanned_by_unexpl)
									//continue;
							}

							//std::cout << "edge does not exist" << std::endl;
							sumofx.clear();
							//since edge between right boundary of first exon and left boundary of second exon does not exist in sg
							//we have to forbid the corresponding ILP solution?

							sumofx += _x[left_seg_idx];
							sumofx += _x[_exon_order[*rb_it]];
							for(int z=_exon_order[*rb_it]+1; z<left_seg_idx; ++z)
								sumofx += 1-_x[z];
							if(Parameters::max_nof_relaxed_edges && spanned_by_unexpl)
							{
								_rx_sgc.push_back(IloBoolVar(_env));
								_model.add(sumofx-(left_seg_idx-_exon_order[*rb_it]+1-1)<=_rx_sgc.back());
                if (Parameters::verbose)
                  cout << "potentially relaxing (exon exon): " << *rb_it << "-" << _ilpData.locus[left_seg_idx] << endl;
								_relaxed_edges[std::make_pair(*rb_it,_ilpData.locus[left_seg_idx])] = _rx_sgc.size()-1;
							}
							else
								_model.add(sumofx-(left_seg_idx-_exon_order[*rb_it]+1-1)<=0);
						}
					}
				}
			}
			for(UInt k=0; k<_ilpData.intron_retention.size(); ++k) //necessary since intron retention may end in interior of another intron retention,
			{													   //such that no "normal" exon exists with the same boundary
				for(std::set<UInt>::iterator rb_it=right_boundaries.begin(); rb_it!=right_boundaries.end(); ++rb_it)
				{
					UInt intret_begin = _ilpData.locus[_ilpData.intron_retention[k].first];
					if(intret_begin > *rb_it && _ilpData.sg.findEdge(*rb_it, intret_begin)==0)
					{
						bool spanned_by_unexpl = false;
						if(Parameters::max_nof_relaxed_edges)
						{
							for(int l=0; l<_ilpData.cov_expl.size(); l++)
							{
								if(!_ilpData.cov_expl[l])
								{
									if(_ilpData.covers[l].exons_1.back()<=*rb_it && _ilpData.covers[l].exons_2.front()>=intret_begin)
									{
										//cout << " relax: " << *rb_it << " " << intret_begin << endl;
										spanned_by_unexpl = true;
										break;
									}
									//cout << _ilpData.covers[k].exons_1.back() << " " << *rb_it << " " << _ilpData.covers[k].exons_2.front() <<  " " << _ilpData.locus[left_seg_idx] << endl;
								}
							}
							//if(spanned_by_unexpl)
								//continue;
						}

						sumofx.clear();
						sumofx += _x[_exon_order[intret_begin]];
						sumofx += _x[_exon_order[*rb_it]];
						for(int z=_exon_order[*rb_it]+1; z<_exon_order[intret_begin]; ++z)
							sumofx += 1-_x[z];
						if(Parameters::max_nof_relaxed_edges && spanned_by_unexpl)
						{
							_rx_sgc.push_back(IloBoolVar(_env));
							_model.add(sumofx-(_exon_order[intret_begin]-_exon_order[*rb_it]+1-1)<=_rx_sgc.back());
              
              if (Parameters::verbose)
                cout << "potentially relaxing (exon, intron_ret):" << *rb_it << "-" << intret_begin << endl;
							
              _relaxed_edges[std::make_pair(*rb_it,intret_begin)] = _rx_sgc.size()-1;
						}
						else
							_model.add(sumofx-(_exon_order[intret_begin]-_exon_order[*rb_it]+1-1)<=0);
					}
				}
			}
		}
		for(UInt i=0; i<_ilpData.intron_retention.size(); ++i)
		{
			UInt intret_end = _ilpData.intron_retention[i].second;
			for(UInt j=i+1; j<_ilpData.intron_retention.size(); ++j)
			{
				UInt intret_start = _ilpData.locus[_ilpData.intron_retention[j].first];
				if(intret_start >  intret_end && _ilpData.sg.findEdge(intret_end, intret_start)==0)
				{
					bool spanned_by_unexpl = false;
					if(Parameters::max_nof_relaxed_edges)
					{
						for(int l=0; l<_ilpData.cov_expl.size(); l++)
						{
							if(!_ilpData.cov_expl[l])
							{
								if(_ilpData.covers[l].exons_1.back()<=intret_end && _ilpData.covers[l].exons_2.front()>=intret_start)
								{
									//cout << " relax: " << intret_end << " " << intret_start << endl;
									spanned_by_unexpl = true;
									break;
								}
								//cout << _ilpData.covers[k].exons_1.back() << " " << *rb_it << " " << _ilpData.covers[k].exons_2.front() <<  " " << _ilpData.locus[left_seg_idx] << endl;
							}
						}
						//if(spanned_by_unexpl)
							//continue;
					}

					sumofx.clear();
					sumofx += _x[_exon_order[intret_start]];
					sumofx += _x[_exon_order[intret_end]];
					for(int z=_exon_order[intret_end]+1; z<_exon_order[intret_start]; ++z)
						sumofx += 1-_x[z];
					if(Parameters::max_nof_relaxed_edges && spanned_by_unexpl)
					{
						_rx_sgc.push_back(IloBoolVar(_env));
						_model.add(sumofx-(_exon_order[intret_start]-_exon_order[intret_end]+1-1)<=_rx_sgc.back());
            
            if (Parameters::verbose)
              cout << "potentially relaxing (intron, intron): " << intret_end << "-" << intret_start << endl;
            
						_relaxed_edges[std::make_pair(intret_end,intret_start)] = _rx_sgc.size()-1;
					}
					else
						_model.add(sumofx-(_exon_order[intret_start]-_exon_order[intret_end]+1-1)<=0);
				}
			}
			for(UInt j=0; j<_ilpData.exon_cluster.size(); j++)
			{
				for(UInt jj=0; jj<_ilpData.exon_cluster[j].size(); jj++)
				{
					UInt left_bound = _ilpData.locus[_ilpData.exon_cluster[j][jj].first];
					if(left_bound > intret_end && _ilpData.sg.findEdge(intret_end, left_bound)==0)
					{
						bool spanned_by_unexpl = false;
						if(Parameters::max_nof_relaxed_edges)
						{
							for(int l=0; l<_ilpData.cov_expl.size(); l++)
							{
								if(!_ilpData.cov_expl[l])
								{
									if(_ilpData.covers[l].exons_1.back()<=intret_end && _ilpData.covers[l].exons_2.front()>=left_bound)
									{
										//cout << " relax: " << intret_end  << " " << left_bound << endl;
										spanned_by_unexpl = true;
										break;
									}
									//cout << _ilpData.covers[k].exons_1.back() << " " << *rb_it << " " << _ilpData.covers[k].exons_2.front() <<  " " << _ilpData.locus[left_seg_idx] << endl;
								}
							}
							//if(spanned_by_unexpl)
								//continue;
						}

						sumofx.clear();
						sumofx += _x[_exon_order[left_bound]];
						sumofx += _x[_exon_order[intret_end]];
						for(int z=_exon_order[intret_end]+1; z<_exon_order[left_bound]; ++z)
							sumofx += 1-_x[z];
						if(Parameters::max_nof_relaxed_edges && spanned_by_unexpl)
						{
							_rx_sgc.push_back(IloBoolVar(_env));
							_model.add(sumofx-(_exon_order[left_bound]-_exon_order[intret_end]+1-1)<=_rx_sgc.back());
              
              if (Parameters::verbose)
                cout << "potentially relaxing: (intron, exon)" << intret_end << "-" << left_bound << endl;
              
							_relaxed_edges[std::make_pair(intret_end,left_bound)] = _rx_sgc.size()-1;
						}
						else
							_model.add(sumofx-(_exon_order[left_bound]-_exon_order[intret_end]+1-1)<=0);
					}
				}
			}
		}
		if(Parameters::max_nof_relaxed_edges && _rx_sgc.size()>0)
		{
			IloExpr sumofrxsgc(_env);
			for(UInt z=0; z<_rx_sgc.size(); z++)
				sumofrxsgc += _rx_sgc[z];
			_model.add(sumofrxsgc<=Parameters::max_nof_relaxed_edges);
			//cout << " bounding " << _rx_sgc.size() << "potential relaxations by: " << relax_unexpl << endl;
		}
	} else
	{
		std::cerr << "splice graph compatibility only implemented for exon aware mode" << std::endl;
		exit(0);
	}
}

const vector<bool>& PriceIlp::getCoverValid() const
{
	return _cover_valid;
}
